/* Shadekeep Mapping Tool ©2018-present Howard A. Kistler */
/* Shamat II ©2024-present Howard A. Kistler */

/* Constants */

const VERSION_NUMBER = '1.0'; // current version number of Shamat II
const LEGACY_VERSION_NUMBER = '2.4'; // last release version of original Shamat software

const PALETTE_SIZE = 41;
const CUSTOM_PALETTE_OFFSET = 100;
const CUSTOM_PALETTE_SIZE = 30;
const MATERIAL_BUTTON_SIZE = '40px';
const MATERIAL_BUTTON_SVG_VIEWBOX = '0 0 91.52 92.04';

const MODE_PAINT = 0;
const MODE_LINE  = 1;
const MODE_RECT  = 2;
const MODE_AREA  = 3;
const MODE_SKIP  = 4;
const MODE_SMART = 5;
const MODE_LROOM = 6;
const MODE_SROOM = 7;
const MODE_CLONE = 30;
const MODE_STAMP = 33;
const MODE_ERASE = 90;
const MODE_BULKE = 99; 

const GRAIN_VERTICAL   = 0;
const GRAIN_HORIZONTAL = 1;

const LINEWEIGHT_PROPORTIONAL = 0;
const LINEWEIGHT_HAIRLINE     = 1;

const LAYER_BASE     = 0;
const LAYER_TERRAIN  = 1;
const LAYER_OVERLAY  = 2;
const LAYER_FEATURE  = 3;
const LAYER_DECOR    = 4;
const LAYER_SYMBOL   = 5;
const LAYER_MEGAHEX  = 6;
const LAYER_ANNOTATE = 7;

const HEX_EMPTY = 0;
const HEX_WHOLE = 1;
const HEX_FHALF = 10;
const HEX_PHALF = 20;
const HEX_APEX  = 30;
const HEX_BULK  = 40;
const HEX_CHANL = 50;
const HEX_JUNCT = 60;
const HEX_CORNR = 70;
const HEX_FINSH = 80;
const HEX_WIDE  = 90;
const HEX_JOIN  = 100;
const HEX_SLICE = 110;
const HEX_TWALL = 120;
const HEX_TEDGE = 130;
const HEX_TCORN = 140;
const HEX_BORDR = 200;
const HEX_LRSID = 210;
const HEX_LRWAL = 220;
const HEX_LRCNR = 230;
const HEX_SRSID = 240;
const HEX_SRCNR = 250;
const HEX_PSEUD = 400;

const MATERIAL_EMPTY = 0;

// Glyph number boundaries, bottom (B) and top (T)
// Note that there is latitude to shift these in future, if one range needs more glyphs
const FEATURE_BOUND_B = 0;
const FEATURE_BOUND_T = 1999;
const TERRAIN_BOUND_B = 2000;
const TERRAIN_BOUND_T = 5999;
const DECOR_BOUND_B   = 6000;
const DECOR_BOUND_T   = 7999;
const SYMBOL_BOUND_B  = 8000;
const SYMBOL_BOUND_T  = 8999;

const FEATURE_EMPTY   = 0;
const FEATURE_DOOR    = 10;
const FEATURE_STAIRS  = 20;
const FEATURE_STAIR   = 30;
const FEATURE_SHAFT   = 40;
const FEATURE_UNDERS  = 50;
const FEATURE_UNDER   = 60;
const FEATURE_BANDS   = 70;
const FEATURE_BAND    = 80;
const FEATURE_DNDOOR  = 90;
const FEATURE_SPIRAL  = 110;
const FEATURE_WINDOW  = 120;
const FEATURE_MIDOOR  = 130;
const FEATURE_MIWIND  = 140;
const FEATURE_AXONS   = 150;
const FEATURE_AXON    = 160;
const FEATURE_TDOOR   = 170;
const FEATURE_TWIND   = 180;
const FEATURE_EDOOR   = 190;
const FEATURE_EWIND   = 200;
const FEATURE_ADOOR   = 210;
const FEATURE_AWIND   = 220;
const FEATURE_PILLAR  = 300;

const TERRAIN_EMPTY    = 0;
const TERRAIN_BOUNDER  = 2000;
const TERRAIN_WATER    = 2100;
const TERRAIN_MOUNTOUT = 2200;
const TERRAIN_MOUNTFIL = 2250;
const TERRAIN_TREES    = 2300;
const TERRAIN_SWAMP    = 2400;
const TERRAIN_PLAINS   = 2500;
const TERRAIN_DESERT   = 2600;
const TERRAIN_CAPITOL  = 3000;
const TERRAIN_CITY     = 3100;
const TERRAIN_VILLAGE  = 3200;
const TERRAIN_FORT     = 3300;
const TERRAIN_MINE     = 3400;
const TERRAIN_RUIN     = 3500;
const TERRAIN_INTEREST = 4000;

const DECOR_EMPTY    = 0;
const DECOR_TRAP     = 6000;
const DECOR_TRAPDOOR = 6100;
const DECOR_FLOORPIT = 6150;
const DECOR_TRIPWIRE = 6200;
const DECOR_CHESTL   = 6300;
const DECOR_CHESTS   = 6350;
const DECOR_KEY      = 6400;
const DECOR_PENTAGRM = 6500;
const DECOR_GATE     = 6550;
const DECOR_LIGHT    = 6600;
const DECOR_SCONCE   = 6650;

const SYMBOL_EMPTY   = 0;
const SYMBOL_ARROWEL = 8000;
const SYMBOL_ARROWES = 8050;
const SYMBOL_ARROWVL = 8100;
const SYMBOL_ARROWVS = 8150;
const SYMBOL_PROHIBL = 8200;
const SYMBOL_PROHIBS = 8250;
const SYMBOL_MARKER  = 8300;
const SYMBOL_EVENT   = 8400;
const SYMBOL_ACCENTL = 8500;
const SYMBOL_ACCENTC = 8510;
const SYMBOL_ACCENTT = 8520;
const SYMBOL_ACCENTD = 8530;
const OLD_SYMBOL_MARKER = 100;
const OLD_SYMBOL_EVENT  = 1000;

const ANNOTATION_BASE = 9000;

const PREVIEW_SCALE = 16;
const PREVIEW_SIZEX = 20;
const PREVIEW_SIZEY = 20;

const COMPONENT_UNSET = -1;
const FEATURE_UNSET   = -9;

const SHIFTMAP_LEFT   = 'sml';
const SHIFTMAP_RIGHT  = 'smr';
const SHIFTMAP_UP     = 'smu';
const SHIFTMAP_DOWN   = 'smd';
const SHIFTMAP_RIGHTK = 'smrk';
const SHIFTMAP_DOWNK  = 'smdk';

const NEWLINE = "\r\n";

const MAPFILE_EXTENSION = '.skmp';
const COLORFILE_EXTENSION = '.skcc';
const EXPORTMAP_EXTENSION = '.json';
const OLD_MAPFILE_EXTENSION = '.skmr';

// Data file constants
const FILEPREAMBLE = 'data:application/json;charset=utf-8,';
const FILEKEYVERS  = 'shamat2ver';
const FILEKEYSIZE  = 'size';
const FILEKEYROWS  = 'rows';
const FILEKEYCOLS  = 'cols';
const FILEKEYNAME  = 'name';
const FILEKEYAUTH  = 'author';
const FILEKEYITER  = 'iteration';
const FILEKEYGRAN  = 'grain';
const FILEKEYLWGT  = 'lineweight';
const FILEKEYSHAD  = 'shade';
const FILEKEYZOOM  = 'zoom';
const FILEKEYTRNO  = 'translucenton';
const FILEKEYTRNA  = 'translucentamt';
const FILEKEYMHEX  = 'mhex';
const FILEKEYMHWT  = 'mhwt';
const FILEKEYMHDT  = 'mhdt';

const DATAHEX      = 'hexes';
const HKEY_X       = 'x';
const HKEY_Y       = 'y';
const HKEY_BASE    = 'b';
const HKEY_TERRAIN = 't';
const HKEY_OVERLAY = 'o';
const HKEY_FEATURE = 'f';
const HKEY_DECOR   = 'd';
const HKEY_SYMBOL  = 's';
const HKEY_ANNOT   = 'a';

const DATAPAIR_CTYPE = 'c';
const DATAPAIR_ROTAT = 'r';
const DATAPAIR_MATER = 'm';

const CUSTOMCOLORS = 'custclrs';

// Backdrop image constants
const BDROPOBJ   = 'backdrop';       // Backdrop data file section header
const BDROPDIMNW = 'naturalWidth';   // img.naturalWidth
const BDROPDIMNH = 'naturalHeight';  // img.naturalHeight
const BDROPDIMRW = 'renderWidth';    // img.width (rendered width)
const BDROPDIMRH = 'renderHeight';   // img.height (rendered height)
const BDROPOFFX  = 'offsetX';        // backdropOffset.x
const BDROPOFFY  = 'offsetY';        // backdropOffset.y
const BDROPDATA  = 'imgdata';        // encoded image data block

// Shamat I file token compatability
const LEGACYFILEPREAMBLE = 'data:text/csv;charset=utf-8,';
const LEGACYFILEKEYVERS  = 'SHAMATVER:';
const LEGACYFILEHEADER   = 'SHADEKEEP:MAP:RAW';
const LEGACYDATAHEADER   = 'X,Y,[Terrains],{Overlays},(Features)';
const LEGACYFILEKEYSIZE  = 'SIZE:';
const LEGACYDATAHEX      = 'H:';

// Shamat I backdrop image constants
const LEGACYBDROPHDR   = '[-BACKDROPIMAGE-]'; // Backdrop data file section header
const LEGACYBDROPDIMNW = 'bdropdimnw';        // img.naturalWidth
const LEGACYBDROPDIMNH = 'bdropdimnh';        // img.naturalHeight
const LEGACYBDROPDIMRW = 'bdropdimrw';        // img.width (rendered width)
const LEGACYBDROPDIMRH = 'bdropdimrh';        // img.height (rendered height)
const LEGACYBDROPOFFX  = 'bdropoffx';         // backdropOffset.x
const LEGACYBDROPOFFY  = 'bdropoffy';         // backdropOffset.y
const LEGACYBDROPDATA  = 'bdropdata';         // encoded image data block

const SEP_DATA = ',';
const SEP_LIST = '|';
const SEP_PAIR = ':';

const ZOOM_MIN  = 4;
const ZOOM_MAX  = 64;
const ZOOM_STEP = 2;

// The six cardinal directions relating to hex orientation
const DIR_HXC0 = 0; // Above in vertical grain, Upper Left in horizontal grain
const DIR_HXC1 = 1; // Upper Right in vertical grain, Upper Right in horizontal grain
const DIR_HXC2 = 2; // Lower Right in vertical grain, Right in horizontal grain
const DIR_HXC3 = 3; // Below in vertical grain, Lower Right in horizontal grain
const DIR_HXC4 = 4; // Lower Left in vertical grain, Lower Left in horizontal grain
const DIR_HXC5 = 5; // Upper Left in vertical grain, Left in horizontal grain
const NO_DIR   = -9;
const OFFGRID  = -1;

const MAX_GRID = 500;

const ANGLE_STEP = ((Math.PI * 2) / 360.0);
const CALCSQRT = Math.sqrt(3);
const CALCTHIN = CALCSQRT / 2;

const DEFAULT_LINE_THICKNESS = 1;

const MEGAHEX_TYPE_STANDARD = 0;
const MEGAHEX_TYPE_SQUARE = 1;
const MEGAHEX_WEIGHT_DEFAULT = 2;
const MEGADOT_WEIGHT_DEFAULT = 2;

const CARDINAL_N = 'N';
const CARDINAL_S = 'S';
const CARDINAL_E = 'E';
const CARDINAL_W = 'W';

const CLR_R = 'R';
const CLR_G = 'G';
const CLR_B = 'B';

const OPAQUE    = 1.0;
var   TRANSLUCE = 0.5;

/* Interface Elements */

const TOOL_BUTTON_PREFIX  = 'toolbtn-';
const TOOL_BUTTON_MAPSETS = TOOL_BUTTON_PREFIX + 'mapsets';
const TOOL_BUTTON_LAYERS  = TOOL_BUTTON_PREFIX + 'layers';
const TOOL_BUTTON_TOOLS   = TOOL_BUTTON_PREFIX + 'tools';
const TOOL_BUTTON_COMPOSE = TOOL_BUTTON_PREFIX + 'compose';
const TOOL_BUTTON_MATSTND = TOOL_BUTTON_PREFIX + 'maters';
const TOOL_BUTTON_MATCUST = TOOL_BUTTON_PREFIX + 'materc';

const LAYER_BUTTON_PREFIX   = 'layer';
const LAYER_BUTTON_BASE     = LAYER_BUTTON_PREFIX + LAYER_BASE;
const LAYER_BUTTON_TERRAIN  = LAYER_BUTTON_PREFIX + LAYER_TERRAIN;
const LAYER_BUTTON_OVERLAY  = LAYER_BUTTON_PREFIX + LAYER_OVERLAY;
const LAYER_BUTTON_FEATURE  = LAYER_BUTTON_PREFIX + LAYER_FEATURE;
const LAYER_BUTTON_DECOR    = LAYER_BUTTON_PREFIX + LAYER_DECOR;
const LAYER_BUTTON_SYMBOL   = LAYER_BUTTON_PREFIX + LAYER_SYMBOL;
const LAYER_BUTTON_ANNOTATE = LAYER_BUTTON_PREFIX + LAYER_ANNOTATE;
const LAYER_BUTTON_MEGAHEX  = LAYER_BUTTON_PREFIX + LAYER_MEGAHEX;

const PAGE_BLOCK_MAPSETS  = 'pblockMapSettings';
const PAGE_BLOCK_LAYERS   = 'pblockLayers';
const PAGE_BLOCK_TOOLS    = 'pblockTools';
const PAGE_BLOCK_COMPOSE  = 'pblockCompose';
const PAGE_BLOCK_BASE     = 'pblockBaseElements';
const PAGE_BLOCK_TERRAIN  = 'pblockTerrain';
const PAGE_BLOCK_OVERLAY  = 'pblockOverlayElements';
const PAGE_BLOCK_FEATURE  = 'pblockFeature';
const PAGE_BLOCK_DECOR    = 'pblockDecor';
const PAGE_BLOCK_SYMBOL   = 'pblockSymbol';
const PAGE_BLOCK_ANNOTATE = 'pblockAnnotate';
const PAGE_BLOCK_MEGAHEX  = 'pblockMegahex';
const PAGE_BLOCK_MATSTND  = 'pblockMaterialStandard';
const PAGE_BLOCK_MATCUST  = 'pblockMaterialCustom';

const LAYER_BUTTONS = [];
LAYER_BUTTONS[LAYER_BASE] = LAYER_BUTTON_BASE;
LAYER_BUTTONS[LAYER_TERRAIN] = LAYER_BUTTON_TERRAIN;
LAYER_BUTTONS[LAYER_OVERLAY] = LAYER_BUTTON_OVERLAY;
LAYER_BUTTONS[LAYER_FEATURE] = LAYER_BUTTON_FEATURE;
LAYER_BUTTONS[LAYER_DECOR] = LAYER_BUTTON_DECOR;
LAYER_BUTTONS[LAYER_SYMBOL] = LAYER_BUTTON_SYMBOL;
LAYER_BUTTONS[LAYER_ANNOTATE] = LAYER_BUTTON_ANNOTATE;
LAYER_BUTTONS[LAYER_MEGAHEX] = LAYER_BUTTON_MEGAHEX

const LAYER_PAGE_BLOCKS = [];
LAYER_PAGE_BLOCKS[LAYER_BASE] = PAGE_BLOCK_BASE;
LAYER_PAGE_BLOCKS[LAYER_TERRAIN] = PAGE_BLOCK_TERRAIN;
LAYER_PAGE_BLOCKS[LAYER_OVERLAY] = PAGE_BLOCK_OVERLAY;
LAYER_PAGE_BLOCKS[LAYER_FEATURE] = PAGE_BLOCK_FEATURE;
LAYER_PAGE_BLOCKS[LAYER_DECOR] = PAGE_BLOCK_DECOR;
LAYER_PAGE_BLOCKS[LAYER_SYMBOL] = PAGE_BLOCK_SYMBOL;
LAYER_PAGE_BLOCKS[LAYER_ANNOTATE] = PAGE_BLOCK_ANNOTATE;
LAYER_PAGE_BLOCKS[LAYER_MEGAHEX] = PAGE_BLOCK_MEGAHEX;

const BUTTON_BLOCKS = [];
BUTTON_BLOCKS[TOOL_BUTTON_MAPSETS] = PAGE_BLOCK_MAPSETS;
BUTTON_BLOCKS[TOOL_BUTTON_LAYERS] = PAGE_BLOCK_LAYERS;
BUTTON_BLOCKS[TOOL_BUTTON_TOOLS] = PAGE_BLOCK_TOOLS;
BUTTON_BLOCKS[TOOL_BUTTON_COMPOSE] = PAGE_BLOCK_COMPOSE;
BUTTON_BLOCKS[TOOL_BUTTON_MATSTND] = PAGE_BLOCK_MATSTND;
BUTTON_BLOCKS[TOOL_BUTTON_MATCUST] = PAGE_BLOCK_MATCUST;
BUTTON_BLOCKS[LAYER_BUTTON_BASE] = PAGE_BLOCK_BASE;
BUTTON_BLOCKS[LAYER_BUTTON_TERRAIN] = PAGE_BLOCK_TERRAIN;
BUTTON_BLOCKS[LAYER_BUTTON_OVERLAY] = PAGE_BLOCK_OVERLAY;
BUTTON_BLOCKS[LAYER_BUTTON_FEATURE] = PAGE_BLOCK_FEATURE;
BUTTON_BLOCKS[LAYER_BUTTON_DECOR] = PAGE_BLOCK_DECOR;
BUTTON_BLOCKS[LAYER_BUTTON_SYMBOL] = PAGE_BLOCK_SYMBOL;
BUTTON_BLOCKS[LAYER_BUTTON_ANNOTATE] = PAGE_BLOCK_ANNOTATE;
BUTTON_BLOCKS[LAYER_BUTTON_MEGAHEX] = PAGE_BLOCK_MEGAHEX;

const MUTABLE_ELEMENTS = [LAYER_BASE, LAYER_OVERLAY, LAYER_FEATURE, LAYER_DECOR, LAYER_SYMBOL];
const TOGGLE_BUTTONS = [TOOL_BUTTON_MAPSETS, TOOL_BUTTON_LAYERS, TOOL_BUTTON_TOOLS, TOOL_BUTTON_COMPOSE];
const MATER_BUTTONS = [TOOL_BUTTON_MATSTND, TOOL_BUTTON_MATCUST];
const MATER_BLOCKS = [PAGE_BLOCK_MATSTND, PAGE_BLOCK_MATCUST];

const LAYER_TOGGLE_PREFIX = 'layertoggle';
const MATERIAL_PREFIX     = 'material';
const CANVAS_PREVIEW      = 'canvas-preview';

const SVG_DECLARATION = 'xmlns="http://www.w3.org/2000/svg" xml:space="preserve"  version="1.1" style="shape-rendering:geometricPrecision; text-rendering:geometricPrecision; image-rendering:optimizeQuality; fill-rule:evenodd; clip-rule:evenodd" xmlns:xlink="http://www.w3.org/1999/xlink"';

const CSS_BLOCK_SHOW          = 'block';
const CSS_BLOCK_HIDE          = 'none';
const CSS_BLOCK_INLINE        = 'inline-block';
const CSS_TOOL_BUTTON         = 'toolicon';
const CSS_TOOL_BUTTON_ACTIVE  = 'toolicon-active';
const CSS_PAL_BUTTON          = 'palettetool';
const CSS_PAL_BUTTON_ACTIVE   = 'palettetool-active';
const CSS_PAL_BUTTON_DISABLED = 'palettetool-disabled';
const CSS_PAL_BUTTON_CUSTOM   = 'palettetool-custom';
const CSS_LYR_BUTTON          = 'layerbutton';
const CSS_LYR_BUTTON_ACTIVE   = 'layerbutton-active';

const THEME_LIGHT = 'light';
const THEME_DARK  = 'dark';

const SIDEBAR_LEFT  = 'left';
const SIDEBAR_RIGHT = 'right';

const EXPORT_NAMES_HEXES = [];
EXPORT_NAMES_HEXES[HEX_EMPTY] = 'empty';
EXPORT_NAMES_HEXES[HEX_WHOLE] = 'whole';
EXPORT_NAMES_HEXES[HEX_FHALF] = 'flat_half';
EXPORT_NAMES_HEXES[HEX_PHALF] = 'point_half';
EXPORT_NAMES_HEXES[HEX_APEX] = 'apex';
EXPORT_NAMES_HEXES[HEX_BULK] = 'bulk';
EXPORT_NAMES_HEXES[HEX_CHANL] = 'channel';
EXPORT_NAMES_HEXES[HEX_JUNCT] = 'junction';
EXPORT_NAMES_HEXES[HEX_CORNR] = 'corner';
EXPORT_NAMES_HEXES[HEX_FINSH] = 'finish';
EXPORT_NAMES_HEXES[HEX_WIDE] = 'wide';
EXPORT_NAMES_HEXES[HEX_JOIN] = 'join';
EXPORT_NAMES_HEXES[HEX_SLICE] = 'slice';
EXPORT_NAMES_HEXES[HEX_TWALL] = 'thin_wall';
EXPORT_NAMES_HEXES[HEX_TEDGE] = 'thin_edge';
EXPORT_NAMES_HEXES[HEX_TCORN] = 'thin_corner';
EXPORT_NAMES_HEXES[HEX_BORDR] = 'hex_border';
EXPORT_NAMES_HEXES[HEX_LRSID] = 'room_side';
EXPORT_NAMES_HEXES[HEX_LRWAL] = 'room_wall';
EXPORT_NAMES_HEXES[HEX_LRCNR] = 'room_corner';
EXPORT_NAMES_HEXES[HEX_SRSID] = 'filled_side';
EXPORT_NAMES_HEXES[HEX_SRCNR] = 'filled_corner';
EXPORT_NAMES_HEXES[HEX_PSEUD] = 'mega_unit';

const EXPORT_NAMES_OBJECTS = [];
EXPORT_NAMES_OBJECTS[0] = 'empty';
EXPORT_NAMES_OBJECTS[FEATURE_DOOR] = 'door';
EXPORT_NAMES_OBJECTS[FEATURE_STAIRS] = 'stairs';
EXPORT_NAMES_OBJECTS[FEATURE_STAIR] = 'stair';
EXPORT_NAMES_OBJECTS[FEATURE_SHAFT] = 'shaft';
EXPORT_NAMES_OBJECTS[FEATURE_UNDERS] = 'unders';
EXPORT_NAMES_OBJECTS[FEATURE_UNDER] = 'under';
EXPORT_NAMES_OBJECTS[FEATURE_BANDS] = 'bands';
EXPORT_NAMES_OBJECTS[FEATURE_BAND] = 'band';
EXPORT_NAMES_OBJECTS[FEATURE_DNDOOR] = 'half_door';
EXPORT_NAMES_OBJECTS[FEATURE_SPIRAL] = 'spiral_stairs';
EXPORT_NAMES_OBJECTS[FEATURE_WINDOW] = 'window';
EXPORT_NAMES_OBJECTS[FEATURE_MIDOOR] = 'mid_door';
EXPORT_NAMES_OBJECTS[FEATURE_MIWIND] = 'mid_window';
EXPORT_NAMES_OBJECTS[FEATURE_AXONS] = 'axons';
EXPORT_NAMES_OBJECTS[FEATURE_AXON] = 'axon';
EXPORT_NAMES_OBJECTS[FEATURE_TDOOR] = 'thin_door';
EXPORT_NAMES_OBJECTS[FEATURE_TWIND] = 'thin_window';
EXPORT_NAMES_OBJECTS[FEATURE_EDOOR] = 'edge_door';
EXPORT_NAMES_OBJECTS[FEATURE_EWIND] = 'edge_window';
EXPORT_NAMES_OBJECTS[FEATURE_ADOOR] = 'angle_door';
EXPORT_NAMES_OBJECTS[FEATURE_AWIND] = 'angle_window';
EXPORT_NAMES_OBJECTS[FEATURE_PILLAR] = 'pillar';
EXPORT_NAMES_OBJECTS[TERRAIN_BOUNDER] = 'bounder';
EXPORT_NAMES_OBJECTS[TERRAIN_WATER] = 'water';
EXPORT_NAMES_OBJECTS[TERRAIN_MOUNTOUT] = 'mountain_outline';
EXPORT_NAMES_OBJECTS[TERRAIN_MOUNTFIL] = 'mountain_fill';
EXPORT_NAMES_OBJECTS[TERRAIN_TREES] = 'forest';
EXPORT_NAMES_OBJECTS[TERRAIN_SWAMP] = 'swamp';
EXPORT_NAMES_OBJECTS[TERRAIN_PLAINS] = 'plains';
EXPORT_NAMES_OBJECTS[TERRAIN_DESERT] = 'desert';
EXPORT_NAMES_OBJECTS[TERRAIN_CAPITOL] = 'capitol';
EXPORT_NAMES_OBJECTS[TERRAIN_CITY] = 'city';
EXPORT_NAMES_OBJECTS[TERRAIN_VILLAGE] = 'village';
EXPORT_NAMES_OBJECTS[TERRAIN_FORT] = 'fort';
EXPORT_NAMES_OBJECTS[TERRAIN_MINE] = 'mine';
EXPORT_NAMES_OBJECTS[TERRAIN_RUIN] = 'ruin';
EXPORT_NAMES_OBJECTS[TERRAIN_INTEREST] = 'interest';
EXPORT_NAMES_OBJECTS[DECOR_TRAP] = 'trap';
EXPORT_NAMES_OBJECTS[DECOR_TRAPDOOR] = 'trapdoor';
EXPORT_NAMES_OBJECTS[DECOR_FLOORPIT] = 'floor_pit';
EXPORT_NAMES_OBJECTS[DECOR_TRIPWIRE] = 'tripwire';
EXPORT_NAMES_OBJECTS[DECOR_CHESTL] = 'chest_large';
EXPORT_NAMES_OBJECTS[DECOR_CHESTS] = 'chest_small';
EXPORT_NAMES_OBJECTS[DECOR_KEY] = 'key';
EXPORT_NAMES_OBJECTS[DECOR_PENTAGRM] = 'pentagram';
EXPORT_NAMES_OBJECTS[DECOR_GATE] = 'gate';
EXPORT_NAMES_OBJECTS[DECOR_LIGHT] = 'light';
EXPORT_NAMES_OBJECTS[DECOR_SCONCE] = 'sconce';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWEL] = 'arrow_edge_large';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWES] = 'arrow_edge_small';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWVL] = 'arrow_vertex_large';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWVS] = 'arrow_vertex_small';
EXPORT_NAMES_OBJECTS[SYMBOL_PROHIBL] = 'prohibit_large';
EXPORT_NAMES_OBJECTS[SYMBOL_PROHIBS] = 'prohibit_small';
EXPORT_NAMES_OBJECTS[SYMBOL_MARKER] = 'marker';
EXPORT_NAMES_OBJECTS[SYMBOL_EVENT] = 'event';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTL] = 'accent_linear';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTC] = 'accent_circle';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTT] = 'accent_triangle';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTD] = 'accent_diamond';
EXPORT_NAMES_OBJECTS[OLD_SYMBOL_MARKER] = 'marker';
EXPORT_NAMES_OBJECTS[OLD_SYMBOL_EVENT] = 'event';

// Debugging Flag
const DEBUG_MODE = true;

/* Collections */

var hexBackground = '#ffffff';

var materialDef = [
	hexBackground,
	'#000000',
	'#ffffff',
	'#cccccc',
	'#ccbcab',
	'#e6cda6',
	'#99a8c2',
	'#add498',
	'#7fc5e8',
	'#aaaaaa',
	'#e3e395',
	'#f78b0f',
	'#c693e6',
	'#5e88f2',
	'#6de8ad',
	'#875822',
	'#f05757',
	'#b07f71',
	'#ba3e12',
	'#cfc50b',
	'#4e9e29',
	'#5a79ad',
	'#6c6c6c',
	'#e394a5',
	'#f6bfff',
	'#bf9567',
	'#db6300',
	'#f0f01f',
	'#a0e35d',
	'#324ccf',
	'#00afef',
	'#ec268f',
	'#8e5aed',
	'#ffffff',
	'#e6e6e6',
	'#cccccc',
	'#b3b3b3',
	'#999999',
	'#808080',
	'#666666',
	'#4d4d4d',
	'#333333',
	'#191919',
	'#000000'
];

var materialNames = [
	'Empty',
	'Black',
	'White',
	'Stone',
	'Earth',
	'Wood',
	'Metal',
	'Verdigris',
	'Azure',
	'Iron',
	'Yellow',
	'Orange',
	'Purple',
	'Blue',
	'Chartreuse',
	'Brown',
	'Crimson',
	'Terracotta',
	'Umber',
	'Gold',
	'Viridian',
	'Lapis',
	'Basalt',
	'Rose',
	'Lilac',
	'Bark',
	'Ochre',
	'Sun',
	'Green',
	'Cobalt',
	'Cyan',
	'Magenta',
	'Violet',
	'Pure White',
	'10% Grey',
	'20% Grey',
	'30% Grey',
	'40% Grey',
	'50% Grey',
	'60% Grey',
	'70% Grey',
	'80% Grey',
	'90% Grey',
	'Pure Black'
];

for(cc = 0; cc < CUSTOM_PALETTE_SIZE; cc++) {
	materialDef[CUSTOM_PALETTE_OFFSET + cc] = '#ffffff';
	materialNames[CUSTOM_PALETTE_OFFSET + cc] = 'Custom' + (cc + 1);
}

var materialSpectrumOrder = [
	 0,
	 1,  2,  3,  9, 22,  6,
	 5,	 4, 25, 15, 17, 23,
	16, 18, 26, 11, 10, 27,
	19,  7, 28, 14, 20,  8,
	30, 13, 21, 29, 24, 12,
	32, 31, 34, 35, 36, 37,
	38, 39, 40, 41, 42
];

var materialSequence = [];
for(m = 0; m < materialDef.length; m++) {
	if(m < materialSpectrumOrder.length) {
		materialSequence[m] = materialSpectrumOrder[m];
	}
	else {
		materialSequence[m] = m;
	}
}

var materialToEdit = null;

var annoSpecialChars = [ 64, 46, 44, 58, 59, 33, 63, 42, 34, 35, 43, 45, 177, 61, 124, 166, 60, 62, 40, 41, 91, 93, 123, 125, 171, 187, 126, 37, 169, 182, 167, 164, 36, 162, 163, 165 ];

var layerVisibility = [];
layerVisibility[LAYER_BASE] = true;
layerVisibility[LAYER_OVERLAY] = true;
layerVisibility[LAYER_FEATURE] = true;
layerVisibility[LAYER_TERRAIN] = true;
layerVisibility[LAYER_SYMBOL] = true;
layerVisibility[LAYER_DECOR] = true;
layerVisibility[LAYER_MEGAHEX] = true;
layerVisibility[LAYER_ANNOTATE] = true;
for(l = 0; l < layerVisibility.length; l++) {
	// this loop fills in skipped numeric indices
	layerVisibility[l] = true;
}

/* Objects */

function Point(x, y) {
	this.x = x;
	this.y = y;
}

function HexElement(typ, mater) {
	this.ctype = typ;
	this.material = mater;
	this.setMaterial = function(m) { this.material = m; }
}

function Hex(x, y) {
	this.x = x;
	this.y = y;
	this.bases = [];
	this.terrains = [];
	this.overlays = [];
	this.features = [];
	this.decors = [];
	this.symbols = [];
	this.annotations = [];
	this.clearBases = function(mater) {
		this.bases = [];
		if(typeof mater != "undefined") {
			this.bases.push(new HexElement(HEX_WHOLE, mater));
		}
		else {
			this.bases.push(new HexElement(HEX_EMPTY, MATERIAL_EMPTY));
		}
	}
	this.clearTerrains = function() {
		this.terrains = [];
	}
	this.clearOverlays = function() {
		this.overlays = [];
	}
	this.clearFeatures = function() {
		this.features = [];
	}
	this.clearDecors = function() {
		this.decors = [];
	}
	this.clearSymbols = function() {
		this.symbols = [];
	}
	this.clearAnnotations = function() {
		this.annotations = [];
	}
	this.clearAll = function() {
		this.clearBases();
		this.clearTerrains();
		this.clearOverlays();
		this.clearFeatures();
		this.clearDecors();
		this.clearSymbols();
		this.clearAnnotations();
	}
	this.addBase = function(typ, mater) {
		var needToAdd = true;
		if(this.bases.length > 0) {
			/* whole and empty hexes are special cases, as they obliterate all other previously assigned component types */
			if(typ == HEX_WHOLE || typ == HEX_EMPTY) {
				for(var i = 0; i < this.bases.length; i++) {
					if(this.bases[i].ctype == typ) {
						if(this.bases[i].material == mater) {
							needToAdd = false;
						}
						else {
							this.bases[i].setMaterial(mater);
							needToAdd = false;
						}
					}
					else {
						this.bases = [];
						needToAdd = true;
					}
				}
			}
			else {
				for(var i = 0; i < this.bases.length; i++) {
					if(this.bases[i].ctype == typ) {
						this.bases.splice(i, 1);
						needToAdd = true;
						i = this.bases.length;
					}
				}
			}
		}
		if(needToAdd) {
			this.bases.push(new HexElement(typ, mater));
		}
	}
	this.addTerrain = function(typ, mater) {
		if(this.terrains.length > 0) {
			for(var i = 0; i < this.terrains.length; i++) {
				if(this.terrains[i].ctype == typ) {
					this.terrains.splice(i, 1);
					i = this.terrains.length;
				}
			}
		}
		this.terrains.push(new HexElement(typ, mater));
	}
	this.addOverlay = function(typ, mater) {
		var needToAdd = true;
		if(this.overlays.length > 0) {
			/* whole and empty hexes are special cases, as they obliterate all other previously assigned component types */
			if(typ == HEX_WHOLE || typ == HEX_EMPTY) {
				for(var i = 0; i < this.overlays.length; i++) {
					if(this.overlays[i].ctype == typ) {
						if(this.overlays[i].material == mater) {
							needToAdd = false;
						}
						else {
							this.overlays[i].setMaterial(mater);
							needToAdd = false;
						}
					}
					else {
						this.overlays = [];
						needToAdd = true;
					}
				}
			}
			else {
				for(var i = 0; i < this.overlays.length; i++) {
					if(this.overlays[i].ctype == typ) {
						this.overlays.splice(i, 1);
						needToAdd = true;
						i = this.overlays.length;
					}
				}
			}
		}
		if(needToAdd) {
			this.overlays.push(new HexElement(typ, mater));
		}
	}
	this.addFeature = function(typ, mater) {
		if(this.features.length > 0) {
			for(var i = 0; i < this.features.length; i++) {
				if(this.features[i].ctype == typ) {
					this.features.splice(i, 1);
					i = this.features.length;
				}
			}
		}
		this.features.push(new HexElement(typ, mater));
	}
	this.addDecor = function(typ, mater) {
		if(this.decors.length > 0) {
			for(var i = 0; i < this.decors.length; i++) {
				if(this.decors[i].ctype == typ) {
					this.decors.splice(i, 1);
					i = this.decors.length;
				}
			}
		}
		this.decors.push(new HexElement(typ, mater));
	}
	this.addSymbol = function(typ, mater) {
		if(this.symbols.length > 0) {
			for(var i = 0; i < this.symbols.length; i++) {
				if(this.symbols[i].ctype == typ) {
					this.symbols.splice(i, 1);
					i = this.symbols.length;
				}
			}
		}
		this.symbols.push(new HexElement(typ, mater));
	}
	this.addAnnotation = function(typ, mater) {
		if(this.annotations.length < 2) {
			this.annotations.push(new HexElement(parseInt(typ), mater));
		}
	}
	this.cloneAt = function(x, y) {
		var newhex = new Hex(x, y);
		for(var b = 0; b < this.bases.length; b++) {
			newhex.addBase(this.bases[b].ctype, this.bases[b].material);
		}
		for(var o = 0; o < this.overlays.length; o++) {
			newhex.addOverlay(this.overlays[o].ctype, this.overlays[o].material);
		}
		for(var f = 0; f < this.terrains.length; f++) {
			newhex.addTerrain(this.terrains[f].ctype, this.terrains[f].material);
		}
		for(var f = 0; f < this.features.length; f++) {
			newhex.addFeature(this.features[f].ctype, this.features[f].material);
		}
		for(var f = 0; f < this.decors.length; f++) {
			newhex.addDecor(this.decors[f].ctype, this.decors[f].material);
		}
		for(var f = 0; f < this.symbols.length; f++) {
			newhex.addSymbol(this.symbols[f].ctype, this.symbols[f].material);
		}
		for(var a = 0; a < this.annotations.length; a++) {
			newhex.addAnnotation(this.annotations[a].ctype, this.annotations[a].material);
		}
		return newhex;
	}
	this.clone = function() {
		return this.cloneAt(this.x, this.y);
	}
}

function RoomDef(x, y, w, h) {
	this.x = x;
	this.y = y;
	this.width = w;
	this.height = h;

	this.startX = this.x;
	this.startY = this.y;
	this.endX   = this.x + (this.width - 1);
	this.endY   = this.y + (this.height - 1);

	this.overlaps = function(room, buffer) {
		if(buffer == null) {
			buffer = 0;
		}
		return !(((this.startX > room.endX + buffer) || (this.endX + buffer < room.startX)) || ((this.startY > room.endY + buffer) || (this.endY + buffer < room.startY)));
	}
}

/* Pseudoconstants */

var drawsize   = 16;
var defCols    = 51;
var defRows    = 34;
var gridx      = defCols;
var gridy      = defRows;
var leftmargin = 8;
var topmargin  = 8;

/* Variables */

var mapGrain = GRAIN_VERTICAL; // GRAIN_VERTICAL || GRAIN_HORIZONTAL
var lineWeight = LINEWEIGHT_PROPORTIONAL; // LINEWEIGHT_PROPORTIONAL || LINEWEIGHT_HAIRLINE
var megahexWeight = MEGAHEX_WEIGHT_DEFAULT;
var megahexDotWeight = MEGADOT_WEIGHT_DEFAULT;

var modeStartX = -1;
var modeStartY = -1;

var drawsize = 16;
var drawscale = drawsize;
var halfscale = drawscale / 2;
var xoffset = drawscale;
var yoffset = (drawscale * CALCTHIN);

var outlineThickness = getOutlineThickness();

var defaultGridGrey = 127;
var defaultGridColor = 'rgb(' + defaultGridGrey + ", " + defaultGridGrey + ", " + defaultGridGrey + ')';

var hexOutline  = defaultGridColor;
var toolOutline = '#000000';

var scratchTemp = 0;

var currentMaterPalette = 0;
var maxMaterPalette     = Math.ceil((materialNames.length - 1) / PALETTE_SIZE) - 1;
var currentCompPalette  = 0;
var maxCompPalette      = 1;
var currentFeatPalette  = 0;
var maxFeatPalette      = 1;
var currentSymbPalette  = 0;
var maxSymbPalette      = 1;

// Active States
var activeMode       = MODE_PAINT;
var activeLayer      = LAYER_BASE;
var activeMaterial   = 1;
var activeBase       = HEX_WHOLE;
var activeTerrain    = TERRAIN_WATER;
var activeOverlay    = HEX_WHOLE;
var activeFeature    = FEATURE_DOOR;
var activeSymbol     = SYMBOL_ARROWEL;
var activeDecor      = DECOR_TRAP;
var activeAnnotation = 9065;
var activePalButton  = LAYER_BUTTON_BASE;
var activePalSection = PAGE_BLOCK_BASE;
var activeTheme      = THEME_LIGHT;
var activeHandedness = SIDEBAR_RIGHT;

var lastHexCol = 0;
var lastHexRow = 0;
var paintOn = false;
var megahexesOn = false;

// Megahex Overlay
var mhexType = MEGAHEX_TYPE_STANDARD;
var mhexImposeX = 1;
var mhexImposeY = -6;
var mhexImposeS = 0;
var mhexOutline = '#000000';

// Backdrop Image
var backdropImage = null;
var backdropShowing = true;
var backdropOffset = new Point(0, 0);
var backdropLocked = false;
var translucencyEnabled = false;

// Dupe and Clone holders
var emptyHex = new Hex(); emptyHex.clearAll();
var dupeHex = null;
var dupeHexOutline = "#fc74d3"; //#ffa4a4

// Undo State
var haveUndo = false;

// File Variables
var defaultMapName = 'default';
var loadedMapName = defaultMapName;
var defaultAuthorName = 'default';
var loadedAuthorName = defaultAuthorName;
var fileIteration = 0;

// Preference Restoration Values
var prefRestoreCreator = defaultAuthorName;
var prefRestoreGridshade = defaultGridGrey;

/* References */

// Divs
var sidebarDiv;
var toolbarDiv;
var paletteDiv;
var itemsDiv;
var shaderDiv;
var overlayDiv;

// Canvases
var cframe;
var canvas;
var cnv;
var buffer;
var buf;
var mhexes;
var mhx;
var preview;
var prv;

/* SVG Interface Reference Variables */

const NARROW_SIDEBAR = 317;
const NARROW_ITEMS = 281;
const WIDE_SIDEBAR = 457;
const WIDE_ITEMS = 421;

var sidebarWidth = WIDE_SIDEBAR;
var toolbarWidth = 36;
var itemsWidth = WIDE_ITEMS;
var toolButtonSize = 32;
var toolButtonViewbox = '0 0 92.44 92.44';

var toolIconNormalFill = 'none';
var toolIconNormalLine = '#ffffff';
var toolIconNormalLineWeight = '3';
var toolIconSelectFill = '#ffffff';
var toolIconSelectLine = 'none';
var toolIconSelectLineWeight = '0';

/* Interface Construction */

function writeTokenizedSvgElement(id, def, wid, hgt, vbox, fill, line, linew) {
	document.write('<svg ' + SVG_DECLARATION + ' id="' + id + '" width="' + wid + '" height="' + hgt + '" viewBox="' + vbox + '">');
	if(fill != null && fill != undefined) {
		if(Array.isArray(fill)) {
			for(a = 0; a < fill.length; a++) {
				def = def.replaceAll('FILLCOLOR' + a, fill[a]);
			}
		}
		else {
			def = def.replaceAll('FILLCOLOR', fill);
		}
	}
	if(line != null && line != undefined) {
		if(Array.isArray(line)) {
			for(a = 0; a < line.length; a++) {
				def = def.replaceAll('LINECOLOR' + a, line[a]);
			}
		}
		else {
			def = def.replaceAll('LINECOLOR', line);
		}
	}
	if(linew != null && linew != undefined) {
		def = def.replaceAll('LINEWIDTH', linew);
	}
	document.write(def);
	document.write('</svg>');
}

function writeSvgCssElement(id, def, wid, hgt, vbox) {
	document.write('<svg ' + SVG_DECLARATION + ' id="' + id + '" width="' + wid + '" height="' + hgt + '" viewBox="' + vbox + '">');
	document.write(def);
	document.write('</svg>');
}

function writeSvgToolElement(id, def, vbox) {
	writeSvgCssElement(id, def, toolButtonSize, toolButtonSize, (vbox != undefined ? vbox : toolButtonViewbox));
}

function writePaletteButton(id, svgIg, def, hoverTip, isActive) {
	if(isActive == 'undefined' || isActive == null) { isActive = (activePalButton == id); }
	document.write('<div role="button" id="' + id + '" class="' + CSS_TOOL_BUTTON + (isActive ? ' ' + CSS_TOOL_BUTTON_ACTIVE : '') + '"' + ((hoverTip != undefined && hoverTip != null && hoverTip > '') ? ' title="' + hoverTip + '"' : '') + ' onclick="processPaletteButton(' + "'" + id + "'" + ');"' + '>');
	writeSvgToolElement(svgIg, def);
	document.write('</div>');
}

function writeLayerButton(id, svgIg, def, hoverTip) {
	document.write('<div role="button" id="' + LAYER_BUTTON_PREFIX + id + '" class="' + CSS_TOOL_BUTTON + ((activeLayer == id) ? ' ' + CSS_TOOL_BUTTON_ACTIVE : '') + '"' + ((hoverTip != undefined && hoverTip != null && hoverTip > '') ? ' title="' + hoverTip + '"' : '') + ' onclick="processPaletteButton(' + "'" + LAYER_BUTTON_PREFIX + id + "'" + ');"' + '>');
	writeSvgToolElement(svgIg, def);
	document.write('</div>');
}

function writeInterfaceSvgTool(id, gid, func, args, wid, hgt, vbox, hover, label, pdef) {
	var elemStr =
		'<div ' +
			'role="button" ' +
			'id="' + id + '" ' +
			'class="' + CSS_PAL_BUTTON + '"' +
			(func != undefined && func != null ? ' onclick="' + func + '(' + (args != undefined && args != null ? (typeof args == 'string' ? "'" + args + "'" : args) : '') + ');" ' : '') +
			(hover != undefined && hover != null ? ' title="' + hover + '" ' : '') +
		'>' +
			'<svg ' + SVG_DECLARATION + ' width="' + wid + '" height="' + hgt + '" viewBox="' + vbox + '">' +
				pdef +
			'</svg>' +
			'<div class="buttonlabel">' + label + '</div>' +
		'</div>';
	document.write(elemStr);
}

function writeMenuToggleButton(id, buttonId, func, def, wid, hgt, vbox, hoverTip) {
	document.write('<div role="button" id="' + buttonId + '" class="menubutton"' + ((hoverTip != undefined && hoverTip != null && hoverTip > '') ? ' title="' + hoverTip + '"' : '') + ((func != '') ? ' onclick="' + func + "()" + ';"' : '') + '>');
	writeSvgCssElement(id, def, wid, hgt, vbox);
	document.write('</div>');
}

function writeMenuLabeledButton(id, buttonId, func, args, def, wid, hgt, vbox, label, hoverTip) {
	document.write('<div role="button" id="' + buttonId + '" class="menulabelbutton"' + ((hoverTip != undefined && hoverTip != null && hoverTip > '') ? ' title="' + hoverTip + '"' : '') + ((func != '') ? ' onclick="' + func + '(' + args + ');"' : '') + '>');
	writeSvgCssElement(id, def, wid, hgt, vbox);
	document.write('<span class="menulabel">' + label + '</span>');
	document.write('</div>');
}

function writeMaterialButton(materialId, isActive) {
	document.write('<div role="button" id="' + MATERIAL_PREFIX + materialId + '" class="' + CSS_PAL_BUTTON + '" onclick="setMaterial(' + materialId + ', false);">');
	writeTokenizedSvgElement('svg' + MATERIAL_PREFIX + materialId, SVGDEF_PALETTE_HEX_COLOR, MATERIAL_BUTTON_SIZE, MATERIAL_BUTTON_SIZE, MATERIAL_BUTTON_SVG_VIEWBOX, materialDef[materialId], '#000000', '2px');
	document.write('<div id="lbl' + MATERIAL_PREFIX + materialId + '" class="buttonlabel">' + materialNames[materialId] + '</div>');
	document.write('</div>');
}

function writeCustomMaterialButton(materialId, isActive) {
	document.write('<div role="button" id="' + MATERIAL_PREFIX + materialId + '" class="' + CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_CUSTOM + '" title="Double-click to edit"' + ' onclick="setMaterial(' + "'" + materialId + "'" + ', true);" ondblclick="editMaterial(' + "'" + materialId + "'" + ');">');
	writeTokenizedSvgElement('svg' + MATERIAL_PREFIX + materialId, SVGDEF_PALETTE_HEX_COLOR_CUSTOM, MATERIAL_BUTTON_SIZE, MATERIAL_BUTTON_SIZE, MATERIAL_BUTTON_SVG_VIEWBOX, null, '#000000', '2px');
	document.write('<div id="lbl' + MATERIAL_PREFIX + materialId + '" class="buttonlabel">' + materialNames[materialId] + '</div>');
	document.write('</div>');
}

function writeCanvasButton(buttonId, canvasId, label, func, args) {
	document.write(
		'<div class="' + CSS_PAL_BUTTON + ' notclickable" id="' + buttonId + '"' + ((func != '') ? ' onclick="' + func + '(' + args + ');"' : '') + '>' +
			'<canvas id="' + canvasId + '" height="40px" width="40px"></canvas>' +
			'<div class="buttonlabel">' + label + '</div>' +
		'</div>');
}

function writeRotateSvgButton(buttonId, shapeId, pdef, isRotateRight) {
	writeInterfaceSvgTool(
		(isRotateRight ? 'rotate_r' : 'rotate_l'),
		(isRotateRight ? 'rotate_clock' : 'rotate_counter'),
		'rotateElement',
		isRotateRight,
		'40px',
		'40px',
		'0 0 29.58 29.58',
		(isRotateRight ? 'Hotkeys: Cursor Right, D' : 'Hotkeys: Cursor Left, A'),
		(isRotateRight ? 'Rotate Right' : 'Rotate Left'),
		pdef
	);
}

function writeFlipSvgButton(buttonId, shapeId, pdef) {
	writeInterfaceSvgTool(buttonId, shapeId, 'flipElement', '', '40px', '40px', '0 0 29.58 29.58', 'Hotkeys: Cursor Down, S', 'Flip', pdef);
}

function writeModeSvgButton(modeId, shapeId, pdef, label, hoverTip) {
	writeInterfaceSvgTool('mode' + modeId, shapeId, 'setPaintMode', modeId, '40px', '40px', '0 0 29.58 29.58', hoverTip, label, pdef);
}

function writeToolbarButton(buttonId, func, args, img, label, wid, hgt, isActive, hoverTip) {
	document.write('<div role="button" id="' + buttonId + '" class="' + CSS_PAL_BUTTON + (isActive ? ' ' + CSS_PAL_BUTTON_ACTIVE : '') + '"' + ((hoverTip != undefined && hoverTip != null && hoverTip > '') ? ' title="' + hoverTip + '"' : '') + ' onclick="' + func + '(' + (typeof args == 'string' ? "'" + args + "'" : args) + ');">');
	document.write('<img src="img/' + img + '.png" width="' + wid + '" height="' + hgt + '"/>');
	document.write('<div class="buttonlabel">' + label + '</div>');
	document.write('</div>');
}

function writeToolButton(toolId, func, args, img, label, hoverTip) {
	writeToolbarButton('tool' + toolId, func, args, img, label, 40, 40, false, hoverTip);
}

function writeModeButton(modeId, img, label, hoverTip) {
	writeToolbarButton('mode' + modeId, 'setPaintMode', modeId, img, label, 40, 40, (activeMode == modeId), hoverTip);
}

function writeBaseButton(baseId, label, isActive) {
	writeToolbarButton('base' + baseId, 'setBase', baseId, 'comp_' + baseId, label, 40, 40, isActive);
}

function writeTerrainButton(terrainId, label, isActive) {
	writeToolbarButton('terrain' + terrainId, 'setTerrain', terrainId, 'terrain_' + terrainId, label, 40, 40, isActive);
}

function writeOverlayButton(overlayId, label, isActive) {
	writeToolbarButton('overlay' + overlayId, 'setOverlay', overlayId, 'comp_' + overlayId, label, 40, 40, isActive);
}

function writeFeatureButton(featureId, label, isActive) {
	writeToolbarButton('feature' + featureId, 'setFeature', featureId, 'feature_' + featureId, label, 40, 40, isActive);
}

function writeDecorButton(decorId, label, isActive) {
	writeToolbarButton('decor' + decorId, 'setDecor', decorId, 'decor_' + decorId, label, 40, 40, isActive);
}

function writeSymbolButton(symbolId, label, isActive) {
	writeToolbarButton('symbol' + symbolId, 'setSymbol', symbolId, 'symbol_' + symbolId, label, 40, 40, isActive);
}

function writeAnnotationButton(annotationId, isActive) {
	document.write('<div role="button" class="annotatetool' + (isActive ? ' annotatetool-active' : '') + '" id="annotate_' + annotationId + '" onclick="setAnnotation(' + "'" + annotationId + "'" + ');"><span class="annosymb">' + (annotationId == ANNOTATION_BASE ? ' ' : String.fromCharCode(annotationId - ANNOTATION_BASE)) + '</span></div>');
}

function writeMegahexButton(mhexId, func, args, label, isActive) {
	writeToolbarButton('mhex' + mhexId, func, args, 'mhex_' + mhexId, label, 40, 40, isActive);
}

function writeComposeButton(id, func, args, img, label, isActive) {
	writeToolbarButton(id, func, args, img, label, 40, 40, isActive);
}

function writeLayerToggleButton(id, def, label, lyr) {
	document.write('<div role="button" id="' + LAYER_TOGGLE_PREFIX + lyr + '" class="' + CSS_LYR_BUTTON + (layerVisibility[lyr] ? ' ' + CSS_LYR_BUTTON_ACTIVE : '') + '" onclick="toggleLayer(' + lyr + ');">');
	writeSvgCssElement(id, def, 20, 20, toolButtonViewbox);
	document.write('<br/>' + label + '</div>');
}

function writePaletteBlockHead(id, header, subid) {
	document.write('<div id="' + id +'" class="paletteblock">');
	document.write('<div class="paletteheader">' + header + '</div>');
	if(subid != null) {
		document.write('<div class="palettebuttons" id="' + subid + '">');
	}
}

function writePaletteBlockTail(id, subid) {
	if(subid != null) {
		document.write('</div>');
	}
	document.write('</div>');
}

function writeToolbarSeparator() {
	document.write('<div class="toolsep"></div>');
}

function activatePalette(paletteButton) {
	document.getElementById(BUTTON_BLOCKS[paletteButton]).style.display = CSS_BLOCK_SHOW;
	document.getElementById(paletteButton).className = CSS_TOOL_BUTTON + ' ' + CSS_TOOL_BUTTON_ACTIVE;
	checkDivScrolling();
}

function deactivatePalette(paletteButton) {
	document.getElementById(BUTTON_BLOCKS[paletteButton]).style.display = CSS_BLOCK_HIDE;
	document.getElementById(paletteButton).className = CSS_TOOL_BUTTON;
	checkDivScrolling();
}

function togglePalette(paletteButton) {
	if(document.getElementById(BUTTON_BLOCKS[paletteButton]).style.display == CSS_BLOCK_SHOW) {
		deactivatePalette(paletteButton);
	}
	else {
		activatePalette(paletteButton);
	}
	checkDivScrolling();
}

function setActivePalette(paletteButton) {
	if(paletteButton in BUTTON_BLOCKS) {
		if(MATER_BUTTONS.includes(paletteButton)) {
			togglePalette(paletteButton);
			for(p = 0; p < MATER_BUTTONS.length; p++) {
				if(MATER_BUTTONS[p] != paletteButton) {
					deactivatePalette(MATER_BUTTONS[p], MATER_BLOCKS[p]);
				}
			}
		}
		else if(BUTTON_BLOCKS[paletteButton] != activePalSection) {
			document.getElementById(activePalSection).style.display = CSS_BLOCK_HIDE;
			document.getElementById(BUTTON_BLOCKS[paletteButton]).style.display = CSS_BLOCK_SHOW;
			activePalSection = BUTTON_BLOCKS[paletteButton];
			if(paletteButton == LAYER_BUTTON_BASE) {
				activeLayer = LAYER_BASE;
			}
			else if(paletteButton == LAYER_BUTTON_TERRAIN) {
				activeLayer = LAYER_TERRAIN;
			}
			else if(paletteButton == LAYER_BUTTON_OVERLAY) {
				activeLayer = LAYER_OVERLAY;
			}
			else if(paletteButton == LAYER_BUTTON_FEATURE) {
				activeLayer = LAYER_FEATURE;
			}
			else if(paletteButton == LAYER_BUTTON_DECOR) {
				activeLayer = LAYER_DECOR;
			}
			else if(paletteButton == LAYER_BUTTON_SYMBOL) {
				activeLayer = LAYER_SYMBOL;
			}
			else if(paletteButton == LAYER_BUTTON_ANNOTATE) {
				activeLayer = LAYER_ANNOTATE;
			}
			else if(paletteButton == LAYER_BUTTON_MEGAHEX) {
				activeLayer = LAYER_MEGAHEX;
			}
			layerVisibility[activeLayer] = true;
			toggleRoomDrawMode(paletteButton != LAYER_BUTTON_BASE && paletteButton != LAYER_BUTTON_OVERLAY);
		}
		checkDivScrolling();
	}
}

function setActiveSelection(pbutton) {
	if(pbutton != activePalButton) {
		document.getElementById(activePalButton).className = CSS_TOOL_BUTTON;
		setActivePalette(pbutton);
		activePalButton = pbutton;
		document.getElementById(activePalButton).className = CSS_TOOL_BUTTON + ' ' + CSS_TOOL_BUTTON_ACTIVE;
		checkDivScrolling();
	}
}

function processPaletteButton(pbutton) {
	if(pbutton != null) {
		if(TOGGLE_BUTTONS.includes(pbutton)) {
			togglePalette(pbutton);
		}
		else if(MATER_BUTTONS.includes(pbutton)) {
			setActivePalette(pbutton);
		}
		else {
			setActiveSelection(pbutton);
		}
		updatePreview();				  
	}
}

/* Input Interceptor & Hotkeys */

const HOTKEY_LEFT     = "hkleft";
const HOTKEY_RIGHT    = "hkright";
const HOTKEY_DOWN     = "hkdown";
const HOTKEY_UP       = "hkup";
const HOTKEY_ZOOM_IN  = "hkzoomin";
const HOTKEY_ZOOM_OUT = "hkzoomout";

const HOTKEY_MAP = {
	'a': HOTKEY_LEFT,
	'd': HOTKEY_RIGHT,
	's': HOTKEY_DOWN,
	'w': HOTKEY_UP,
	'ArrowLeft': HOTKEY_LEFT,
	'ArrowRight': HOTKEY_RIGHT,
	'ArrowDown': HOTKEY_DOWN,
	'ArrowUp': HOTKEY_UP,
	'+': HOTKEY_ZOOM_IN,
	'=': HOTKEY_ZOOM_IN,
	'-': HOTKEY_ZOOM_OUT
};

const KEY_IGNORE = ['Shift', 'Alt', 'Control', 'Meta', 'NumLock', 'ScrollLock'];

function hotkeyPress(event) {
	if(document.activeElement.type == 'text') { return; }
	if(activeLayer == LAYER_ANNOTATE) {
		if(KEY_IGNORE.includes(event.key)) { return; }
		debugMsg(event.key.charCodeAt());
		annonum = 9000 + event.key.charCodeAt();
		if(event.keyCode == 32) {
			annonum = 9000;
		}
		if(document.getElementById('annotate_' + annonum) != null) {
		}
		else {
			annonum = 9000 + event.keyCode;
		}
		if(document.getElementById('annotate_' + annonum) != null) {
			setAnnotation(annonum);
			doubleBuffer();
			plotTool(cnv, lastHexCol, lastHexRow);
		}
		return;
	}
	if(HOTKEY_MAP.hasOwnProperty(event.key)) {
		event.preventDefault();
		refreshTool = false;
		switch(HOTKEY_MAP[event.key]) {
			case HOTKEY_LEFT:
				if(MUTABLE_ELEMENTS.includes(activeLayer)) {
					rotateElement(false);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_W);
				}
				break;
			case HOTKEY_RIGHT:
				if(MUTABLE_ELEMENTS.includes(activeLayer)) {
					rotateElement(true);
					refreshTool = true;
				}
				else if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_E);
				}
				break;
			case HOTKEY_DOWN:
				if(MUTABLE_ELEMENTS.includes(activeLayer)) {
					flipElement();
					refreshTool = true;
				}
				else if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_S);
				}
				break;
			case HOTKEY_UP:
				if(activeLayer == LAYER_MEGAHEX) {
					shiftMegahexes(CARDINAL_N);
				}
				break;
			case HOTKEY_ZOOM_IN:
				zoom(true);
				break;
			case HOTKEY_ZOOM_OUT:
				zoom(false);
				break;
			default:
				debugMsg(HOTKEY_MAP[event.key]);
				break;
		}
		if(refreshTool) {
			doubleBuffer();
			plotTool(cnv, lastHexCol, lastHexRow);
		}
		if(typeof event.stopPropagation != 'undefined') {
			event.stopPropagation();
		}
		else {
			event.cancelBubble = true;
		}
	}
}

/* Utility Functions */

function checkSimpleString(str) {
	if(str != 'undefined' && str != null && str.length > 0) {
		for(var s = 0; s < str.length; s++) {
			if(str.charCodeAt(s) < 32 || str.charCodeAt(s) > 126) {
				return false;
			}
		}
		return true;
	}
	return false;
}

function debugMsg(msg) {
	if(DEBUG_MODE == true) {
		console.log(msg);
	}
}

/* Interface */

function initialiseInterface() {
	sidebarDiv = document.getElementById('sidebar');
	sidebarDiv.setAttribute('style','width:' + sidebarWidth + 'px');
	sidebarDiv.style.width = sidebarWidth+'px';
	toolbarDiv = document.getElementById('toolbar');
	toolbarDiv.setAttribute('style','width:' + toolbarWidth + 'px');
	toolbarDiv.style.width = toolbarWidth+'px';
	itemsDiv = document.getElementById('item-panel');
	itemsDiv.setAttribute('style','width:' + itemsDiv + 'px');
	itemsDiv.style.width = itemsWidth+'px';
	shaderDiv = document.getElementById('shader');
	overlayDiv = document.getElementById('overlay');
	initialise();
	loadPrefs();
	activatePalette(TOOL_BUTTON_MAPSETS);
}

function checkDivScrolling() {
	var toolScrollbarWidth = toolbarDiv.offsetWidth - toolbarDiv.clientWidth;
	var itemScrollbarWidth = itemsDiv.offsetWidth - itemsDiv.clientWidth;
	if(toolbarDiv.offsetWidth > toolbarDiv.clientWidth) {
		toolScrollbarWidth = toolbarDiv.offsetWidth - toolbarDiv.clientWidth;
		toolbarDiv.setAttribute('style','width:' + toolbarWidth + toolScrollbarWidth + 'px');
		toolbarDiv.style.width = toolbarWidth + toolScrollbarWidth + 'px';
	}
	else {
		if(toolbarDiv.clientWidth > toolbarWidth) {
			toolbarDiv.setAttribute('style','width:' + toolbarWidth + 'px');
			toolbarDiv.style.width = toolbarWidth + 'px';
		}
	}
	if(itemsDiv.offsetWidth > itemsDiv.clientWidth) {
		itemScrollbarWidth = itemsDiv.offsetWidth - itemsDiv.clientWidth;
		itemsDiv.setAttribute('style','width:' + itemsWidth + itemScrollbarWidth + 'px');
		itemsDiv.style.width = itemsWidth + itemScrollbarWidth + 'px';
	}
	else {
		if(itemsDiv.offsetWidth > itemsWidth) {
			itemsDiv.setAttribute('style','width:' + itemsWidth + 'px');
			itemsDiv.style.width = itemsWidth + 'px';
		}
	}
	sidebarDiv.setAttribute('style','width:' + sidebarWidth + toolScrollbarWidth + itemScrollbarWidth + 'px');
	sidebarDiv.style.width = sidebarWidth + toolScrollbarWidth + itemScrollbarWidth + 'px';
}

function refreshDisplaySpace() {
	var pagex = window.innerWidth || document.documentElement.clientWidth || document.getElementsByTagName('body')[0].clientWidth;
	var pagey = window.innerHeight || document.documentElement.clientHeight || document.getElementsByTagName('body')[0].clientHeight;
	var usableHeight = pagey - 36;
	var displayHeight = Math.max(usableHeight, 820);
	document.getElementById('framework').setAttribute('style','height:' + displayHeight + 'px');
	document.getElementById('framework').style.height = displayHeight + "px";
	checkDivScrolling();
}

/* Preferences */

function setTheme(themename) {
	const htmlTag = document.getElementsByTagName("html")[0];
	htmlTag.setAttribute("data-theme", themename);
	window.localStorage.setItem("data-theme", themename);
	document.getElementById("svg_theme").innerHTML = (themename == THEME_DARK ? SVGDEF_THEME_DARK : SVGDEF_THEME_LIGHT);
	activeTheme = themename;
}

function toggleTheme() {
	if(activeTheme == THEME_LIGHT) {
		setTheme(THEME_DARK);
	}
	else {
		setTheme(THEME_LIGHT);
	}
	updatePreview();
}

function setHandedness(handopt) {
	const htmlTag = document.getElementsByTagName("html")[0];
	htmlTag.setAttribute("handedness", handopt);
	window.localStorage.setItem("handedness", handopt);
	document.getElementById("svg_handed").innerHTML = (handopt == SIDEBAR_LEFT ? SVGDEF_HANDED_LEFT : SVGDEF_HANDED_RIGHT);
	activeHandedness = handopt;
}

function toggleHandedness() {
	if(activeHandedness == SIDEBAR_RIGHT) {
		setHandedness(SIDEBAR_LEFT);
	}
	else {
		setHandedness(SIDEBAR_RIGHT);
	}
}

function loadPrefs() {
	const storedThemePref = window.localStorage.getItem("data-theme");
	if(storedThemePref != null) {
		setTheme(storedThemePref);
	}
	const storedHandedPref = window.localStorage.getItem("handedness");
	if(storedHandedPref != null) {
		setHandedness(storedHandedPref);
	}
}

/* Methods */

function initialiseGrid() {
	hexarray = [];
	for(x = 0; x < gridx; x++) {
		hexarray[x] = [];
		for(y = 0; y < gridy; y++) {
			var hex = new Hex(x, y);
			hex.addBase(HEX_EMPTY, MATERIAL_EMPTY);
			hexarray[x][y] = hex;
		}
	}
	hexundo = hexarray.slice();
}

function initialiseFont(ctx, scl) {
	var fontStr = "bold 1px RobotoMono";
	ctx.font = fontStr;
	ctx.fillStyle = 0;
	ctx.textAlign = "center";
	ctx.textBaseline = "middle";
	ctx.fillText('A', -10, -10);
}

function initialise() {
	// initialise graphical components
	cframe = document.getElementById('canvas-frame');
	canvas = document.getElementById('canvas');
	cnv = canvas.getContext('2d');
	buffer = document.getElementById('buffer');
	buf = buffer.getContext('2d');
	mhexes = document.getElementById('mhexes');
	mhx = mhexes.getContext('2d');
	preview = document.getElementById(CANVAS_PREVIEW);
	prv = preview.getContext('2d');
	updatePreview();
	// install mouse listeners
	canvas.addEventListener('mousemove',  function(evt) { processMouseMove(evt); });
	canvas.addEventListener('mousedown',  function(evt) { togglePaintMode(evt, true); });
	canvas.addEventListener('mouseup',    function(evt) { togglePaintMode(evt, false); });
	canvas.addEventListener('mouseleave', function(evt) { togglePaintMode(evt, false); hideCursor(evt); });
	document.getElementById('mapname').value = defaultMapName;
	document.getElementById('mapauthor').value = defaultAuthorName;
	document.getElementById('maprows').value = gridy;
	document.getElementById('mapcols').value = gridx;
	// force load the annotation font
	initialiseFont(cnv, drawsize);
	// initialise the grid
	initialiseGrid();
	// prep the canvas
	resizeCanvas();
	drawMap();
	refreshDisplaySpace();
}

function doubleBuffer() {
	cnv.clearRect(0, 0, canvas.width, canvas.height);
	if(backdropShowing && backdropImage != null) {
		if(backdropImage.src != null && backdropImage.src > '') {
			cnv.drawImage(backdropImage, backdropOffset.x, backdropOffset.y, backdropImage.width, backdropImage.height);
		}
	}
	if(translucencyEnabled) { cnv.globalAlpha = TRANSLUCE; }
	cnv.drawImage(buffer, 0, 0);
	if(megahexesOn && layerVisibility[LAYER_MEGAHEX]) {
		cnv.drawImage(mhexes, 0, 0);
	}
	if(translucencyEnabled) { cnv.globalAlpha = OPAQUE; }
}

function getOutlineThickness() {
	if(lineWeight == LINEWEIGHT_PROPORTIONAL) {
		return Math.max(drawscale / 16, 1);
	}
	else if(lineWeight == LINEWEIGHT_HAIRLINE) {
		return 1;
	}
	else {
		return DEFAULT_LINE_THICKNESS;
	}
}

function getHexCol(x, r) {
	if(mapGrain == GRAIN_VERTICAL) {
		return Math.round((x - (leftmargin + xoffset)) / (drawscale * 1.5));
	}
	else {
		return Math.round((x - (leftmargin + xoffset) - ((r % 2) * (drawscale * CALCTHIN))) / (drawscale * CALCSQRT));
	}
}
function getHexRow(y, c) {
	if(mapGrain == GRAIN_VERTICAL) {
		return Math.round((y - (topmargin + yoffset) - ((c % 2) * (drawscale * CALCTHIN))) / (drawscale * CALCSQRT));
	}
	else {
		return Math.round((y - (topmargin + yoffset)) / (drawscale * 1.5));
	}
}

function getHexRawX(row, col, scl, xoff, lmrg) {
	if(mapGrain == GRAIN_VERTICAL) {
		return (col * (scl * 1.5)) + xoff + lmrg;
	}
	else {
		shft = (row % 2) * (scl * CALCTHIN);
		return (col * scl * CALCSQRT) + xoff + shft + lmrg;
	}
}
function getHexRawY(row, col, scl, yoff, tmrg) {
	if(mapGrain == GRAIN_VERTICAL) {
		shft = (col % 2) * (scl * CALCTHIN);
		return (row * scl * CALCSQRT) + yoff + shft + tmrg;
	}
	else {
		return (row * (scl * 1.5)) + yoff + tmrg;
	}
}

function getHexX(row, col) { return getHexRawX(row, col, drawscale, xoffset, leftmargin); }
function getHexY(row, col) { return getHexRawY(row, col, drawscale, yoffset, topmargin); }

function getHexAt(x, y, direction) {
	var pt = new Point(x, y);
	if(mapGrain == GRAIN_VERTICAL) {
		if(x % 2 == 0) {
			switch(direction) {
				case DIR_HXC0:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC3:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				default:
					break;
			}
		}
		else {
			switch(direction) {
				case DIR_HXC0:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC3:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				default:
					break;
			}
		}
	}
	else {
		if(y % 2 == 0) {
			switch(direction) {
				case DIR_HXC0:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC3:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				default:
					break;
			}
		}
		else {
			switch(direction) {
				case DIR_HXC0:
					pt.x = x;
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC1:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y - 1) < 0 ? OFFGRID : (y - 1));
					break;
				case DIR_HXC2:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = y;
					break;
				case DIR_HXC3:
					pt.x = ((x + 1) >= hexarray.length ? OFFGRID : x + 1);
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC4:
					pt.x = x;
					pt.y = ((y + 1) >= hexarray[x].length ? OFFGRID : y + 1);
					break;
				case DIR_HXC5:
					pt.x = ((x - 1) < 0 ? OFFGRID : (x - 1));
					pt.y = y;
					break;
				default:
					break;
			}
		}
	}
	return pt;
}

function getMouseHex(x, y) {
	var boundbox = cframe.getBoundingClientRect();
	var mousex = (x - boundbox.left) + cframe.scrollLeft;
	var mousey = (y - boundbox.top) + cframe.scrollTop;
	var col, row;
	if(mapGrain == GRAIN_VERTICAL) {
		col = getHexCol(mousex, null);
		row = getHexRow(mousey, col);
	}
	else {
		row = getHexRow(mousey, null);
		col = getHexCol(mousex, row);
	}
	var mouseHex = { x: col, y: row };
	return mouseHex;
}

function getHexPoints(dscale) {
	var hscale = Math.max(dscale / 2, 1);
	var qscale = Math.max(dscale / 4, 1);
	var escale = Math.max(dscale / 8, 1);
	// returns a set of points with the apex points, the midpoints for each face, and the sub-mid points for each apex-midpoint pair
	// return the origin (0,0) as the 0 index item, the remaining points are the hexagon starting at the upper left and moving clockwise
	if(mapGrain == GRAIN_VERTICAL) {
		pptx = [
			0,
			-hscale,
			-qscale,
			0,
			qscale,
			hscale,
			hscale + escale,
			hscale + qscale,
			hscale + qscale + escale,
			dscale,
			hscale + qscale + escale,
			hscale + qscale,
			hscale + escale,
			hscale,
			qscale,
			0,
			-qscale,
			-hscale,
			-(hscale + escale),
			-(hscale + qscale),
			-(hscale + qscale + escale),
			-dscale,
			-(hscale + qscale + escale),
			-(hscale + qscale),
			-(hscale + escale)
		];
		ppty = [
			0,
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN),
			-(hscale * CALCTHIN),
			-(qscale * CALCTHIN),
			0,
			(qscale * CALCTHIN),
			(hscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(hscale * CALCTHIN),
			(qscale * CALCTHIN),
			0,
			-(qscale * CALCTHIN),
			-(hscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN)
		];
	}
	else {
		pptx = [
			0,
			-(dscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN),
			-(hscale * CALCTHIN),
			-(qscale * CALCTHIN),
			0,
			(qscale * CALCTHIN),
			(hscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			(dscale * CALCTHIN),
			((hscale + qscale) * CALCTHIN),
			(hscale * CALCTHIN),
			(qscale * CALCTHIN),
			0,
			-(qscale * CALCTHIN),
			-(hscale * CALCTHIN),
			-((hscale + qscale) * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN),
			-(dscale * CALCTHIN)
		];
		ppty = [
			0,
			-hscale,
			-(hscale + escale),
			-(hscale + qscale),
			-(hscale + qscale + escale),
			-dscale,
			-(hscale + qscale + escale),
			-(hscale + qscale),
			-(hscale + escale),
			-hscale,
			-qscale,
			0,
			qscale,
			hscale,
			(hscale + escale),
			(hscale + qscale),
			(hscale + qscale + escale),
			dscale,
			(hscale + qscale + escale),
			(hscale + qscale),
			(hscale + escale),
			hscale,
			qscale,
			0,
			-qscale
		];
	}
	return [pptx, ppty];
}

function getHexFinePoints(ctrx, ctry, dscale) {
	var hscale = Math.max(dscale / 2, 1);
	var qscale = Math.max(dscale / 4, 1);
	var escale = Math.max(dscale / 8, 1);
	// return the centerpoint as the 0 index item, the remaining points are the hexagon starting at the upper left and moving clockwise
	// returns a large set of points, with the apex points, the midpoints for each face, and the sub-mid points for each apex-midpoint pair
	// there points are translated in reference to the hex centerpoint
	if(mapGrain == GRAIN_VERTICAL) {
		pptx = [
			ctrx,
			ctrx - hscale,
			ctrx - qscale,
			ctrx,
			ctrx + qscale,
			ctrx + hscale,
			ctrx + hscale + escale,
			ctrx + hscale + qscale,
			ctrx + hscale + qscale + escale,
			ctrx + dscale,
			ctrx + hscale + qscale + escale,
			ctrx + hscale + qscale,
			ctrx + hscale + escale,
			ctrx + hscale,
			ctrx + qscale,
			ctrx,
			ctrx - qscale,
			ctrx - hscale,
			ctrx - (hscale + escale),
			ctrx - (hscale + qscale),
			ctrx - (hscale + qscale + escale),
			ctrx - dscale,
			ctrx - (hscale + qscale + escale),
			ctrx - (hscale + qscale),
			ctrx - (hscale + escale)
		];
		ppty = [
			ctry,
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - (dscale * CALCTHIN),
			ctry - ((hscale + qscale) * CALCTHIN),
			ctry - (hscale * CALCTHIN),
			ctry - (qscale * CALCTHIN),
			ctry,
			ctry + (qscale * CALCTHIN),
			ctry + (hscale * CALCTHIN),
			ctry + ((hscale + qscale) * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + (dscale * CALCTHIN),
			ctry + ((hscale + qscale) * CALCTHIN),
			ctry + (hscale * CALCTHIN),
			ctry + (qscale * CALCTHIN),
			ctry,
			ctry - (qscale * CALCTHIN),
			ctry - (hscale * CALCTHIN),
			ctry - ((hscale + qscale) * CALCTHIN)
		];
	}
	else {
		pptx = [
			ctrx,
			ctrx - (dscale * CALCTHIN),
			ctrx - ((hscale + qscale) * CALCTHIN),
			ctrx - (hscale * CALCTHIN),
			ctrx - (qscale * CALCTHIN),
			ctrx,
			ctrx + (qscale * CALCTHIN),
			ctrx + (hscale * CALCTHIN),
			ctrx + ((hscale + qscale) * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + (dscale * CALCTHIN),
			ctrx + ((hscale + qscale) * CALCTHIN),
			ctrx + (hscale * CALCTHIN),
			ctrx + (qscale * CALCTHIN),
			ctrx,
			ctrx - (qscale * CALCTHIN),
			ctrx - (hscale * CALCTHIN),
			ctrx - ((hscale + qscale) * CALCTHIN),
			ctrx - (dscale * CALCTHIN),
			ctrx - (dscale * CALCTHIN),
			ctrx - (dscale * CALCTHIN),
			ctrx - (dscale * CALCTHIN)
		];
		ppty = [
			ctry,
			ctry - hscale,
			ctry - (hscale + escale),
			ctry - (hscale + qscale),
			ctry - (hscale + qscale + escale),
			ctry - dscale,
			ctry - (hscale + qscale + escale),
			ctry - (hscale + qscale),
			ctry - (hscale + escale),
			ctry - hscale,
			ctry - qscale,
			ctry,
			ctry + qscale,
			ctry + hscale,
			ctry + (hscale + escale),
			ctry + (hscale + qscale),
			ctry + (hscale + qscale + escale),
			ctry + dscale,
			ctry + (hscale + qscale + escale),
			ctry + (hscale + qscale),
			ctry + (hscale + escale),
			ctry + hscale,
			ctry + qscale,
			ctry,
			ctry - qscale
		];
	}
	return [pptx, ppty];
}

function getHexCoords(scaling) {
	return getHexPoints(scaling);
}

function getHexFineCoords(col, row) {
	plotx = getHexX(row, col);
	ploty = getHexY(row, col);
	return getHexFinePoints(plotx, ploty, drawscale);
}

function getHexRawCoords(scl, xoff, yoff) {
	plotx = getHexRawX(0, 0, scl, xoff, 0);
	ploty = getHexRawY(0, 0, scl, yoff, 0);
	return getHexFinePoints(plotx, ploty, scl);
}

function getNorthernHexPoints(ctrx, ctry, dscale) {
	var hscale = Math.max(dscale / 2, 1);
	var qscale = Math.max(dscale / 4, 1);
	var escale = Math.max(dscale / 8, 1);
	pptx = [
		ctrx,
		ctrx - hscale,
		ctrx - qscale,
		ctrx,
		ctrx + qscale,
		ctrx + hscale,
		ctrx + hscale + escale,
		ctrx + hscale + qscale,
		ctrx + hscale + qscale + escale,
		ctrx + dscale,
		ctrx + hscale + qscale + escale,
		ctrx + hscale + qscale,
		ctrx + hscale + escale,
		ctrx + hscale,
		ctrx + qscale,
		ctrx,
		ctrx - qscale,
		ctrx - hscale,
		ctrx - (hscale + escale),
		ctrx - (hscale + qscale),
		ctrx - (hscale + qscale + escale),
		ctrx - dscale,
		ctrx - (hscale + qscale + escale),
		ctrx - (hscale + qscale),
		ctrx - (hscale + escale)
	];
	ppty = [
		ctry,
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - (dscale * CALCTHIN),
		ctry - ((hscale + qscale) * CALCTHIN),
		ctry - (hscale * CALCTHIN),
		ctry - (qscale * CALCTHIN),
		ctry,
		ctry + (qscale * CALCTHIN),
		ctry + (hscale * CALCTHIN),
		ctry + ((hscale + qscale) * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + (dscale * CALCTHIN),
		ctry + ((hscale + qscale) * CALCTHIN),
		ctry + (hscale * CALCTHIN),
		ctry + (qscale * CALCTHIN),
		ctry,
		ctry - (qscale * CALCTHIN),
		ctry - (hscale * CALCTHIN),
		ctry - ((hscale + qscale) * CALCTHIN)
	];
	return [pptx, ppty];
}

function rotatePoint(pt, angle) {
	var sine = Math.sin(angle * ANGLE_STEP);
	var cosine = Math.cos(angle * ANGLE_STEP);
	var rotx = (pt.x * cosine) - (pt.y * sine);
	var roty = (pt.x * sine) + (pt.y * cosine);
	return new Point(rotx, roty);
}

function rotatePolygon(pts, angle) {
	if(angle == 0) { return pts; }
	var rotpoly = [];
	for(var p = 0; p < pts.length; p++) {
		rotpoly.push(rotatePoint(pts[p], angle));
	}
	return rotpoly;
}

function renderPolygon(ctx, poly, ctrx, ctry, style, isFilled, lineThick) {
	ctx.beginPath();
	ctx.moveTo(poly[0].x + ctrx, poly[0].y + ctry);
	for(var p = 1; p < poly.length; p++) {
		ctx.lineTo(poly[p].x + ctrx, poly[p].y + ctry);
	}
	ctx.closePath();
	if(isFilled) {
		ctx.fillStyle = style;
		ctx.fill();
	}
	else {
		ctx.lineWidth = (lineThick === undefined ? outlineThickness : lineThick);
		ctx.strokeStyle = style;
		ctx.stroke();
		ctx.lineWidth = DEFAULT_LINE_THICKNESS;
	}
}

function drawPolygon(ctx, poly, ctrx, ctry, style, lineThick) {
	renderPolygon(ctx, poly, ctrx, ctry, style, false, lineThick);
}

function fillPolygon(ctx, poly, ctrx, ctry, style) {
	renderPolygon(ctx, poly, ctrx, ctry, style, true);
}

function drawComponent(ctx, orix, oriy, pptx, ppty, ctyp, matr) {
	ctx.fillStyle = matr;
	var rotation = ctyp % 10;
	var basetyp = ctyp - rotation;
	switch(basetyp) {
		case HEX_EMPTY:
		case HEX_WHOLE:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[9], ppty[9]),
				new Point(pptx[13], ppty[13]),
				new Point(pptx[17], ppty[17]),
				new Point(pptx[21], ppty[21])
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_FHALF:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[9], ppty[9]),
				new Point(pptx[21], ppty[21])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_PHALF:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[3], ppty[3]),
				new Point(pptx[15], ppty[15]),
				new Point(pptx[17], ppty[17]),
				new Point(pptx[21], ppty[21])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_APEX:
			var poly = [
				new Point(pptx[21], ppty[21]),
				new Point(pptx[1], ppty[1]),
				new Point(pptx[17], ppty[17])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_BULK:
			var poly = [
				new Point(pptx[21], ppty[21]),
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[13], ppty[13]),
				new Point(pptx[17], ppty[17])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_TWALL:
			var poly = [
				new Point(pptx[24], ppty[24]),
				new Point(pptx[6], ppty[6]),
				new Point(pptx[8], ppty[8]),
				new Point(pptx[22], ppty[22])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_TEDGE:
			var poly = [
				new Point(pptx[6], ppty[6]),
				new Point(pptx[8], ppty[8]),
				new Point(pptx[10], ppty[10]),
				new Point(pptx[12], ppty[12])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_TCORN:
			var poly = [];
			if(mapGrain == GRAIN_VERTICAL) {
				if(rotation == 1) {
					poly = [
						new Point(pptx[10], ppty[10]),
						new Point(pptx[12], ppty[12]),
						new Point(pptx[12], ppty[10])
					];
				}
				else if(rotation == 2) {
					poly = [
						new Point(pptx[14], ppty[14]),
						new Point(pptx[16], ppty[16]),
						new Point(pptx[15], ppty[12])
					];
				}
				else if(rotation == 3) {
					poly = [
						new Point(pptx[18], ppty[18]),
						new Point(pptx[20], ppty[20]),
						new Point(pptx[18], ppty[20])
					];
				}
				else if(rotation == 4) {
					poly = [
						new Point(pptx[22], ppty[22]),
						new Point(pptx[24], ppty[24]),
						new Point(pptx[24], ppty[22])
					];
				}
				else if(rotation == 5) {
					poly = [
						new Point(pptx[2], ppty[2]),
						new Point(pptx[4], ppty[4]),
						new Point(pptx[3], ppty[6])
					];
				}
				else {
					poly = [
						new Point(pptx[6], ppty[6]),
						new Point(pptx[8], ppty[8]),
						new Point(pptx[6], ppty[8])
					];
				}
			}
			else {
				if(rotation == 1) {
					poly = [
						new Point(pptx[14], ppty[14]),
						new Point(pptx[16], ppty[16]),
						new Point(pptx[16], ppty[14])
					];
				}
				else if(rotation == 2) {
					poly = [
						new Point(pptx[18], ppty[18]),
						new Point(pptx[20], ppty[20]),
						new Point(pptx[18], ppty[20])
					];
				}
				else if(rotation == 3) {
					poly = [
						new Point(pptx[22], ppty[22]),
						new Point(pptx[24], ppty[24]),
						new Point(pptx[20], ppty[23])
					];
				}
				else if(rotation == 4) {
					poly = [
						new Point(pptx[2], ppty[2]),
						new Point(pptx[4], ppty[4]),
						new Point(pptx[4], ppty[2])
					];
				}
				else if(rotation == 5) {
					poly = [
						new Point(pptx[6], ppty[6]),
						new Point(pptx[8], ppty[8]),
						new Point(pptx[6], ppty[8])
					];
				}
				else {
					poly = [
						new Point(pptx[10], ppty[10]),
						new Point(pptx[12], ppty[12]),
						new Point(pptx[8], ppty[11])
					];
				}
			}
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_CHANL:
			var poly = [
				new Point(pptx[2], ppty[2]),
				new Point(pptx[4], ppty[4]),
				new Point(pptx[14], ppty[14]),
				new Point(pptx[16], ppty[16])
			];
			poly = rotatePolygon(poly, rotation * 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_JUNCT:
			ctx.beginPath();
			ctx.arc(orix, oriy, (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10])), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			var poly = [
				new Point(pptx[4], ppty[4]),
				new Point(pptx[2], ppty[2]),
				new Point(((pptx[2] + pptx[16]) / 2), ((ppty[2] + ppty[16]) / 2)),
				new Point(((pptx[14] + pptx[4]) / 2), ((ppty[14] + ppty[4]) / 2))
			];
			poly = rotatePolygon(poly, rotation * 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_SLICE:
			var poly = [
				new Point(0, 0),
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_WIDE:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[13], ppty[13]),
				new Point(pptx[17], ppty[17])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 0 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_JOIN:
			ctx.beginPath();
			ctx.arc(orix, oriy, (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[1]) : (ppty[11] - ppty[9])), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(((pptx[5] + pptx[13]) / 2), ((ppty[5] + ppty[13]) / 2)),
				new Point(((pptx[1] + pptx[17]) / 2), ((ppty[1] + ppty[17]) / 2))
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 0 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_CORNR:
			ctx.beginPath();
			if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[7] + orix, ppty[7] + oriy);
					ctx.lineTo(orix, oriy);
					ctx.lineTo(pptx[23] + orix, ppty[23] + oriy);
				}
				else {
					ctx.moveTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[13] + oriy);
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[7] + orix, ppty[7] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[15] + orix, ppty[15] + oriy);
					ctx.lineTo(orix, oriy);
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[9] + oriy);
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[11] + orix, ppty[11] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[19] + orix, ppty[19] + oriy);
					ctx.lineTo(orix, oriy);
				}
				else {
					ctx.moveTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[1] + oriy);
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
				else {
					ctx.moveTo(pptx[19] + orix, ppty[19] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[3] + orix, ppty[3] + oriy);
					ctx.lineTo(orix, oriy);
				}
			}
			else {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
				else {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
			}
			ctx.closePath();
			ctx.fill();
			break;
		case HEX_FINSH:
			ctx.beginPath();
			if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(Math.round((pptx[3] + pptx[0]) / 2) + orix, Math.round((ppty[3] + ppty[0]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[9] + oriy);
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(Math.round((pptx[11] + pptx[0]) / 2) + orix, ppty[11] + oriy);
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[9] + orix, ppty[9] + oriy);
					ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(Math.round((pptx[5] + pptx[13]) / 2) + orix, Math.round((ppty[5] + ppty[13]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[13] + oriy);
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[13] + orix, ppty[13] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(Math.round((pptx[15] + pptx[0]) / 2) + orix, Math.round((ppty[15] + ppty[0]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[17] + orix, ppty[21] + oriy);
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[17] + orix, ppty[17] + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
				}
				else {
					ctx.moveTo(pptx[21] + orix, ppty[21] + oriy);
					ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(Math.round((pptx[23] + pptx[0]) / 2) + orix, ppty[23] + oriy);
				}
			}
			else {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(Math.round((pptx[1] + pptx[17]) / 2) + orix, Math.round((ppty[1] + ppty[17]) / 2) + oriy);
					ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				}
				else {
					ctx.moveTo(pptx[1] + orix, ppty[1] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
					ctx.lineTo(pptx[5] + orix, ppty[1] + oriy);
				}
			}
			ctx.closePath();
			ctx.fill();
			break;
		case HEX_BORDR:
			var poly = [
				new Point(pptx[1], ppty[1]),
				new Point(pptx[5], ppty[5]),
				new Point(pptx[6], ppty[6]),
				new Point(pptx[24], ppty[24])
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case HEX_LRSID:
			if(rotation == 0 || rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(((pptx[2] + pptx[3]) / 2) + orix, ((ppty[2] + ppty[3]) / 2) + oriy);
				ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
				ctx.lineTo(((pptx[14] + pptx[15]) / 2) + orix, ((ppty[14] + ppty[15]) / 2) + oriy);
				ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, ((ppty[15] + ppty[16]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1 || rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(((pptx[6] + pptx[7]) / 2) + orix, ((ppty[6] + ppty[7]) / 2) + oriy);
				ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
				ctx.lineTo(((pptx[18] + pptx[19]) / 2) + orix, ((ppty[18] + ppty[19]) / 2) + oriy);
				ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2 || rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(((pptx[10] + pptx[11]) / 2) + orix, ((ppty[10] + ppty[11]) / 2) + oriy);
				ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[12] + ppty[11]) / 2) + oriy);
				ctx.lineTo(((pptx[22] + pptx[23]) / 2) + orix, ((ppty[22] + ppty[23]) / 2) + oriy);
				ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case HEX_LRWAL:
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo(pptx[8] + orix, ppty[8] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[22] + orix, ppty[22] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo(pptx[12] + orix, ppty[12] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[2] + orix, ppty[2] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo(pptx[6] + orix, ppty[6] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[16] + orix, ppty[16] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(pptx[10] + orix, ppty[10] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[20] + orix, ppty[20] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(pptx[14] + orix, ppty[14] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[24] + orix, ppty[24] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(pptx[18] + orix, ppty[18] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[4] + orix, ppty[4] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case HEX_LRCNR:
			if(rotation == 0) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[8]) + orix, (ppty[8]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[8]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo((pptx[6]) + orix, ((ppty[10] + ppty[11]) / 2) + oriy);
					ctx.lineTo((pptx[12]) + orix, (ppty[12]) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[11] + pptx[10]) / 2) + orix, ((ppty[11] + ppty[10]) / 2) + oriy);
					ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[11] + pptx[10]) / 2) + orix, ((ppty[11] + ppty[10]) / 2) + oriy);
					ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[16] + pptx[15]) / 2) + orix, ((ppty[16] + ppty[15]) / 2) + oriy);
					ctx.lineTo(((pptx[14] + pptx[15]) / 2) + orix, ((ppty[14] + ppty[15]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[14] + pptx[4]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[14] + ppty[4]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[16] + pptx[2]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[16] + ppty[2]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[10]) + orix, (ppty[10]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[10]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo((pptx[14]) + orix, (ppty[14]) + oriy);
					ctx.lineTo((pptx[17]) + orix, (ppty[12]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[20]) + orix, (ppty[20]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[20]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[17] + pptx[16]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[24]) + orix, (ppty[24]) + oriy);
					ctx.lineTo((pptx[4]) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[23] + pptx[22]) / 2) + orix, ((ppty[23] + ppty[22]) / 2) + oriy);
					ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, ((ppty[3] + ppty[2]) / 2) + oriy);
					ctx.lineTo((Math.round((((pptx[3] + pptx[2]) / 2) + ((pptx[16] + pptx[15]) / 2)) / 2)) + orix, (Math.round((((ppty[3] + ppty[2]) / 2) + ((ppty[16] + ppty[15]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((((pptx[3] + pptx[4]) / 2) + ((pptx[14] + pptx[15]) / 2)) / 2)) + orix, (Math.round((((ppty[3] + ppty[4]) / 2) + ((ppty[14] + ppty[15]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[23] + pptx[22]) / 2) + orix, ((ppty[23] + ppty[22]) / 2) + oriy);
					ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[22]) + orix, (ppty[22]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[22]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo(((pptx[5] + pptx[6]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[2]) + orix, (ppty[2]) + oriy);
					ctx.lineTo((pptx[5]) + orix, (ppty[10]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			break;
		case HEX_SRSID:
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo(((pptx[2] + pptx[3]) / 2) + orix, ((ppty[2] + ppty[3]) / 2) + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, ((ppty[15] + ppty[16]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo(((pptx[6] + pptx[7]) / 2) + orix, ((ppty[6] + ppty[7]) / 2) + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo(((pptx[10] + pptx[11]) / 2) + orix, ((ppty[10] + ppty[11]) / 2) + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(((pptx[14] + pptx[15]) / 2) + orix, ((ppty[14] + ppty[15]) / 2) + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(((pptx[18] + pptx[19]) / 2) + orix, ((ppty[18] + ppty[19]) / 2) + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(((pptx[22] + pptx[23]) / 2) + orix, ((ppty[22] + ppty[23]) / 2) + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[12] + ppty[11]) / 2) + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case HEX_SRCNR:
			if(rotation == 0) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[2]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[5]) + orix, (ppty[5]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 1) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[6]) / 2) + orix, ((ppty[7] + ppty[6]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[11] + pptx[12]) / 2) + orix, ((ppty[11] + ppty[12]) / 2) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[11] + pptx[10]) / 2) + orix, ((ppty[11] + ppty[10]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[16] + pptx[2]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[16] + ppty[2]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[16] + pptx[15]) / 2) + orix, ((ppty[16] + ppty[15]) / 2) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 2) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[16]) / 2) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[9]) + orix, (ppty[9]) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[20]) / 2) + orix, ((ppty[19] + ppty[20]) / 2) + oriy);
					ctx.lineTo(((pptx[4] + pptx[5]) / 2) + orix, (Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[13]) + orix, (ppty[13]) + oriy);
					ctx.lineTo((pptx[17]) + orix, (ppty[17]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 3) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[15]) + oriy);
					ctx.lineTo(((pptx[15] + pptx[14]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[17]) + orix, (ppty[17]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo(((pptx[17] + pptx[16]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 4) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (pptx[3] - pptx[2]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[19] + pptx[18]) / 2) + orix, ((ppty[19] + ppty[18]) / 2) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[23] + pptx[24]) / 2) + orix, ((ppty[23] + ppty[24]) / 2) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.arc(pptx[0] + orix, ppty[0] + oriy, (ppty[11] - ppty[10]) / 2, 0, Math.PI*2);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, ((ppty[3] + ppty[4]) / 2) + oriy);
					ctx.lineTo((Math.round((((pptx[3] + pptx[4]) / 2) + ((pptx[14] + pptx[15]) / 2)) / 2)) + orix, (Math.round((((ppty[3] + ppty[4]) / 2) + ((ppty[14] + ppty[15]) / 2)) / 2)) + oriy);
					ctx.lineTo((Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2)) + orix, (Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2)) + oriy);
					ctx.lineTo(((pptx[23] + pptx[22]) / 2) + orix, ((ppty[23] + ppty[22]) / 2) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.closePath();
					ctx.fill();
					ctx.beginPath();
					ctx.closePath();
					ctx.fill();
				}
			}
			else if(rotation == 5) {
				if(mapGrain == GRAIN_VERTICAL) {
					ctx.beginPath();
					ctx.moveTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[3]) + oriy);
					ctx.lineTo(((pptx[3] + pptx[4]) / 2) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[21]) + orix, (ppty[21]) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
				else {
					ctx.beginPath();
					ctx.moveTo(((pptx[7] + pptx[8]) / 2) + orix, ((ppty[7] + ppty[8]) / 2) + oriy);
					ctx.lineTo(((pptx[5] + pptx[6]) / 2) + orix, (Math.round((ppty[0] + ((ppty[11] + ppty[12]) / 2)) / 2)) + oriy);
					ctx.lineTo((pptx[1]) + orix, (ppty[1]) + oriy);
					ctx.lineTo((pptx[5]) + orix, (ppty[5]) + oriy);
					ctx.closePath();
					ctx.fill();
				}
			}
			break;
		case HEX_PSEUD:
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[17] + orix, ppty[17] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[21] + orix, ppty[21] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[1] + orix, ppty[1] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo(pptx[0] + orix, ppty[0] + oriy);
				ctx.lineTo(pptx[5] + orix, ppty[5] + oriy);
				ctx.lineTo(pptx[9] + orix, ppty[9] + oriy);
				ctx.lineTo(pptx[13] + orix, ppty[13] + oriy);
				ctx.closePath();
				ctx.fill();
			}
			break;
		default:
			break;
	}
}

function drawBase(ctx, orix, oriy, pptx, ppty, ctyp, matr) {
	drawComponent(ctx, orix, oriy, pptx, ppty, ctyp, matr);
}

function drawOverlay(ctx, orix, oriy, pptx, ppty, ctyp, matr) {
	drawComponent(ctx, orix, oriy, pptx, ppty, ctyp, matr);
}

function drawTerrain(ctx, pptx, ppty, scl, ctyp, matr) {
	var orix = pptx[0];
	var oriy = ppty[0];
	var rotation  = ctyp % 10;
	var symbology = ctyp - rotation;
	ctx.fillStyle = matr;
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
	switch(symbology) {
		case TERRAIN_BOUNDER:
			var poly = [
				new Point( 0,             -0.75 * scl),
				new Point( 0.1875 * scl,  -0.1875 * scl),
				new Point( 0.75 * scl,    -0.1875 * scl),
				new Point( 0.28125 * scl,  0.140625 * scl),
				new Point( 0.46875 * scl,  0.703125 * scl),
				new Point( 0,              0.328125 * scl),
				new Point(-0.46875 * scl,  0.703125 * scl),
				new Point(-0.28125 * scl,  0.140625 * scl),
				new Point(-0.75 * scl,    -0.1875 * scl),
				new Point(-0.1875 * scl,  -0.1875 * scl)
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_WATER:
			ctx.beginPath();
			if(mapGrain == GRAIN_HORIZONTAL) {
				var pt = getLineIntersection(new Point(pptx[9], ppty[7]), new Point(pptx[8], ppty[9]), new Point(pptx[9], ppty[9]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[8], ppty[9]);
				ctx.lineTo(pptx[10], ppty[10]);
				ctx.lineTo(pptx[8], ppty[11]);
				ctx.lineTo(pptx[12], ppty[12]);
				ctx.lineTo(pptx[8], ppty[13]);
				pt = getLineIntersection(new Point(pptx[8], ppty[13]), new Point(pptx[9], ppty[15]), new Point(pptx[13], ppty[13]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[8], ppty[7]), new Point(pptx[7], ppty[9]), new Point(pptx[9], ppty[9]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[7], ppty[9]);
				ctx.lineTo(pptx[8], ppty[10]);
				ctx.lineTo(pptx[7], ppty[11]);
				ctx.lineTo(pptx[8], ppty[12]);
				ctx.lineTo(pptx[7], ppty[13]);
				pt = getLineIntersection(new Point(pptx[7], ppty[13]), new Point(pptx[14], ppty[15]), new Point(pptx[13], ppty[13]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[7], ppty[7]);
				ctx.lineTo(pptx[6], ppty[9]);
				ctx.lineTo(pptx[7], ppty[10]);
				ctx.lineTo(pptx[6], ppty[11]);
				ctx.lineTo(pptx[7], ppty[12]);
				ctx.lineTo(pptx[6], ppty[13]);
				ctx.lineTo(pptx[15], ppty[15]);
				ctx.moveTo(pptx[5], ppty[5]);
				ctx.lineTo(pptx[6], ppty[7]);
				ctx.lineTo(pptx[5], ppty[9]);
				ctx.lineTo(pptx[6], ppty[10]);
				ctx.lineTo(pptx[5], ppty[11]);
				ctx.lineTo(pptx[6], ppty[12]);
				ctx.lineTo(pptx[5], ppty[13]);
				ctx.lineTo(pptx[6], ppty[15]);
				ctx.lineTo(pptx[17], ppty[17]);
				pt = getLineIntersection(new Point(pptx[4], ppty[5]), new Point(pptx[5], ppty[3]), new Point(pptx[1], ppty[1]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[5], ppty[3]);
				ctx.lineTo(pptx[4], ppty[9]);
				ctx.lineTo(pptx[5], ppty[10]);
				ctx.lineTo(pptx[4], ppty[11]);
				ctx.lineTo(pptx[5], ppty[12]);
				ctx.lineTo(pptx[4], ppty[13]);
				ctx.lineTo(pptx[5], ppty[19]);
				pt = getLineIntersection(new Point(pptx[5], ppty[19]), new Point(pptx[4], ppty[17]), new Point(pptx[21], ppty[21]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[3], ppty[5]), new Point(pptx[4], ppty[3]), new Point(pptx[1], ppty[1]), new Point(pptx[5], ppty[5]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[4], ppty[3]);
				ctx.lineTo(pptx[3], ppty[9]);
				ctx.lineTo(pptx[4], ppty[10]);
				ctx.lineTo(pptx[3], ppty[11]);
				ctx.lineTo(pptx[4], ppty[12]);
				ctx.lineTo(pptx[3], ppty[13]);
				ctx.lineTo(pptx[4], ppty[19]);
				pt = getLineIntersection(new Point(pptx[4], ppty[19]), new Point(pptx[19], ppty[17]), new Point(pptx[21], ppty[21]), new Point(pptx[17], ppty[17]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[3], ppty[3]);
				ctx.lineTo(pptx[2], ppty[9]);
				ctx.lineTo(pptx[3], ppty[10]);
				ctx.lineTo(pptx[2], ppty[11]);
				ctx.lineTo(pptx[3], ppty[12]);
				ctx.lineTo(pptx[2], ppty[13]);
				ctx.lineTo(pptx[19], ppty[19]);
				ctx.moveTo(pptx[1], ppty[1]);
				ctx.lineTo(pptx[2], ppty[10]);
				ctx.lineTo(pptx[1], ppty[11]);
				ctx.lineTo(pptx[2], ppty[12]);
				ctx.lineTo(pptx[21], ppty[21]);
			}
			else {
				var pt = getLineIntersection(new Point(pptx[23], ppty[1]), new Point(pptx[1], ppty[24]), new Point(pptx[1], ppty[1]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[1], ppty[24]);
				ctx.lineTo(pptx[2], ppty[2]);
				ctx.lineTo(pptx[3], ppty[24]);
				ctx.lineTo(pptx[4], ppty[4]);
				ctx.lineTo(pptx[5], ppty[24]);
				pt = getLineIntersection(new Point(pptx[5], ppty[24]), new Point(pptx[7], ppty[1]), new Point(pptx[5], ppty[5]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[23], ppty[24]), new Point(pptx[1], ppty[23]), new Point(pptx[1], ppty[1]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[1], ppty[23]);
				ctx.lineTo(pptx[2], ppty[24]);
				ctx.lineTo(pptx[3], ppty[23]);
				ctx.lineTo(pptx[4], ppty[24]);
				ctx.lineTo(pptx[5], ppty[23]);
				pt = getLineIntersection(new Point(pptx[5], ppty[23]), new Point(pptx[7], ppty[6]), new Point(pptx[5], ppty[5]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[23], ppty[23]);
				ctx.lineTo(pptx[1],  ppty[22]);
				ctx.lineTo(pptx[2],  ppty[23]);
				ctx.lineTo(pptx[3],  ppty[22]);
				ctx.lineTo(pptx[4],  ppty[23]);
				ctx.lineTo(pptx[5],  ppty[22]);
				ctx.lineTo(pptx[7],  ppty[7]);
				ctx.moveTo(pptx[21], ppty[21]);
				ctx.lineTo(pptx[23], ppty[22]);
				ctx.lineTo(pptx[1],  ppty[21]);
				ctx.lineTo(pptx[2],  ppty[22]);
				ctx.lineTo(pptx[3],  ppty[21]);
				ctx.lineTo(pptx[4],  ppty[22]);
				ctx.lineTo(pptx[5],  ppty[21]);
				ctx.lineTo(pptx[7],  ppty[22]);
				ctx.lineTo(pptx[9],  ppty[9]);
				pt = getLineIntersection(new Point(pptx[21], ppty[20]), new Point(pptx[19], ppty[21]), new Point(pptx[17], ppty[17]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[19], ppty[21]);
				ctx.lineTo(pptx[1],  ppty[20]);
				ctx.lineTo(pptx[2],  ppty[21]);
				ctx.lineTo(pptx[3],  ppty[20]);
				ctx.lineTo(pptx[4],  ppty[21]);
				ctx.lineTo(pptx[5],  ppty[20]);
				ctx.lineTo(pptx[11], ppty[21]);
				pt = getLineIntersection(new Point(pptx[11], ppty[21]), new Point(pptx[9], ppty[20]), new Point(pptx[13], ppty[13]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				pt = getLineIntersection(new Point(pptx[21], ppty[19]), new Point(pptx[19], ppty[20]), new Point(pptx[17], ppty[17]), new Point(pptx[21], ppty[21]));
				ctx.moveTo(pt.x, pt.y);
				ctx.lineTo(pptx[19], ppty[20]);
				ctx.lineTo(pptx[1],  ppty[19]);
				ctx.lineTo(pptx[2],  ppty[20]);
				ctx.lineTo(pptx[3],  ppty[19]);
				ctx.lineTo(pptx[4],  ppty[20]);
				ctx.lineTo(pptx[5],  ppty[19]);
				ctx.lineTo(pptx[11], ppty[20]);
				pt = getLineIntersection(new Point(pptx[11], ppty[20]), new Point(pptx[9], ppty[11]), new Point(pptx[13], ppty[13]), new Point(pptx[9], ppty[9]));
				ctx.lineTo(pt.x, pt.y);
				ctx.moveTo(pptx[19], ppty[19]);
				ctx.lineTo(pptx[1],  ppty[18]);
				ctx.lineTo(pptx[2],  ppty[19]);
				ctx.lineTo(pptx[3],  ppty[18]);
				ctx.lineTo(pptx[4],  ppty[19]);
				ctx.lineTo(pptx[5],  ppty[18]);
				ctx.lineTo(pptx[11], ppty[11]);
				ctx.moveTo(pptx[17], ppty[17]);
				ctx.lineTo(pptx[2],  ppty[18]);
				ctx.lineTo(pptx[3],  ppty[17]);
				ctx.lineTo(pptx[4],  ppty[18]);
				ctx.lineTo(pptx[13], ppty[13]);
			}
			ctx.strokeStyle = matr;
			ctx.lineWidth = outlineThickness;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_MOUNTOUT:
		case TERRAIN_MOUNTFIL:
			var xunit = scl * 0.125;
			var yunit = scl * 0.2;
			var radunit = xunit * 0.9;
			ctx.lineWidth = outlineThickness;
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0], radunit * 4, radunit * 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (xunit * 3), ppty[0] - (yunit * 2), radunit * 2, radunit, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (xunit * 5), ppty[0], radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (xunit * 2), ppty[0] + (yunit * 2), radunit * 2, radunit, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (xunit * 2), ppty[0] - (yunit * 3), radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (xunit * 5), ppty[0] - yunit, radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (xunit * 2), ppty[0] + (yunit * 3), radunit, radunit / 2, -45 * ANGLE_STEP, 0, Math.PI*2);
			if(symbology == TERRAIN_MOUNTOUT) { ctx.strokeStyle = matr; ctx.stroke(); } else { ctx.fillStyle = matr; ctx.fill(); }
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_TREES:
			var unitx = (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10]));
			var unity = (mapGrain == GRAIN_VERTICAL ? (ppty[7] - ppty[6]) : (ppty[11] - ppty[10]) * CALCTHIN);
			var arcunitx = unitx / 2.5;
			var arcunity = unity / 2.5;
			ctx.fillStyle = matr;
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0], arcunitx * 4.0, arcunity * 4.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - unitx, ppty[0], arcunitx * 3.0, arcunity * 3.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0] - (unity * 2.0), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.0), ppty[0] - unity, arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (unitx * 2.0), ppty[0] - (unity * 0.5), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.0), ppty[0] + unity, arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (unitx * 2.0), ppty[0] + unity, arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + unitx, ppty[0] + (unity * 2.0), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0], ppty[0] + (unity * 2.5), arcunitx * 2.0, arcunity * 2.0, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - unitx, ppty[0] - (unity * 2.0), arcunitx * 1.5, arcunity * 1.5, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + unitx, ppty[0] - (unity * 2.0), arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - (unitx * 2.0), ppty[0] - (unity * 1.5), arcunitx * 1.5, arcunity * 1.5, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.5), ppty[0], arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] - unitx, ppty[0] + (unity * 2.0), arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			ctx.beginPath();
			ctx.ellipse(pptx[0] + (unitx * 2.0), ppty[0] + (unity * 2.0), arcunitx, arcunity, 0, 0, Math.PI*2);
			ctx.fill();
			break;
		case TERRAIN_SWAMP:
			var mxunit = scl * 0.125;
			var myunit = scl * 0.1;
			ctx.lineWidth = outlineThickness;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.moveTo(pptx[0] + mxunit, ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] + mxunit, ppty[0] - (myunit * 6));
			ctx.moveTo(pptx[0], ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] - mxunit, ppty[0] - (myunit * 6));
			ctx.moveTo(pptx[0] + (mxunit * 2), ppty[0] - (myunit * 4));
			ctx.lineTo(pptx[0] + (mxunit * 3), ppty[0] - (myunit * 6));
			ctx.stroke();
			ctx.beginPath();
			ctx.moveTo(pptx[0] - (mxunit * 1), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 2), ppty[0] + (myunit * 2));
			ctx.moveTo(pptx[0] - (mxunit * 3), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 3), ppty[0] + (myunit * 2));
			ctx.moveTo(pptx[0] - (mxunit * 5), ppty[0]);
			ctx.lineTo(pptx[0] - (mxunit * 4), ppty[0] + (myunit * 2));
			ctx.stroke();
			ctx.beginPath();
			ctx.moveTo(pptx[0] + mxunit, ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 2), ppty[0] + (myunit * 4));
			ctx.moveTo(pptx[0] + (mxunit * 3), ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 3), ppty[0] + (myunit * 4));
			ctx.moveTo(pptx[0] + (mxunit * 5), ppty[0] + (myunit * 2));
			ctx.lineTo(pptx[0] + (mxunit * 4), ppty[0] + (myunit * 4));
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_PLAINS:
			ctx.lineWidth = outlineThickness;
			ctx.beginPath();
			ctx.moveTo(pptx[7], ppty[7]);
			ctx.lineTo(pptx[7] - ((pptx[7] - pptx[19]) * 0.25), ppty[7] + ((ppty[19] - ppty[7])  * 0.25));
			ctx.moveTo(pptx[0], ppty[0]);
			ctx.lineTo(pptx[7] - ((pptx[7] - pptx[19]) * 0.75), ppty[7] + ((ppty[19] - ppty[7])  * 0.75));
			ctx.moveTo(pptx[5] - ((pptx[5] - pptx[21]) * 0.25), ppty[5] + ((ppty[21] - ppty[5])  * 0.25));
			ctx.lineTo(pptx[5] - ((pptx[5] - pptx[21]) * 0.5), ppty[5] + ((ppty[21] - ppty[5])  * 0.5));
			ctx.moveTo(pptx[5] - ((pptx[5] - pptx[21]) * 0.75), ppty[5] + ((ppty[21] - ppty[5])  * 0.75));
			ctx.lineTo(pptx[21], ppty[21]);
			ctx.moveTo(pptx[9] - ((pptx[9] - pptx[17]) * 0.25), ppty[9] + ((ppty[17] - ppty[9])  * 0.25));
			ctx.lineTo(pptx[9] - ((pptx[9] - pptx[17]) * 0.5), ppty[9] + ((ppty[17] - ppty[9])  * 0.5));
			ctx.moveTo(pptx[9] - ((pptx[9] - pptx[17]) * 0.75), ppty[9] + ((ppty[17] - ppty[9])  * 0.75));
			ctx.lineTo(pptx[17], ppty[17]);
			ctx.strokeStyle = matr;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case TERRAIN_DESERT:
			var unitx = (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10]));
			var unity = (mapGrain == GRAIN_VERTICAL ? (ppty[7] - ppty[6]) : (ppty[11] - ppty[10]) * CALCTHIN);
			var sandsize = Math.max(scl * 0.125, 1);
			var grainsize = Math.max(sandsize * 0.66, 1);
			ctx.fillStyle = matr;
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 2.0), ppty[0] - (unity * 3.0), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + unitx, ppty[0] - (unity * 2.5), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - unitx, ppty[0] - (unity * 1.3), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 3.15), ppty[0] - unity, sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.0), ppty[0], sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0], ppty[0] + (unity * 0.7), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 1.25), ppty[0] + (unity * 1.6), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 0.5), ppty[0] + (unity * 2.2), sandsize, sandsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 0.5), ppty[0] - (unity * 2.8), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - unitx, ppty[0] - (unity * 2.75), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.25), ppty[0] - (unity * 2.25), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.75), ppty[0] - (unity * 1.25), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 1.8), ppty[0] - (unity * 1.5), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 0.5), ppty[0] - (unitx * 0.5), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - unitx, ppty[0], grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 2.85), ppty[0] + unity, grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] - (unitx * 2.0), ppty[0] + (unity * 2.1), grainsize, grainsize);
			ctx.fill();
			ctx.beginPath();
			ctx.rect(pptx[0] + (unitx * 2.0), ppty[0] + (unity * 2.25), grainsize, grainsize);
			ctx.fill();
			break;
		case TERRAIN_CAPITOL:
			var csize = scl * 0.75;
			var poly = [
				new Point( 0,               -0.75     * csize),
				new Point( 0.1875  * csize, -0.1875   * csize),
				new Point( 0.75    * csize, -0.1875   * csize),
				new Point( 0.28125 * csize,  0.140625 * csize),
				new Point( 0.46875 * csize,  0.703125 * csize),
				new Point( 0,                0.328125 * csize),
				new Point(-0.46875 * csize,  0.703125 * csize),
				new Point(-0.28125 * csize,  0.140625 * csize),
				new Point(-0.75    * csize, -0.1875   * csize),
				new Point(-0.1875  * csize, -0.1875   * csize)
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_CITY:
			var cunit = scl * 0.45;
			ctx.beginPath();
			ctx.rect(pptx[0] - cunit, ppty[0] - cunit, cunit * 2, cunit * 2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case TERRAIN_VILLAGE:
			var vunit = scl * 0.45;
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], vunit, 0, Math.PI*2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case TERRAIN_FORT:
			var cunit = scl * 0.45;
			var qunit = cunit * 0.5;
			var poly = [
				new Point(-cunit, -cunit),
				new Point(-cunit + qunit, -cunit),
				new Point(-cunit + qunit, -cunit + qunit),
				new Point( cunit - qunit, -cunit + qunit),
				new Point( cunit - qunit, -cunit),
				new Point( cunit, -cunit),
				new Point( cunit,  cunit),
				new Point(-cunit,  cunit)
			];
			drawPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_MINE:
			var mxunit = scl * 0.125;
			var myunit = scl * 0.2;
			var poly = [
				new Point(-mxunit * 0.75, -myunit * 2.5),
				new Point( mxunit * 0.75, -myunit * 2.5),
				new Point( mxunit * 0.75, myunit * 0.5),
				new Point( mxunit * 2.0, 0),
				new Point( mxunit * 2.0, myunit),
				new Point( mxunit,       myunit * 2.0),
				new Point( 0,            myunit * 2.5),
				new Point(-mxunit,       myunit * 2.0),
				new Point(-mxunit * 2.0, myunit),
				new Point(-mxunit * 2.0, 0),
				new Point(-mxunit * 0.75, myunit * 0.5)
			];
			drawPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_RUIN:
			var vunit = scl * 0.45;
			var poly = [
				new Point(0, -vunit),
				new Point(-vunit, vunit * 0.72),
				new Point(vunit, vunit * 0.72)
			];
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case TERRAIN_INTEREST:
			var isize = scl * 0.75;
			var poly = [
				new Point( 0,               -0.75     * isize),
				new Point( 0.1875  * isize, -0.1875   * isize),
				new Point( 0.75    * isize, -0.1875   * isize),
				new Point( 0.28125 * isize,  0.140625 * isize),
				new Point( 0.46875 * isize,  0.703125 * isize),
				new Point( 0,                0.328125 * isize),
				new Point(-0.46875 * isize,  0.703125 * isize),
				new Point(-0.28125 * isize,  0.140625 * isize),
				new Point(-0.75    * isize, -0.1875   * isize),
				new Point(-0.1875  * isize, -0.1875   * isize)
			];
			drawPolygon(ctx, poly, orix, oriy, matr);
			break;
		default:
			break;
	}
}

function drawFeature(ctx, pptx, ppty, scl, ctyp, matr) {
	var orix = pptx[0];
	var oriy = ppty[0];
	var rotation  = ctyp % 10;
	var symbology = ctyp - rotation;
	ctx.fillStyle = matr;
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
	switch(symbology) {
		case FEATURE_DOOR:
			var thick = Math.max(Math.round(scl / 4), 1);
			var diffX = (pptx[17] - pptx[1]) * 1.0;
			var diffY = (ppty[17] - ppty[1]) * 1.0;
			var segStepX = (Math.abs(diffX) > Math.abs(diffY) ? 1.0 * Math.sign(diffX) : diffX / diffY);
			var segStepY = (Math.abs(diffY) > Math.abs(diffX) ? 1.0 * Math.sign(diffY) : diffY / diffX);
			var poly = [
				new Point(pptx[1] - orix, (ppty[1]) - oriy),
				new Point(pptx[5] - orix, (ppty[5]) - oriy),
				new Point((pptx[5] + Math.round(segStepX * thick)) - orix, (ppty[5] + Math.round(segStepY * thick)) - oriy),
				new Point((pptx[1] + Math.round(segStepX * thick)) - orix, (ppty[1] + Math.round(segStepY * thick)) - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 60 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_STAIRS:
			var diffX = (pptx[15] - pptx[3]) / 13.0;
			var diffY = (ppty[15] - ppty[3]) / 13.0;
			var poly1 = [
				new Point((pptx[2] - orix) + (diffX * 1.0), (ppty[2] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 1.0), (ppty[4] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 3.0), (ppty[4] - oriy) + (diffY * 3.0)),
				new Point((pptx[2] - orix) + (diffX * 3.0), (ppty[2] - oriy) + (diffY * 3.0))
			];
			var poly2 = [
				new Point((pptx[2] - orix) + (diffX * 4.0), (ppty[2] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 4.0), (ppty[4] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 6.0), (ppty[4] - oriy) + (diffY * 6.0)),
				new Point((pptx[2] - orix) + (diffX * 6.0), (ppty[2] - oriy) + (diffY * 6.0))
			];
			var poly3 = [
				new Point((pptx[2] - orix) + (diffX * 7.0), (ppty[2] - oriy) + (diffY * 7.0)),
				new Point((pptx[4] - orix) + (diffX * 7.0), (ppty[4] - oriy) + (diffY * 7.0)),
				new Point((pptx[4] - orix) + (diffX * 9.0), (ppty[4] - oriy) + (diffY * 9.0)),
				new Point((pptx[2] - orix) + (diffX * 9.0), (ppty[2] - oriy) + (diffY * 9.0))
			];
			var poly4 = [
				new Point((pptx[2] - orix) + (diffX * 10.0), (ppty[2] - oriy) + (diffY * 10.0)),
				new Point((pptx[4] - orix) + (diffX * 10.0), (ppty[4] - oriy) + (diffY * 10.0)),
				new Point((pptx[4] - orix) + (diffX * 12.0), (ppty[4] - oriy) + (diffY * 12.0)),
				new Point((pptx[2] - orix) + (diffX * 12.0), (ppty[2] - oriy) + (diffY * 12.0))
			];
			poly1 = rotatePolygon(poly1, (rotation * 60));
			poly2 = rotatePolygon(poly2, (rotation * 60));
			poly3 = rotatePolygon(poly3, (rotation * 60));
			poly4 = rotatePolygon(poly4, (rotation * 60));
			fillPolygon(ctx, poly1, orix, oriy, matr);
			fillPolygon(ctx, poly2, orix, oriy, matr);
			fillPolygon(ctx, poly3, orix, oriy, matr);
			fillPolygon(ctx, poly4, orix, oriy, matr);
			break;
		case FEATURE_STAIR:
			var diffX = (pptx[15] - pptx[3]) / 13.0;
			var diffY = (ppty[15] - ppty[3]) / 13.0;
			var poly1 = [
				new Point((pptx[2] - orix) + (diffX * 1.0), (ppty[2] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 1.0), (ppty[4] - oriy) + (diffY * 1.0)),
				new Point((pptx[4] - orix) + (diffX * 3.0), (ppty[4] - oriy) + (diffY * 3.0)),
				new Point((pptx[2] - orix) + (diffX * 3.0), (ppty[2] - oriy) + (diffY * 3.0))
			];
			var poly2 = [
				new Point((pptx[2] - orix) + (diffX * 4.0), (ppty[2] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 4.0), (ppty[4] - oriy) + (diffY * 4.0)),
				new Point((pptx[4] - orix) + (diffX * 6.0), (ppty[4] - oriy) + (diffY * 6.0)),
				new Point((pptx[2] - orix) + (diffX * 6.0), (ppty[2] - oriy) + (diffY * 6.0))
			];
			poly1 = rotatePolygon(poly1, (rotation * 60));
			poly2 = rotatePolygon(poly2, (rotation * 60));
			fillPolygon(ctx, poly1, orix, oriy, matr);
			fillPolygon(ctx, poly2, orix, oriy, matr);
			break;
		case FEATURE_SHAFT:
			ctx.beginPath();
			ctx.arc(orix, oriy, Math.round(scl / 2), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			ctx.lineWidth = outlineThickness;
			ctx.strokeStyle = hexOutline;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case FEATURE_UNDERS:
			var diffY = (ppty[15] - ppty[3]) / 9.0;
			var symbRadius = Math.max(scl / 10, 1);
			var symbCenter = new Point(0.0, -(diffY * 3.0));
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = -Math.round(diffY);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = Math.round(diffY);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = Math.round(diffY * 3.0);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			break;
		case FEATURE_UNDER:
			var diffY = (ppty[15] - ppty[3]) / 9.0;
			var symbRadius = Math.max(scl / 10, 1);
			var symbCenter = new Point(0.0, -(diffY * 3.0));
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			symbCenter.x = 0;
			symbCenter.y = -Math.round(diffY);
			symbCenter = rotatePoint(symbCenter, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + symbCenter.x, oriy + symbCenter.y, symbRadius, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			break;
		case FEATURE_BANDS:
			if(rotation == 0 || rotation == 3) {
				ctx.beginPath();
				ctx.moveTo((pptx[2] + pptx[3]) / 2, (ppty[2] + ppty[3]) / 2);
				ctx.lineTo((pptx[3] + pptx[4]) / 2, (ppty[3] + ppty[4]) / 2);
				ctx.lineTo((pptx[14] + pptx[15]) / 2, (ppty[14] + ppty[15]) / 2);
				ctx.lineTo((pptx[15] + pptx[16]) / 2, (ppty[15] + ppty[16]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1 || rotation == 4) {
				ctx.beginPath();
				ctx.moveTo((pptx[6] + pptx[7]) / 2, (ppty[6] + ppty[7]) / 2);
				ctx.lineTo((pptx[7] + pptx[8]) / 2, (ppty[7] + ppty[8]) / 2);
				ctx.lineTo((pptx[18] + pptx[19]) / 2, (ppty[18] + ppty[19]) / 2);
				ctx.lineTo((pptx[19] + pptx[20]) / 2, (ppty[19] + ppty[20]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2 || rotation == 5) {
				ctx.beginPath();
				ctx.moveTo((pptx[10] + pptx[11]) / 2, (ppty[10] + ppty[11]) / 2);
				ctx.lineTo((pptx[11] + pptx[12]) / 2, (ppty[12] + ppty[11]) / 2);
				ctx.lineTo((pptx[22] + pptx[23]) / 2, (ppty[22] + ppty[23]) / 2);
				ctx.lineTo((pptx[23] + pptx[24]) / 2, (ppty[23] + ppty[24]) / 2);
				ctx.closePath();
				ctx.fill();
			}
			break;
		case FEATURE_BAND:
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10])) / 2, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo((pptx[3] + pptx[4]) / 2, (ppty[3] + ppty[4]) / 2);
				ctx.lineTo((pptx[3] + pptx[2]) / 2, (ppty[3] + ppty[2]) / 2);
				ctx.lineTo(Math.round((((pptx[3] + pptx[2]) / 2) + ((pptx[16] + pptx[15]) / 2)) / 2), Math.round((((ppty[3] + ppty[2]) / 2) + ((ppty[16] + ppty[15]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[3] + pptx[4]) / 2) + ((pptx[14] + pptx[15]) / 2)) / 2), Math.round((((ppty[3] + ppty[4]) / 2) + ((ppty[14] + ppty[15]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo((pptx[7] + pptx[6]) / 2, (ppty[7] + ppty[6]) / 2);
				ctx.lineTo((pptx[7] + pptx[8]) / 2, (ppty[7] + ppty[8]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2), Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2), Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo((pptx[11] + pptx[10]) / 2, (ppty[11] + ppty[10]) / 2);
				ctx.lineTo((pptx[11] + pptx[12]) / 2, (ppty[11] + ppty[12]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2), Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2), Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo((pptx[16] + pptx[15]) / 2, (ppty[16] + ppty[15]) / 2);
				ctx.lineTo((pptx[14] + pptx[15]) / 2, (ppty[14] + ppty[15]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[14] + pptx[4]) / 2)) / 2), Math.round((ppty[0] + ((ppty[14] + ppty[4]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[16] + pptx[2]) / 2)) / 2), Math.round((ppty[0] + ((ppty[16] + ppty[2]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo((pptx[19] + pptx[18]) / 2, (ppty[19] + ppty[18]) / 2);
				ctx.lineTo((pptx[19] + pptx[20]) / 2, (ppty[19] + ppty[20]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[20] + pptx[6]) / 2)) / 2), Math.round((ppty[0] + ((ppty[20] + ppty[6]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[8] + pptx[18]) / 2)) / 2), Math.round((ppty[0] + ((ppty[8] + ppty[18]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo((pptx[23] + pptx[22]) / 2, (ppty[23] + ppty[22]) / 2);
				ctx.lineTo((pptx[23] + pptx[24]) / 2, (ppty[23] + ppty[24]) / 2);
				ctx.lineTo(Math.round((pptx[0] + ((pptx[24] + pptx[10]) / 2)) / 2), Math.round((ppty[0] + ((ppty[24] + ppty[10]) / 2)) / 2));
				ctx.lineTo(Math.round((pptx[0] + ((pptx[12] + pptx[22]) / 2)) / 2), Math.round((ppty[0] + ((ppty[12] + ppty[22]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			break;
		case FEATURE_AXONS:
			var poly = [
				new Point(((pptx[24] + pptx[1]) / 2) - orix, ((ppty[24] + ppty[1]) / 2) - oriy),
				new Point(pptx[1] - orix, ppty[1] - oriy),
				new Point(((pptx[1] + pptx[2]) / 2) - orix, ((ppty[1] + ppty[2]) / 2) - oriy),
				new Point(((pptx[12] + pptx[13]) / 2) - orix, ((ppty[12] + ppty[13]) / 2) - oriy),
				new Point(pptx[13] - orix, ppty[13] - oriy),
				new Point(((pptx[13] + pptx[14]) / 2) - orix, ((ppty[13] + ppty[14]) / 2) - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_AXON:
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], (mapGrain == GRAIN_VERTICAL ? (ppty[24] - ppty[1]) : (pptx[5] - pptx[4])) / 2, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			if(rotation == 0) {
				ctx.beginPath();
				ctx.moveTo((pptx[24] + pptx[1]) / 2, (ppty[24] + ppty[1]) / 2);
				ctx.lineTo(pptx[1], ppty[1]);
				ctx.lineTo((pptx[1] + pptx[2]) / 2, (ppty[1] + ppty[2]) / 2);
				ctx.lineTo(Math.round((((pptx[1] + pptx[2]) / 2) + ((pptx[13] + pptx[12]) / 2)) / 2), Math.round((((ppty[1] + ppty[2]) / 2) + ((ppty[13] + ppty[12]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[1] + pptx[24]) / 2) + ((pptx[13] + pptx[14]) / 2)) / 2), Math.round((((ppty[1] + ppty[24]) / 2) + ((ppty[13] + ppty[14]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 1) {
				ctx.beginPath();
				ctx.moveTo((pptx[4] + pptx[5]) / 2, (ppty[4] + ppty[5]) / 2);
				ctx.lineTo(pptx[5], ppty[5]);
				ctx.lineTo((pptx[5] + pptx[6]) / 2, (ppty[5] + ppty[6]) / 2);
				ctx.lineTo(Math.round((((pptx[5] + pptx[6]) / 2) + ((pptx[17] + pptx[16]) / 2)) / 2), Math.round((((ppty[5] + ppty[6]) / 2) + ((ppty[17] + ppty[16]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[5] + pptx[4]) / 2) + ((pptx[17] + pptx[18]) / 2)) / 2), Math.round((((ppty[5] + ppty[4]) / 2) + ((ppty[17] + ppty[18]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 2) {
				ctx.beginPath();
				ctx.moveTo((pptx[8] + pptx[9]) / 2, (ppty[8] + ppty[9]) / 2);
				ctx.lineTo(pptx[9], ppty[9]);
				ctx.lineTo((pptx[9] + pptx[10]) / 2, (ppty[9] + ppty[10]) / 2);
				ctx.lineTo(Math.round((((pptx[9] + pptx[10]) / 2) + ((pptx[21] + pptx[20]) / 2)) / 2), Math.round((((ppty[9] + ppty[10]) / 2) + ((ppty[21] + ppty[20]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[9] + pptx[8]) / 2) + ((pptx[21] + pptx[22]) / 2)) / 2), Math.round((((ppty[9] + ppty[8]) / 2) + ((ppty[21] + ppty[22]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 3) {
				ctx.beginPath();
				ctx.moveTo((pptx[12] + pptx[13]) / 2, (ppty[12] + ppty[13]) / 2);
				ctx.lineTo(pptx[13], ppty[13]);
				ctx.lineTo((pptx[13] + pptx[14]) / 2, (ppty[13] + ppty[14]) / 2);
				ctx.lineTo(Math.round((((pptx[1] + pptx[24]) / 2) + ((pptx[13] + pptx[14]) / 2)) / 2), Math.round((((ppty[1] + ppty[24]) / 2) + ((ppty[13] + ppty[14]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[1] + pptx[2]) / 2) + ((pptx[13] + pptx[12]) / 2)) / 2), Math.round((((ppty[1] + ppty[2]) / 2) + ((ppty[13] + ppty[12]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 4) {
				ctx.beginPath();
				ctx.moveTo((pptx[16] + pptx[17]) / 2, (ppty[16] + ppty[17]) / 2);
				ctx.lineTo(pptx[17], ppty[17]);
				ctx.lineTo((pptx[17] + pptx[18]) / 2, (ppty[17] + ppty[18]) / 2);
				ctx.lineTo(Math.round((((pptx[5] + pptx[4]) / 2) + ((pptx[17] + pptx[18]) / 2)) / 2), Math.round((((ppty[5] + ppty[4]) / 2) + ((ppty[17] + ppty[18]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[5] + pptx[6]) / 2) + ((pptx[17] + pptx[16]) / 2)) / 2), Math.round((((ppty[5] + ppty[6]) / 2) + ((ppty[17] + ppty[16]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			else if(rotation == 5) {
				ctx.beginPath();
				ctx.moveTo((pptx[20] + pptx[21]) / 2, (ppty[20] + ppty[21]) / 2);
				ctx.lineTo(pptx[21], ppty[21]);
				ctx.lineTo((pptx[21] + pptx[22]) / 2, (ppty[21] + ppty[22]) / 2);
				ctx.lineTo(Math.round((((pptx[9] + pptx[8]) / 2) + ((pptx[21] + pptx[22]) / 2)) / 2), Math.round((((ppty[9] + ppty[8]) / 2) + ((ppty[21] + ppty[22]) / 2)) / 2));
				ctx.lineTo(Math.round((((pptx[9] + pptx[10]) / 2) + ((pptx[21] + pptx[20]) / 2)) / 2), Math.round((((ppty[9] + ppty[10]) / 2) + ((ppty[21] + ppty[20]) / 2)) / 2));
				ctx.closePath();
				ctx.fill();
			}
			break;
		case FEATURE_DNDOOR:
			var thick = Math.max((scl / 8), 1);
			var dunit = scl / 4;
			var yoff  = -(scl - (dunit / 2));
			var poly = [
				new Point(dunit * -2.0, yoff),
				new Point(dunit * -1.0, yoff),
				new Point(dunit * -1.0, yoff + thick),
				new Point(dunit *  1.0, yoff + thick),
				new Point(dunit *  1.0, yoff),
				new Point(dunit *  2.0, yoff),
				new Point(dunit *  2.0, yoff + thick),
				new Point(dunit *  1.0 + thick, yoff + thick),
				new Point(dunit *  1.0 + thick, yoff + thick * 2.0),
				new Point(dunit * -1.0 - thick, yoff + thick * 2.0),
				new Point(dunit * -1.0 - thick, yoff + thick),
				new Point(dunit * -2.0, yoff + thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_SPIRAL:
			var minrad = scl * 0.2;
			var maxrad = scl * 0.6;
			ctx.lineWidth = outlineThickness;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], minrad, 0, Math.PI*2);
			ctx.closePath();
			ctx.stroke();
			ctx.beginPath();
			ctx.arc(pptx[0], ppty[0], maxrad, 0, Math.PI*2);
			ctx.closePath();
			ctx.stroke();
			for(var sstp = 0; sstp < 360; sstp += 45) {
				ctx.moveTo(pptx[0] + (minrad * Math.cos(sstp * ANGLE_STEP)), ppty[0]  + (minrad * Math.sin(sstp * ANGLE_STEP)));
				ctx.lineTo(pptx[0] + (maxrad * Math.cos(sstp * ANGLE_STEP)), ppty[0]  + (maxrad * Math.sin(sstp * ANGLE_STEP)));
				ctx.stroke();
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case FEATURE_WINDOW:
			var thick = Math.max((scl / 16), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[2]) : (ppty[11] - ppty[10]));
			var yoff  = -(scl - (dunit / 2));
			var poly = [
				new Point(dunit *  2.0, yoff),
				new Point(dunit *  2.0, yoff + thick * 4.0),
				new Point(dunit *  1.0 + thick, yoff + thick * 4.0),
				new Point(dunit *  1.0 + thick, yoff + thick),
				new Point(dunit *  0.0 + thick, yoff + thick),
				new Point(dunit *  0.0 + thick, yoff + thick * 3.0),
				new Point(dunit *  0.0 - thick, yoff + thick * 3.0),
				new Point(dunit *  0.0 - thick, yoff + thick),
				new Point(dunit * -1.0 - thick, yoff + thick),
				new Point(dunit * -1.0 - thick, yoff + thick * 3.0),
				new Point(dunit *  1.0 + thick, yoff + thick * 3.0),
				new Point(dunit *  1.0 + thick, yoff + thick * 4.0),
				new Point(dunit * -2.0, yoff + thick * 4.0),
				new Point(dunit * -2.0, yoff)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_MIDOOR:
			var thick = Math.max((scl / 8), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, 0),
				new Point(dunit *  2.0, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit * -2.0, thick),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -1.0, thick),
				new Point(dunit *  1.0, thick),
				new Point(dunit *  1.0, 0),
				new Point(dunit *  1.0, -thick),
				new Point(dunit * -1.0, -thick),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -2.0, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  2.0, -thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_VERTICAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_MIWIND:
			var thick = Math.max((scl / 16), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, -thick * 2.0),
				new Point(dunit *  2.0,  thick * 2.0),
				new Point(dunit *  1.0 + thick,  thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  0.0 + thick, -thick),
				new Point(dunit *  0.0 + thick,  thick),
				new Point(dunit *  0.0 - thick,  thick),
				new Point(dunit *  0.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, thick * 2.0),
				new Point(dunit * -2.0, thick * 2.0),
				new Point(dunit * -2.0, -thick * 2.0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_VERTICAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_TDOOR:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[24] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[22] - orix, ppty[22] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[5] - orix, (ppty[6] - oriy) + thick),
				new Point(pptx[1] - orix, (ppty[24] - oriy) + thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, (ppty[8] - oriy) - thick),
				new Point(pptx[1] - orix, (ppty[22] - oriy) - thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_TWIND:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[24] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[22] - orix, ppty[22] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[24] - oriy),
				new Point(pptx[5] - orix, ppty[6] - oriy),
				new Point(pptx[5] - orix, (ppty[6] - oriy) + thick),
				new Point(pptx[1] - orix, (ppty[24] - oriy) + thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[1] - orix, ppty[22] - oriy),
				new Point(pptx[5] - orix, ppty[8] - oriy),
				new Point(pptx[5] - orix, (ppty[8] - oriy) - thick),
				new Point(pptx[1] - orix, (ppty[22] - oriy) - thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point((pptx[3] - orix) - thick, ppty[6] - oriy),
				new Point((pptx[3] - orix) + thick, ppty[6] - oriy),
				new Point((pptx[3] - orix) + thick, ppty[8] - oriy),
				new Point((pptx[3] - orix) - thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_EDOOR:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[10] - orix, ppty[10] - oriy),
				new Point(pptx[12] - orix, ppty[12] - oriy),
				new Point(pptx[12] - orix, ppty[10] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[8] - orix, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[6] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_EWIND:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[6] - orix, ppty[6] - oriy),
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[10] - orix, ppty[10] - oriy),
				new Point(pptx[12] - orix, ppty[12] - oriy),
				new Point(pptx[12] - orix, ppty[10] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[8] - orix, ppty[8] - oriy),
				new Point(pptx[8] - orix, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[10] - oriy),
				new Point((pptx[8] - orix) - thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[6] - orix, ppty[8] - oriy),
				new Point(pptx[6] - orix, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[10] - oriy),
				new Point((pptx[6] - orix) + thick, ppty[8] - oriy)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = [
				new Point(pptx[6] - orix, (ppty[9] - oriy) - thick),
				new Point(pptx[6] - orix, (ppty[9] - oriy) + thick),
				new Point(pptx[8] - orix, (ppty[9] - oriy) + thick),
				new Point(pptx[8] - orix, (ppty[9] - oriy) - thick)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_ADOOR:
			var thick = Math.max((scl / 8), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, 0),
				new Point(dunit *  2.0, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick * 2.0),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit * -2.0, thick),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -1.0, thick),
				new Point(dunit *  1.0, thick),
				new Point(dunit *  1.0, 0),
				new Point(dunit *  1.0, -thick),
				new Point(dunit * -1.0, -thick),
				new Point(dunit * -1.0, 0),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -2.0, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit * -1.0 - thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick * 2.0),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  2.0, -thick)
			];
			poly = rotatePolygon(poly, (mapGrain == GRAIN_VERTICAL ? 60 : 30));
			poly[21] = new Point(((pptx[12] + pptx[13]) / 2) - orix, ((ppty[12] + ppty[13]) / 2) - oriy);
			poly[0] = new Point(pptx[13] - orix, ppty[13] - oriy);
			poly[1] = new Point(((pptx[13] + pptx[14]) / 2) - orix, ((ppty[13] + ppty[14]) / 2) - oriy);
			poly[6] = new Point(((pptx[24] + pptx[1]) / 2) - orix, ((ppty[24] + ppty[1]) / 2) - oriy);
			poly[7] = new Point(pptx[1] - orix, ppty[1] - oriy);
			poly[15] = new Point(pptx[1] - orix, ppty[1] - oriy);
			poly[16] = new Point(((pptx[1] + pptx[2]) / 2) - orix, ((ppty[1] + ppty[2]) / 2) - oriy);
			poly = rotatePolygon(poly, (rotation * 60));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_AWIND:
			var thick = Math.max((scl / 8), 1);
			var halfthick = Math.max((thick / 2), 1);
			var dunit = (mapGrain == GRAIN_VERTICAL ? (ppty[15] - ppty[3]) : (pptx[12] - pptx[23])) / 4;
			var poly = [
				new Point(dunit *  2.0, 0),
				new Point(dunit *  2.0, thick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit *  1.0 + thick, -halfthick),
				new Point(halfthick, -halfthick),
				new Point(halfthick, halfthick),
				new Point(-halfthick, halfthick),
				new Point(-halfthick, -halfthick),
				new Point(dunit * -1.0 - thick, -halfthick),
				new Point(dunit * -1.0 - thick, halfthick),
				new Point(dunit *  1.0 + thick, halfthick),
				new Point(dunit *  1.0 + thick, thick),
				new Point(dunit * -1.0 - thick, thick),
				new Point(dunit * -2.0, thick),
				new Point(dunit * -2.0, 0),
				new Point(dunit * -2.0, -thick),
				new Point(dunit * -1.0 - thick, -thick),
				new Point(dunit *  1.0 + thick, -thick),
				new Point(dunit *  2.0, -thick)
			];
			poly = rotatePolygon(poly, (mapGrain == GRAIN_VERTICAL ? 60 : 30));
			poly[18] = new Point(((pptx[12] + pptx[13]) / 2) - orix, ((ppty[12] + ppty[13]) / 2) - oriy);
			poly[0] = new Point(pptx[13] - orix, ppty[13] - oriy);
			poly[1] = new Point(((pptx[13] + pptx[14]) / 2) - orix, ((ppty[13] + ppty[14]) / 2) - oriy);
			poly[13] = new Point(((pptx[24] + pptx[1]) / 2) - orix, ((ppty[24] + ppty[1]) / 2) - oriy);
			poly[14] = new Point(pptx[1] - orix, ppty[1] - oriy);
			poly[15] = new Point(((pptx[1] + pptx[2]) / 2) - orix, ((ppty[1] + ppty[2]) / 2) - oriy);
			poly = rotatePolygon(poly, (rotation * 60));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case FEATURE_PILLAR:
			ctx.beginPath();
			ctx.arc(orix, oriy, (mapGrain == GRAIN_VERTICAL ? (pptx[3] - pptx[1]) : (ppty[11] - ppty[9])), 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			break;
		default:
			break;
	}
}

function drawDecor(ctx, pptx, ppty, scl, ctyp, matr) {
	var orix = pptx[0];
	var oriy = ppty[0];
	var rotation  = ctyp % 10;
	var symbology = ctyp - rotation;
	ctx.fillStyle = matr;
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
	switch(symbology) {
		case DECOR_TRAP:
			var thick = Math.max((scl / 12), 1);
			var tunit = scl / 3;
			var qunit = scl / 6;
			var ounit = scl / 10;
			var sunit = scl / 16;
			var rotangle = rotation * 60;
			var poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, -tunit),
				new Point(tunit, tunit),
				new Point(-tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-qunit, -qunit),
				new Point(qunit, qunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-qunit, qunit),
				new Point(qunit, -qunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			break;
		case DECOR_TRAPDOOR:
			var thick = Math.max((scl / 12), 1);
			var tunit = scl / 3;
			var qunit = scl / 4;
			var ounit = scl / 10;
			var sunit = scl / 16;
			var rotangle = rotation * 60;
			var poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, -tunit),
				new Point(tunit, tunit),
				new Point(-tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			ctx.beginPath();
			var dotpoint = rotatePoint(new Point(0, ounit), rotangle);
			ctx.arc(orix + dotpoint.x, oriy + dotpoint.y, ounit, 0, Math.PI*2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case DECOR_FLOORPIT:
			var thick = Math.max((scl / 12), 1);
			var tunit = scl / 3;
			var qunit = scl / 4;
			var ounit = scl / 6;
			var sunit = scl / 16;
			var rotangle = rotation * 60;
			var poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, -tunit),
				new Point(tunit, tunit),
				new Point(-tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-tunit, -tunit),
				new Point(tunit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-ounit, -ounit),
				new Point(tunit, -ounit),
				new Point(tunit, tunit),
				new Point(-ounit, tunit)
			];
			poly = rotatePolygon(poly, rotangle);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case DECOR_TRIPWIRE:
			var thick = Math.max((scl / 12), 1);
			var funit = scl / 9;
			var rotangle = (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0);
			var poly = [
				new Point(-(funit * 9), 0),
				new Point(-(funit * 7), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-(funit * 5), 0),
				new Point(-(funit * 3), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point(-(funit * 1), 0),
				new Point((funit * 1), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point((funit * 3), 0),
				new Point((funit * 5), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			poly = [
				new Point((funit * 7), 0),
				new Point((funit * 9), 0)
			];
			poly = rotatePolygon(poly, rotangle);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			break;
		case DECOR_CHESTL:
			var thick = Math.max((scl / 12), 1);
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var poly = [
					new Point(pptx[1] - orix, ppty[7] - oriy),
					new Point(pptx[5] - orix, ppty[7] - oriy),
					new Point(pptx[5] - orix, ppty[11] - oriy),
					new Point(pptx[1] - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] - orix, ppty[7] - oriy),
					new Point(pptx[2] - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[7] - oriy),
					new Point(pptx[4] - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[10] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[10] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[11] - oriy),
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[11] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			else {
				var poly = [
					new Point(pptx[3] - orix, ppty[9] - oriy),
					new Point(pptx[7] - orix, ppty[9] - oriy),
					new Point(pptx[7] - orix, ppty[13] - oriy),
					new Point(pptx[3] - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[9] - oriy),
					new Point(pptx[4] - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[6] - orix, ppty[9] - oriy),
					new Point(pptx[6] - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[12] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[12] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[13] - oriy),
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[13] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			break;
		case DECOR_CHESTS:
			var thick = Math.max((scl / 12), 1);
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var poly = [
					new Point(pptx[1] - orix, ppty[8] - oriy),
					new Point(pptx[5] - orix, ppty[8] - oriy),
					new Point(pptx[5] - orix, ppty[10] - oriy),
					new Point(pptx[1] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] - orix, ppty[8] - oriy),
					new Point(pptx[2] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[8] - oriy),
					new Point(pptx[4] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[9] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[9] - oriy),
					new Point(pptx[3] + ((pptx[4] - pptx[3]) / 2) - orix, ppty[10] - oriy),
					new Point(pptx[2] + ((pptx[3] - pptx[2]) / 2) - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			else {
				var poly = [
					new Point(pptx[3] - orix, ppty[10] - oriy),
					new Point(pptx[7] - orix, ppty[10] - oriy),
					new Point(pptx[7] - orix, ppty[12] - oriy),
					new Point(pptx[3] - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] - orix, ppty[10] - oriy),
					new Point(pptx[4] - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[6] - orix, ppty[10] - oriy),
					new Point(pptx[6] - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				drawPolygon(ctx, poly, orix, oriy, matr, thick);
				poly = [
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[11] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[11] - oriy),
					new Point(pptx[5] + ((pptx[6] - pptx[5]) / 2) - orix, ppty[12] - oriy),
					new Point(pptx[4] + ((pptx[5] - pptx[4]) / 2) - orix, ppty[12] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr, thick);
			}
			break;
		case DECOR_KEY:
			var thick = Math.max((scl / 12), 1);
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var ringwidth = (pptx[2] - pptx[24]) / 2;
				var centerpoint = rotatePoint(new Point((pptx[24] + ringwidth) - orix, ppty[21] - oriy), rotangle);
				ctx.beginPath();
				ctx.arc(centerpoint.x + orix, centerpoint.y + oriy, ringwidth, 0, Math.PI*2);
				ctx.closePath();
				ctx.stroke();
				var linepoint1 = rotatePoint(new Point(pptx[2] - orix, ppty[9] - oriy), rotangle);
				var linepoint2 = rotatePoint(new Point((pptx[5] - orix) + Math.max(thick / 2, 1), ppty[9] - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				var toothbase = ppty[10] + ((ppty[11] - ppty[10]) / 2);
				linepoint1 = rotatePoint(new Point(pptx[5] - orix, ppty[9] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point(pptx[5] - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				linepoint1 = rotatePoint(new Point(pptx[4] - orix, ppty[9] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point(pptx[4] - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			}
			else {
				var ringwidth = (ppty[24] - ppty[2]) / 2;
				var centerpoint = rotatePoint(new Point((pptx[2] + ringwidth) - orix, ppty[23] - oriy), rotangle);
				ctx.beginPath();
				ctx.arc(centerpoint.x + orix, centerpoint.y + oriy, ringwidth, 0, Math.PI*2);
				ctx.closePath();
				ctx.stroke();
				var linepoint1 = rotatePoint(new Point((pptx[2] - orix) + (ringwidth * 2), ppty[11] - oriy), rotangle);
				var linepoint2 = rotatePoint(new Point(((pptx[8] - ((pptx[8] - pptx[7]) / 2)) - orix) + Math.max(thick / 2, 1), ppty[11] - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				var toothbase = ppty[12] + ((ppty[13] - ppty[12]) / 2);
				linepoint1 = rotatePoint(new Point((pptx[8] - ((pptx[8] - pptx[7]) / 2)) - orix, ppty[11] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point((pptx[8] - ((pptx[8] - pptx[7]) / 2)) - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				linepoint1 = rotatePoint(new Point((pptx[7] - ((pptx[7] - pptx[6]) / 2)) - orix, ppty[11] - oriy), rotangle);
				linepoint2 = rotatePoint(new Point((pptx[7] - ((pptx[7] - pptx[6]) / 2)) - orix, toothbase - oriy), rotangle);
				ctx.beginPath();
				ctx.moveTo(linepoint1.x + orix, linepoint1.y + oriy);
				ctx.lineTo(linepoint2.x + orix, linepoint2.y + oriy);
				ctx.closePath();
				ctx.stroke();
				ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			}
			break;
		case DECOR_PENTAGRM:
			var dunit = scl / 2;
			var thick = scl / 16;
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.arc(orix, oriy, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.stroke();
			var currLineCap = ctx.lineCap;
			ctx.lineCap = "round";
			var starrot = (rotation * 60);
			var starpoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - dunit), starrot);
			for(var pointloop = 0; pointloop < 5; pointloop++) {
				ctx.beginPath();
				ctx.moveTo(orix + starpoint.x, oriy + starpoint.y);
				starrot = starrot + 144;
				starpoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - dunit), starrot);
				ctx.lineTo(orix + starpoint.x, oriy + starpoint.y);
				ctx.closePath();
				ctx.stroke();
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			ctx.lineCap = ctx.lineCap;
			break;
		case DECOR_GATE:
			var xunit = scl / 2;
			var yunit = scl / 2;
			var mxunit = scl / 9;
			var myunit = scl / 12;
			var thick = scl / 32;
			var rotradians = (rotation * 30) * (Math.PI / 180);
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			for(var el = 0; el < 5; el++) {
				ctx.beginPath();
				ctx.ellipse(orix, oriy, xunit, yunit, rotradians, 0, Math.PI*2);
				ctx.closePath();
				ctx.stroke();
				xunit = xunit - mxunit;
				yunit = yunit - myunit;
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		case DECOR_LIGHT:
			var dunit = scl / 6;
			var sunit = dunit * 2;
			var eunit = dunit * 3;
			var thick = scl / 16;
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.arc(orix, oriy, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.fill();
			var currLineCap = ctx.lineCap;
			ctx.lineCap = "round";
			var baserot = (rotation * 10) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0);
			var spoint = null;
			var epoint = null;
			for(var pointloop = 0; pointloop < 6; pointloop++) {
				spoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - sunit), baserot);
				epoint = rotatePoint(new Point(pptx[0] - orix, (ppty[0] - oriy) - eunit), baserot);
				ctx.beginPath();
				ctx.moveTo(orix + spoint.x, oriy + spoint.y);
				ctx.lineTo(orix + epoint.x, oriy + epoint.y);
				ctx.closePath();
				ctx.stroke();
				baserot = baserot + 60;
			}
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			ctx.lineCap = ctx.lineCap;
			break;
		case DECOR_SCONCE:
			var dunit = scl / 6;
			var sunit = dunit * 2;
			var eunit = dunit * 3;
			var thick = scl / 16;
			var rotangle = (rotation * 60);
			var rotradians = (rotangle - (mapGrain == GRAIN_HORIZONTAL ? 30 : 0)) * (Math.PI / 180);
			var mountpoint = rotatePoint(new Point(pptx[3] - orix, ppty[3] - oriy), rotangle);
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.beginPath();
			ctx.ellipse(mountpoint.x + orix, mountpoint.y + oriy, dunit, dunit, rotradians, 0, Math.PI);
			ctx.closePath();
			ctx.fill();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			break;
		default:
			break;
	}
}

function drawSymbol(ctx, pptx, ppty, scl, ctyp, matr) {
	var orix = pptx[0];
	var oriy = ppty[0];
	var rotation  = ctyp % 10;
	var symbology = ctyp - rotation;
	ctx.fillStyle = matr;
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
	switch(symbology) {
		case SYMBOL_ARROWEL:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -(dunit * 2.5)),
				new Point(dunit * 2.5, 0),
				new Point(dunit, 0),
				new Point(dunit, (dunit * 2.5)),
				new Point(-dunit, (dunit * 2.5)),
				new Point(-dunit, 0),
				new Point(-(dunit * 2.5), 0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ARROWES:
			var dunit = scl / 4;
			var vunit = scl / 3;
			var poly = [
				new Point(0, -(dunit * 1.5) - vunit),
				new Point(dunit * 1.5, - vunit),
				new Point(dunit * 0.5, - vunit),
				new Point(dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, - vunit),
				new Point(-(dunit * 1.5), - vunit)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ARROWVL:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -(dunit * 2.5)),
				new Point(dunit * 2.5, 0),
				new Point(dunit, 0),
				new Point(dunit, (dunit * 2.5)),
				new Point(-dunit, (dunit * 2.5)),
				new Point(-dunit, 0),
				new Point(-(dunit * 2.5), 0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ARROWVS:
			var dunit = scl / 4;
			var vunit = scl / 3;
			var poly = [
				new Point(0, -(dunit * 1.5) - vunit),
				new Point(dunit * 1.5, - vunit),
				new Point(dunit * 0.5, - vunit),
				new Point(dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, (dunit * 1.5) - vunit),
				new Point(-dunit * 0.5, - vunit),
				new Point(-(dunit * 1.5), - vunit)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ACCENTL:
			var rotangle = rotation * 60;
			if(mapGrain == GRAIN_VERTICAL) {
				var poly = [
					new Point(pptx[1] - orix, ppty[6] - oriy),
					new Point(pptx[5] - orix, ppty[6] - oriy),
					new Point(pptx[4] - orix, ppty[7] - oriy),
					new Point(pptx[2] - orix, ppty[7] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr);
			}
			else {
				var poly = [
					new Point(pptx[8] - orix, ppty[9] - oriy),
					new Point(pptx[8] - orix, ppty[13] - oriy),
					new Point(pptx[7] - orix, ppty[12] - oriy),
					new Point(pptx[7] - orix, ppty[10] - oriy)
				];
				poly = rotatePolygon(poly, rotangle);
				fillPolygon(ctx, poly, orix, oriy, matr);
			}
			break;
		case SYMBOL_ACCENTC:
			var vunit = scl * 0.7;
			var cunit = scl * 0.15;
			ctx.beginPath();
			var dotpoint = rotatePoint(new Point(0, -vunit), (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			ctx.arc(orix + dotpoint.x, oriy + dotpoint.y, cunit, 0, Math.PI*2);
			ctx.fillStyle = matr;
			ctx.fill();
			break;
		case SYMBOL_ACCENTT:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -dunit * 3.25),
				new Point(dunit, -dunit * 2),
				new Point(-dunit, -dunit * 2)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_ACCENTD:
			var dunit = scl / 4;
			var poly = [
				new Point(0, -dunit * 3.25),
				new Point(dunit * 0.5, -dunit * 2.5),
				new Point(0, -dunit * 1.75),
				new Point(-dunit * 0.5, -dunit * 2.5)
			];
			poly = rotatePolygon(poly, (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0) + 30);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_PROHIBL:
			var dunit = scl / 2;
			var thick = scl / 12;
			ctx.beginPath();
			ctx.arc(orix, oriy, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			var poly = [
				new Point(-dunit, 0),
				new Point(dunit, 0)
			];
			poly = rotatePolygon(poly, (rotation * 60) + 60);
			drawPolygon(ctx, poly, orix, oriy, matr, thick);
			break;
		case SYMBOL_PROHIBS:
			var dunit = scl / 4;
			var vunit = scl / 2.5;
			var thick = scl / 16;
			var ctrpoint = rotatePoint(new Point(0, -vunit), (rotation * 60) + (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			ctx.beginPath();
			ctx.arc(orix + ctrpoint.x, oriy + ctrpoint.y, dunit, 0, Math.PI*2);
			ctx.closePath();
			ctx.lineWidth = thick;
			ctx.strokeStyle = matr;
			ctx.stroke();
			ctx.lineWidth = DEFAULT_LINE_THICKNESS;
			var poly = [
				new Point(-dunit, 0),
				new Point(dunit, 0)
			];
			poly = rotatePolygon(poly, 60);
			drawPolygon(ctx, poly, orix + ctrpoint.x, oriy + ctrpoint.y, matr, thick);
			break;
		case SYMBOL_MARKER:
			var poly = [
				new Point( 0,             -0.75 * scl),
				new Point( 0.1875 * scl,  -0.1875 * scl),
				new Point( 0.75 * scl,    -0.1875 * scl),
				new Point( 0.28125 * scl,  0.140625 * scl),
				new Point( 0.46875 * scl,  0.703125 * scl),
				new Point( 0,              0.328125 * scl),
				new Point(-0.46875 * scl,  0.703125 * scl),
				new Point(-0.28125 * scl,  0.140625 * scl),
				new Point(-0.75 * scl,    -0.1875 * scl),
				new Point(-0.1875 * scl,  -0.1875 * scl)
			];
			if(mapGrain == GRAIN_HORIZONTAL) { poly = rotatePolygon(poly, 35); }
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		case SYMBOL_EVENT:
			hexpts = getNorthernHexPoints(orix, oriy, scl);
			pptx = hexpts[0];
			ppty = hexpts[1];
			var thick = Math.max((scl / 16), 1);
			var poly = [
				new Point(pptx[4] - orix, (ppty[7] - oriy)),
				new Point((pptx[3] - orix) - (thick / 2), (ppty[7] - oriy)),
				new Point((pptx[3] - orix) - (thick / 2), (ppty[6] - oriy)),
				new Point((pptx[3] - orix) + (thick / 2), (ppty[6] - oriy)),
				new Point((pptx[3] - orix) + (thick / 2), (ppty[7] - oriy)),
				new Point(pptx[2] - orix, (ppty[23] - oriy)),
				new Point(pptx[2] - orix, (ppty[23] - oriy) + thick),
				new Point(pptx[4] - orix, (ppty[7] - oriy) + thick)
			];
			poly = rotatePolygon(poly, (mapGrain == GRAIN_HORIZONTAL ? 30 : 0));
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			poly = rotatePolygon(poly, 60);
			fillPolygon(ctx, poly, orix, oriy, matr);
			break;
		default:
			break;
	}
}

function drawAnnotation(ctx, pptx, ppty, scl, ctyp, matr) {
	var fontStr = "bold " + scl + "px RobotoMono";
	var orix = pptx[0];
	var oriy = ppty[0];
	ctx.font = fontStr;
	ctx.fillStyle = matr;
	ctx.textAlign = "center";
	ctx.textBaseline = "middle";
	ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
}

function drawAnnotations(ctx, pptx, ppty, scl, annots) {
	var fontStr = "bold " + scl + "px RobotoMono";
	if(annots.length > 1) {
		// draw first character of annotation pair
		var ctyp = annots[0].ctype;
		var matr = materialDef[annots[0].material];
		var orix = pptx[0] - (scl / 3);
		var oriy = ppty[0];
		ctx.font = fontStr;
		ctx.fillStyle = matr;
		ctx.textAlign = "center";
		ctx.textBaseline = "middle";
		ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
		// draw second character of annotation pair
		ctyp = annots[1].ctype;
		matr = materialDef[annots[1].material];
		orix = pptx[0] + (scl / 3);
		oriy = ppty[0];
		ctx.font = fontStr;
		ctx.fillStyle = matr;
		ctx.textAlign = "center";
		ctx.textBaseline = "middle";
		ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
	}
	else {
		// draw single annotation character
		var ctyp = annots[0].ctype;
		var matr = materialDef[annots[0].material];
		var orix = pptx[0];
		var oriy = ppty[0];
		ctx.font = fontStr;
		ctx.fillStyle = matr;
		ctx.textAlign = "center";
		ctx.textBaseline = "middle";
		ctx.fillText(String.fromCharCode(ctyp - ANNOTATION_BASE), orix, oriy);
	}
}

function drawAnnotationEraser(ctx, pptx, ppty, scl) {
	var cunit = scl * 0.5;
	var xs = pptx[0] - cunit;
	var ys = ppty[0] - cunit;
	var xe = pptx[0] + cunit;
	var ye = ppty[0] + cunit;
	ctx.beginPath();
	for(yv = ys; yv < ye; yv = yv + 3.0) {
		ctx.moveTo(xs, yv);
		ctx.lineTo(xe, yv);
	}
	ctx.moveTo(xe, ye);
	ctx.closePath();
	ctx.strokeStyle = materialDef[3];
	ctx.stroke();
}

function drawHexOutline(ctx, pptx, ppty, clr, wid) {
	if(typeof clr != "undefined") {} else { clr = hexOutline; }
	if(typeof wid != "undefined") {} else { wid = outlineThickness; }
	ctx.beginPath();
	ctx.moveTo(pptx[1], ppty[1]);
	ctx.lineTo(pptx[5], ppty[5]);
	ctx.lineTo(pptx[9], ppty[9]);
	ctx.lineTo(pptx[13], ppty[13]);
	ctx.lineTo(pptx[17], ppty[17]);
	ctx.lineTo(pptx[21], ppty[21]);
	ctx.closePath();
	ctx.strokeStyle = clr;
	ctx.lineWidth = wid;
	ctx.stroke();
}

function drawMegahex(ctx, col, row, clr, thick, dotwt) {
	var hexpts;
	ctx.fillStyle = clr;
	ctx.strokeStyle = clr;
	ctx.lineWidth = thick;
	// draw center hex
	var hcol = col;
	var hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
	}
	// draw hex above
	hcol = col - (mapGrain == GRAIN_HORIZONTAL ? ((row - 1) % 2) : 0);
	hrow = row - 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw hex upper-left (need to calculate based on row/column)
	hcol = col - 1;
	hrow = row - (mapGrain == GRAIN_HORIZONTAL ? 0 : ((col - 1) % 2));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.stroke();
	}
	// draw hex upper-right (need to calculate based on row/column)
	hcol = col + (mapGrain == GRAIN_HORIZONTAL ? (row % 2) : 1);
	hrow = row - (mapGrain == GRAIN_HORIZONTAL ? 1 : ((col + 1) % 2));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw hex below
	hcol = col + (mapGrain == GRAIN_HORIZONTAL ? Math.abs(row % 2) : 0);
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	// draw hex lower-left
	hcol = col - (mapGrain == GRAIN_HORIZONTAL ? ((row + 1) % 2) : 1);
	hrow = row + 1 - (mapGrain == GRAIN_HORIZONTAL ? 0 : ((col - 1) % 2));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw hex upper-right
	hcol = col + 1;
	hrow = row + (mapGrain == GRAIN_HORIZONTAL ? 0 : (1 - ((col + 1) % 2)));
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.stroke();
	}
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
}

function drawSquareMegahexVertical(ctx, col, row, clr, thick, dotwt) {
	var hexpts;
	ctx.fillStyle = clr;
	ctx.strokeStyle = clr;
	ctx.lineWidth = thick;
	// draw center upper hex
	var hcol = col;
	var hrow = row + ((col + 1) % 2);
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.stroke();
	}
	// draw center lower hex
	hcol = col;
	hrow = row + 1 + ((col + 1) % 2);
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.stroke();
	}
	// draw upper left hex
	hcol = col - 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw center left hex
	hcol = col - 1;
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw lower left hex
	hcol = col - 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw upper right hex
	hcol = col + 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw center right hex
	hcol = col + 1;
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw lower right hex
	hcol = col + 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
}

function drawSquareMegahexHorizontal(ctx, col, row, clr, thick, dotwt) {
	var hexpts;
	ctx.fillStyle = clr;
	ctx.strokeStyle = clr;
	ctx.lineWidth = thick;
	// draw center left hex
	var hcol = col - ((row + 1) % 2);
	var hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.stroke();
	}
	// draw center right hex
	hcol = col + (row % 2);
	hrow = row + 1;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.stroke();
	}
	// draw upper center hex
	hcol = col;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw lower center hex
	hcol = col;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	// draw upper left hex
	hcol = col - 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.stroke();
	}
	// draw lower left hex
	hcol = col - 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.lineTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.stroke();
	}
	// draw upper right hex
	hcol = col + 1;
	hrow = row;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][1], hexpts[1][1]);
		ctx.lineTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.stroke();
	}
	// draw lower right hex
	hcol = col + 1;
	hrow = row + 2;
	if(hcol >= 0 && hcol < gridx && hrow >= 0 && hrow < gridy) {
		hexpts = getHexFineCoords(hcol, hrow);
		ctx.beginPath();
		ctx.arc(hexpts[0][0], hexpts[1][0], dotwt, 0, Math.PI*2);
		ctx.closePath();
		ctx.fill();
		ctx.beginPath();
		ctx.moveTo(hexpts[0][5], hexpts[1][5]);
		ctx.lineTo(hexpts[0][9], hexpts[1][9]);
		ctx.lineTo(hexpts[0][13], hexpts[1][13]);
		ctx.lineTo(hexpts[0][17], hexpts[1][17]);
		ctx.lineTo(hexpts[0][21], hexpts[1][21]);
		ctx.stroke();
	}
	ctx.lineWidth = DEFAULT_LINE_THICKNESS;
}

function drawMegahexGrid(grain, startX, startY, offsetFactor) {
	var plotX = startX;
	var plotY = startY;
	if(grain == GRAIN_HORIZONTAL) {
		var xOffset = 0;
		var yOffset = 0;
		var nextIncr = (startY == -1);
		while(plotY < gridy + 2) {
			while(plotX < gridx + 2) {
				drawMegahex(mhx, plotX, plotY + yOffset, mhexOutline, megahexWeight, megahexDotWeight);
				if(yOffset == 0) {
					plotX += (2 + Math.abs(plotY % 2));
					yOffset = 1;
				}
				else if(yOffset == 1) {
					plotX += 2;
					yOffset = -1;
				}
				else {
					plotX += (3 - Math.abs(plotY % 2));
					yOffset = 0;
				}
			}
			plotY += 3;
			if(nextIncr) { xOffset++; }
			if(xOffset > 1) {
				xOffset = -5;
				nextIncr = true;
			}
			nextIncr = !nextIncr;
			yOffset = 0;
			plotX = startX + xOffset;
		}
	}
	else {
		var prestep = offsetFactor;
		while(plotY < gridy + 5) {
			var step = 0;
			var stepBias = prestep;
			var loopRise = 0;
			var colpos = plotX;
			while(colpos < gridx + 2) {
				drawMegahex(mhx, colpos, plotY - Math.floor((step + stepBias) / 2) + loopRise, mhexOutline, megahexWeight, megahexDotWeight);
				colpos += 3;
				step++;
				if(step > 5) { step = 0; colpos--; stepBias = Math.abs(stepBias - 1); if(stepBias == 0) { loopRise--; } }
			}
			if(plotX == -1) {
				plotX = 1;
				plotY += 2;
				prestep = 0;
			}
			else if(plotX == 1) {
				plotX = 0;
				plotY += 3;
				prestep = 1;
			}
			else if(plotX == 0) {
				plotX = -1;
				plotY += 2;
				prestep = 0;
			}
		}
	}
}

function drawSquareMegahexRow(colpos, rowpos, isoddrow) {
	var drawSquare = isoddrow;
	while(colpos < gridx + 2) {
		if(drawSquare) {
			drawSquareMegahexVertical(mhx, colpos, rowpos, mhexOutline, megahexWeight, megahexDotWeight);
		}
		else {
			drawMegahex(mhx, colpos, rowpos + 1, mhexOutline, megahexWeight, megahexDotWeight);
		}
		drawSquare = !drawSquare;
		colpos += 3;
	}
}

function drawSquareMegahexCol(colpos, rowpos, isoddrow) {
	var drawSquare = isoddrow;
	while(rowpos < gridy + 2) {
		if(drawSquare) {
			drawSquareMegahexHorizontal(mhx, colpos, rowpos, mhexOutline, megahexWeight, megahexDotWeight);
		}
		else {
			drawMegahex(mhx, colpos, rowpos + 1, mhexOutline, megahexWeight, megahexDotWeight);
		}
		drawSquare = !drawSquare;
		rowpos += 3;
	}
}

function correctMegahexOffsetsForGrain() {
	if(mapGrain == GRAIN_HORIZONTAL) {
		if(mhexImposeY < -2) {
			mhexImposeY = 0;
		}
		else if(mhexImposeY > 0) {
			mhexImposeY = -2;
		}
		if(mhexImposeX < -6) {
			mhexImposeX = 0;
		}
		else if(mhexImposeX > 0) {
			mhexImposeX = -6;
		}
	}
	else {
		if(mhexImposeY < -6) {
			mhexImposeY = 0;
		}
		else if(mhexImposeY > 0) {
			mhexImposeY = -6;
		}
		if(mhexImposeX < -1) {
			mhexImposeX = 1;
		}
		else if(mhexImposeX > 1) {
			mhexImposeX = -1;
		}
	}
}

function shiftMegahexes(dir) {
	if(mapGrain == GRAIN_HORIZONTAL) {
		switch(dir) {
			case CARDINAL_N:
				mhexImposeY -= 1;
				break;
			case CARDINAL_S:
				mhexImposeY += 1;
				break;
			case CARDINAL_W:
				mhexImposeX -= 1;
				break;
			case CARDINAL_E:
				mhexImposeX += 1;
				break;
			default:
				break;
		}
	}
	else {
		switch(dir) {
			case CARDINAL_N:
				mhexImposeY -= 1;
				break;
			case CARDINAL_S:
				mhexImposeY += 1;
				break;
			case CARDINAL_W:
				mhexImposeX -= 1;
				break;
			case CARDINAL_E:
				mhexImposeX += 1;
				break;
			default:
				break;
		}
	}
	correctMegahexOffsetsForGrain();
	mhexImposeS = 0;
	if(mhexImposeX == 0) { mhexImposeS = 1; }
	if(mhexImposeY == 0 && mhexImposeX != 0) { mhexImposeS = 0; }
	imposeMegahexes();
}

function imposeMegahexes() {
	mhx.clearRect(0, 0, mhexes.width, mhexes.height);
	if(mhexType == MEGAHEX_TYPE_SQUARE) {
		if(mapGrain == GRAIN_HORIZONTAL) {
			var colnum = 1;
			var oddrow = true;
			while(colnum < gridx) {
				drawSquareMegahexCol(colnum, 0, oddrow);
				colnum += 2 + (oddrow ? 0 : 1);
				oddrow = !oddrow;
			}
		}
		else {
			var rownum = 0;
			var oddrow = true;
			while(rownum < gridy) {
				drawSquareMegahexRow(1, rownum, oddrow);
				rownum += 2 + (oddrow ? 0 : 1);
				oddrow = !oddrow;
			}
		}
	}
	else {
		drawMegahexGrid(mapGrain, mhexImposeX, mhexImposeY, mhexImposeS);
	}
	doubleBuffer();
}

function toggleMegahexes(show) {
	megahexesOn = show;
	drawMap();
	if(show) {
		correctMegahexOffsetsForGrain();
		imposeMegahexes();
		document.getElementById('mhexshow').className = CSS_PAL_BUTTON + (mhexType == MEGAHEX_TYPE_STANDARD ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
		document.getElementById('mhexsqr').className = CSS_PAL_BUTTON + (mhexType == MEGAHEX_TYPE_SQUARE ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
	}
	else {
		document.getElementById('mhexshow').className = CSS_PAL_BUTTON;
		document.getElementById('mhexsqr').className = CSS_PAL_BUTTON;
	}
	cnv.clearRect(0, 0, canvas.width, canvas.height);
	doubleBuffer();
}

function setMegahexType(mhtype) {
	mhexType = mhtype;
}

function standardMegahexes() {
	if(megahexesOn && mhexType == MEGAHEX_TYPE_STANDARD) {
		toggleMegahexes(false);
	}
	else {
		setMegahexType(MEGAHEX_TYPE_STANDARD);
		toggleMegahexes(true);
	}
}

function squareMegahexes() {
	if(megahexesOn && mhexType == MEGAHEX_TYPE_SQUARE) {
		toggleMegahexes(false);
	}
	else {
		setMegahexType(MEGAHEX_TYPE_SQUARE);
		toggleMegahexes(true);
	}
}

function plotHex(ctx, x, y, isCursor) {
	var hexcor = getHexCoords(drawscale);
	var hexptx = getHexX(y, x);
	var hexpty = getHexY(y, x);
	var hexfin = getHexFineCoords(x, y);
	var hexptr = hexarray[x][y];
	var hexoln;
	var hexwgt;
	var isDrawCursor = (isCursor && activeMode != MODE_CLONE);

	/* if duping a hex, use that hex as the definition */
	if(isCursor && activeMode == MODE_CLONE && dupeHex != null) {
		hexptr = dupeHex;
	}
	/* conditional drawing */
	if(isCursor && activeMode == MODE_CLONE && dupeHex == null) {
		hexoln = dupeHexOutline;
		hexwgt = outlineThickness + 1;
	}
	else if(isCursor && activeMode == MODE_BULKE) {
		/* check if we are bulk erasing the hex */
		drawBase(ctx, hexptx, hexpty, hexcor[0], hexcor[1], HEX_WHOLE, materialDef[MATERIAL_EMPTY]);
	}
	else {
		/* draw base components */
		if(layerVisibility[LAYER_BASE]) {
			if(hexptr.bases.length > 0) {
				for(var i = 0; i < hexptr.bases.length; i++) {
					drawBase(ctx, hexptx, hexpty, hexcor[0], hexcor[1], hexptr.bases[i].ctype, materialDef[hexptr.bases[i].material]);
				}
			}
			if(isDrawCursor && activeLayer == LAYER_BASE) {
				if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
				drawBase(ctx, hexptx, hexpty, hexcor[0], hexcor[1], activeBase, (activeBase == HEX_EMPTY || activeMode == MODE_ERASE ? materialDef[MATERIAL_EMPTY] : materialDef[activeMaterial]));
				if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
			}
		}

		/* draw terrain components */
		if(layerVisibility[LAYER_TERRAIN]) {
			if(hexptr.terrains.length > 0) {
				for(var i = 0; i < hexptr.terrains.length; i++) {
					drawTerrain(ctx, hexfin[0], hexfin[1], drawscale, hexptr.terrains[i].ctype, materialDef[hexptr.terrains[i].material]);
				}
			}
			if(isDrawCursor && activeLayer == LAYER_TERRAIN) {
				if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
				drawTerrain(ctx, hexfin[0], hexfin[1], drawscale, activeTerrain, (activeTerrain == TERRAIN_EMPTY || activeMode == MODE_ERASE ? materialDef[MATERIAL_EMPTY] : materialDef[activeMaterial]));
				if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
			}
		}

		/* draw overlay components */
		if(layerVisibility[LAYER_OVERLAY]) {
			if(hexptr.overlays.length > 0) {
				for(var i = 0; i < hexptr.overlays.length; i++) {
					drawOverlay(ctx, hexptx, hexpty, hexcor[0], hexcor[1], hexptr.overlays[i].ctype, materialDef[hexptr.overlays[i].material]);
				}
			}
			if(isDrawCursor && activeLayer == LAYER_OVERLAY) {
				if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
				drawOverlay(ctx, hexptx, hexpty, hexcor[0], hexcor[1], activeOverlay, (activeOverlay == HEX_EMPTY || activeMode == MODE_ERASE ? materialDef[MATERIAL_EMPTY] : materialDef[activeMaterial]));
				if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
			}
		}

		/* draw feature components */
		if(layerVisibility[LAYER_FEATURE]) {
			if(hexptr.features.length > 0) {
				for(var i = 0; i < hexptr.features.length; i++) {
					drawFeature(ctx, hexfin[0], hexfin[1], drawscale, hexptr.features[i].ctype, materialDef[hexptr.features[i].material]);
				}
			}
			if(isDrawCursor && activeLayer == LAYER_FEATURE) {
				if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
				drawFeature(ctx, hexfin[0], hexfin[1], drawscale, activeFeature, (activeFeature == FEATURE_EMPTY || activeMode == MODE_ERASE ? materialDef[MATERIAL_EMPTY] : materialDef[activeMaterial]));
				if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
			}
		}

		/* draw decor components */
		if(layerVisibility[LAYER_DECOR]) {
			if(hexptr.decors.length > 0) {
				for(var i = 0; i < hexptr.decors.length; i++) {
					drawDecor(ctx, hexfin[0], hexfin[1], drawscale, hexptr.decors[i].ctype, materialDef[hexptr.decors[i].material]);
				}
			}
			if(isDrawCursor && activeLayer == LAYER_DECOR) {
				if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
				drawDecor(ctx, hexfin[0], hexfin[1], drawscale, activeDecor, (activeDecor == DECOR_EMPTY || activeMode == MODE_ERASE ? materialDef[MATERIAL_EMPTY] : materialDef[activeMaterial]));
				if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
			}
		}

		/* draw symbol components */
		if(layerVisibility[LAYER_SYMBOL]) {
			if(hexptr.symbols.length > 0) {
				for(var i = 0; i < hexptr.symbols.length; i++) {
					drawSymbol(ctx, hexfin[0], hexfin[1], drawscale, hexptr.symbols[i].ctype, materialDef[hexptr.symbols[i].material]);
				}
			}
			if(isDrawCursor && activeLayer == LAYER_SYMBOL) {
				if(translucencyEnabled) { ctx.globalAlpha = TRANSLUCE; }
				drawSymbol(ctx, hexfin[0], hexfin[1], drawscale, activeSymbol, (activeSymbol == SYMBOL_EMPTY || activeMode == MODE_ERASE ? materialDef[MATERIAL_EMPTY] : materialDef[activeMaterial]));
				if(translucencyEnabled) { ctx.globalAlpha = OPAQUE; }
			}
		}

		/* draw annotation components */
		if(layerVisibility[LAYER_ANNOTATE]) {
			if(hexptr.annotations.length > 0) {
				drawAnnotations(ctx, hexfin[0], hexfin[1], drawscale, hexptr.annotations);
			}
			if(isDrawCursor && activeLayer == LAYER_ANNOTATE) {
				if(activeAnnotation == ANNOTATION_BASE || activeMode == MODE_ERASE) {
					drawAnnotationEraser(ctx, hexfin[0], hexfin[1], drawscale);
				}
				else {
					drawAnnotation(ctx, hexfin[0], hexfin[1], drawscale, activeAnnotation, materialDef[activeMaterial]);
				}
			}
		}
	}

	/* draw outline */
	drawHexOutline(ctx, hexfin[0], hexfin[1], hexoln, hexwgt);
}

function plotHexes(ctx) {
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			plotHex(ctx, x, y, false);
		}
	}
}

function plotTool(ctx, x, y) {
	plotHex(ctx, x, y, true);
	if(modeStartX > -1 && modeStartY > -1) {
		if(activeMode == MODE_LINE) {
			var slope = getSlope(modeStartX, modeStartY, x, y);
			if(mapGrain == GRAIN_VERTICAL) {
				if(slope <= 0.25) { y = modeStartY; }
				else if(slope >= 0.75) { x = modeStartX; }
				else { y = modeStartY + (Math.round(Math.abs(modeStartX - x) * 0.5) * (y < modeStartY ? -1 : 1)) + (y < modeStartY ? ((modeStartX % 2) * Math.abs((x % 2) - 1)) : -(x % 2) + Math.abs((modeStartX % 2) * (x % 2))); }
			}
			else {
				if(slope <= 1.0) { y = modeStartY; }
				else if(slope >= 3.0) { x = modeStartX; }
				else { x = modeStartX + (Math.round(Math.abs(modeStartY - y) * 0.5) * (x < modeStartX ? -1 : 1)) + (x < modeStartX ? ((modeStartY % 2) * Math.abs((y % 2) - 1)) : -(y % 2) + Math.abs((modeStartY % 2) * (y % 2))); }
			}
			ctx.beginPath();
			ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
			ctx.lineTo(getHexX(y, x), getHexY(y, x));
			ctx.closePath();
			ctx.strokeStyle = toolOutline;
			ctx.stroke();
		}
		else if(activeMode == MODE_RECT) {
			ctx.strokeStyle = toolOutline;
			ctx.rect(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX), getHexX(y, x) - getHexX(modeStartY, modeStartX), getHexY(y, x) - getHexY(modeStartY, modeStartX));
			ctx.stroke();
		}
		else if(activeMode == MODE_AREA) {
			ctx.fillStyle = toolOutline;
			ctx.fillRect(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX), getHexX(y, x) - getHexX(modeStartY, modeStartX), getHexY(y, x) - getHexY(modeStartY, modeStartX));
		}
		else if(activeMode == MODE_SKIP) {
			if(mapGrain == GRAIN_VERTICAL) {
				ctx.beginPath();
				ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
				ctx.lineTo(getHexX(y, x), getHexY(modeStartY, modeStartX));
				ctx.closePath();
				ctx.strokeStyle = toolOutline;
				ctx.stroke();
			}
			else {
				ctx.beginPath();
				ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
				ctx.lineTo(getHexX(modeStartY, modeStartX), getHexY(y, x));
				ctx.closePath();
				ctx.strokeStyle = toolOutline;
				ctx.stroke();
			}
		}
		else if(activeMode == MODE_SMART) {
			ctx.beginPath();
			ctx.moveTo(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX));
			ctx.lineTo(getHexX(y, x), getHexY(modeStartY, modeStartX));
			ctx.closePath();
			ctx.strokeStyle = toolOutline;
			ctx.stroke();
		}
		else if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) {
			if(Math.abs(modeStartX - x) % 2 > 0) {
				if(x < modeStartX) {
					if(x > 0) {
						x = x - 1;
					}
					else {
						x = x + 1;
					}
				}
				else {
					if(x < (gridx - 1)) {
						x = x + 1;
					}
					else {
						x = x - 1;
					}
				}
			}
			ctx.strokeStyle = toolOutline;
			ctx.rect(getHexX(modeStartY, modeStartX), getHexY(modeStartY, modeStartX), getHexX(y, x) - getHexX(modeStartY, modeStartX), getHexY(y, x) - getHexY(modeStartY, modeStartX));
			ctx.stroke();
		}
		else if(activeMode == MODE_CLONE) {
			if(dupeHex != null) {
				plotHex(ctx, x, y, true);
			}
		}
	}
}

function hideCursor(evt) {
	doubleBuffer();
}

function drawMap() {
	buf.clearRect(0, 0, buffer.width, buffer.height);
	plotHexes(buf);
	doubleBuffer();
}

function setPaintMode(pmode) {
	if(pmode != activeMode) {
		modeButton = document.getElementById("mode" + pmode);
		if(true == modeButton.disabled) {
			return;
		}
		document.getElementById("mode" + activeMode).className = CSS_PAL_BUTTON;
		activeMode = pmode;
		modeButton.className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_ACTIVE;
		if(pmode == MODE_LROOM || pmode == MODE_SROOM) {
			if(activeLayer == LAYER_BASE) {
				setBase(HEX_WHOLE);
			}
			else if(activeLayer == LAYER_OVERLAY) {
				setOverlay(HEX_WHOLE);
			}
			else {
				processPaletteButton(LAYER_BUTTONS[LAYER_BASE]);
				setBase(HEX_WHOLE);
			}
		}
		endModeDraw();
		if((activeMode == MODE_ERASE || activeMode == MODE_BULKE) && activeLayer == LAYER_ANNOTATE) { setAnnotation(ANNOTATION_BASE); }
	}
	dupeHex = null;
}

function resetMode(mode) {
	if(
		activeMode == MODE_LROOM ||
		activeMode == MODE_SROOM ||
		activeMode == MODE_ERASE ||
		activeMode == MODE_BULKE ||
		activeMode == MODE_CLONE
	) {
		setPaintMode(mode);
	}
}

function getBaseGlyph(currGlyph) {
	return (currGlyph < 10 ? currGlyph : (currGlyph - (currGlyph % 10)));
}

function setMaterial(typ, isCustom) {
	if(typ != activeMaterial) {
		document.getElementById("material" + activeMaterial).className = CSS_PAL_BUTTON + (isCustom ? ' ' + CSS_PAL_BUTTON_CUSTOM : '');
		activeMaterial = typ;
		document.getElementById("material" + typ).className = CSS_PAL_BUTTON + (isCustom ? ' ' + CSS_PAL_BUTTON_CUSTOM : '') + ' ' + CSS_PAL_BUTTON_ACTIVE;
		updatePreview();
	}
}

function setBase(typ) {
	if(typ != getBaseGlyph(activeBase)) {
		document.getElementById("base" + getBaseGlyph(activeBase)).className = CSS_PAL_BUTTON;
		activeBase = typ;
		document.getElementById("base" + typ).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_ACTIVE;
		updatePreview();
		resetMode(MODE_PAINT);
	}
}

function setOverlay(typ) {
	if(typ != getBaseGlyph(activeOverlay)) {
		document.getElementById("overlay" + getBaseGlyph(activeOverlay)).className = CSS_PAL_BUTTON;
		activeOverlay = typ;
		document.getElementById("overlay" + typ).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_ACTIVE;
		updatePreview();
		resetMode(MODE_PAINT);
	}
}

function setTerrain(typ) {
	if(typ != activeTerrain) {
		document.getElementById("terrain" + activeTerrain).className = CSS_PAL_BUTTON;
		activeTerrain = typ;
		document.getElementById("terrain" + typ).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_ACTIVE;
		updatePreview();
		resetMode(MODE_PAINT);
	}
}

function setFeature(typ) {
	if(typ != getBaseGlyph(activeFeature)) {
		document.getElementById("feature" + getBaseGlyph(activeFeature)).className = CSS_PAL_BUTTON;
		activeFeature = typ;
		document.getElementById("feature" + typ).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_ACTIVE;
		updatePreview();
		resetMode(MODE_PAINT);
	}
}

function setDecor(typ) {
	if(typ != getBaseGlyph(activeDecor)) {
		document.getElementById("decor" + getBaseGlyph(activeDecor)).className = CSS_PAL_BUTTON;
		activeDecor = typ;
		document.getElementById("decor" + typ).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_ACTIVE;
		updatePreview();
		resetMode(MODE_PAINT);
	}
}

function setSymbol(typ) {
	if(typ != getBaseGlyph(activeSymbol)) {
		document.getElementById("symbol" + getBaseGlyph(activeSymbol)).className = CSS_PAL_BUTTON;
		activeSymbol = typ;
		document.getElementById("symbol" + typ).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_ACTIVE;
		updatePreview();
		resetMode(MODE_PAINT);
	}
}

function updatePreview() {
	scratchTemp = outlineThickness;
	outlineThickness = DEFAULT_LINE_THICKNESS;
	var hexraw = getHexRawCoords(PREVIEW_SCALE, PREVIEW_SIZEX, PREVIEW_SIZEY);
	var hexcor = getHexCoords(PREVIEW_SCALE);
	var hexptx = hexraw[0];
	var hexpty = hexraw[1];
	prv.clearRect(0, 0, PREVIEW_SIZEX * 2, PREVIEW_SIZEY * 2);
	if(activeLayer == LAYER_BASE) {
		drawBase(prv, hexptx[0], hexpty[0], hexcor[0], hexcor[1], activeBase, materialDef[activeMaterial]);
	}
	else if(activeLayer == LAYER_TERRAIN) {
		drawTerrain(prv, hexraw[0], hexraw[1], PREVIEW_SCALE, activeTerrain, materialDef[activeMaterial]);
	}
	else if(activeLayer == LAYER_OVERLAY) {
		drawOverlay(prv, hexptx[0], hexpty[0], hexcor[0], hexcor[1], activeOverlay, materialDef[activeMaterial]);
	}
	else if(activeLayer == LAYER_FEATURE) {
		drawFeature(prv, hexraw[0], hexraw[1], PREVIEW_SCALE, activeFeature, materialDef[activeMaterial]);
	}
	else if(activeLayer == LAYER_DECOR) {
		drawDecor(prv, hexraw[0], hexraw[1], PREVIEW_SCALE, activeDecor, materialDef[activeMaterial]);
	}
	else if(activeLayer == LAYER_SYMBOL) {
		drawSymbol(prv, hexraw[0], hexraw[1], PREVIEW_SCALE, activeSymbol, materialDef[activeMaterial]);
	}
	drawHexOutline(prv, hexraw[0], hexraw[1], (activeTheme == THEME_LIGHT ? '#404040' : '#c0c0c0'));
	outlineThickness = scratchTemp;
}

function calcRotation(elem, clockwise) {
	var rotatedElement = elem;
	var baseElement = getBaseGlyph(elem);
	if(elem >= 10) {
		if(typeof clockwise != "undefined") {} else { clockwise = true; }
		var currRotation = elem % 10;
		if(clockwise) {
			currRotation++;
		}
		else {
			currRotation--;
		}
		while(currRotation < 0) {
			currRotation += 6;
		}
		while(currRotation >= 6) {
			currRotation -= 6;
		}
		rotatedElement = baseElement + currRotation;
	}
	return rotatedElement;
}

function rotateElement(clockwise) {
	if(activeLayer == LAYER_BASE) {
		activeBase = calcRotation(activeBase, clockwise);
	}
	else if(activeLayer == LAYER_OVERLAY) {
		activeOverlay = calcRotation(activeOverlay, clockwise);
	}
	else if(activeLayer == LAYER_FEATURE) {
		activeFeature = calcRotation(activeFeature, clockwise);
	}
	else if(activeLayer == LAYER_DECOR) {
		activeDecor = calcRotation(activeDecor, clockwise);
	}
	else if(activeLayer == LAYER_SYMBOL) {
		activeSymbol = calcRotation(activeSymbol, clockwise);
	}
	else {
		return;
	}
	updatePreview();
}

function calculateFlip(elem) {
	var flippedElement = elem;
	var baseElement = getBaseGlyph(elem);
	if(elem >= 10) {
		var currRotation = elem % 10;
		currRotation = currRotation + 3;
		while(currRotation < 0) {
			currRotation += 6;
		}
		while(currRotation >= 6) {
			currRotation -= 6;
		}
		flippedElement = baseElement + currRotation;
	}
	return flippedElement;
}

function flipElement() {
	if(activeLayer == LAYER_BASE) {
		activeBase = calculateFlip(activeBase);
	}
	else if(activeLayer == LAYER_OVERLAY) {
		activeOverlay = calculateFlip(activeOverlay);
	}
	else if(activeLayer == LAYER_FEATURE) {
		activeFeature = calculateFlip(activeFeature);
	}
	else if(activeLayer == LAYER_DECOR) {
		activeDecor = calculateFlip(activeDecor);
	}
	else if(activeLayer == LAYER_SYMBOL) {
		activeSymbol = calculateFlip(activeSymbol);
	}
	else {
		return;
	}
	updatePreview();
}

function setHex(col, row) {
	if(activeLayer == LAYER_MEGAHEX) { return; }
	if(col >= 0 && col < hexarray.length && row >= 0 && row < hexarray[0].length) {
		if(activeMode == MODE_CLONE && dupeHex != null) {
			hexarray[col][row] = dupeHex.cloneAt(col, row);
		}
		else {
			if(activeLayer == LAYER_BASE) {
				if(activeBase == HEX_EMPTY || activeMaterial == MATERIAL_EMPTY) {
					hexarray[col][row].clearBases();
				}
				else {
					hexarray[col][row].addBase(activeBase, activeMaterial);
				}
			}
			if(activeLayer == LAYER_TERRAIN) {
				if(activeTerrain == TERRAIN_EMPTY || activeMaterial == MATERIAL_EMPTY) {
					hexarray[col][row].clearTerrains();
				}
				else {
					hexarray[col][row].addTerrain(activeTerrain, activeMaterial);
				}
			}
			if(activeLayer == LAYER_OVERLAY) {
				if(activeOverlay == HEX_EMPTY || activeMaterial == MATERIAL_EMPTY) {
					hexarray[col][row].clearOverlays();
				}
				else {
					hexarray[col][row].addOverlay(activeOverlay, activeMaterial);
				}
			}
			if(activeLayer == LAYER_FEATURE) {
				if(activeFeature == FEATURE_EMPTY || activeMaterial == MATERIAL_EMPTY) {
					hexarray[col][row].clearFeatures();
				}
				else {
					hexarray[col][row].addFeature(activeFeature, activeMaterial);
				}
			}
			if(activeLayer == LAYER_DECOR) {
				if(activeDecor == DECOR_EMPTY || activeMaterial == MATERIAL_EMPTY) {
					hexarray[col][row].clearDecors();
				}
				else {
					hexarray[col][row].addDecor(activeDecor, activeMaterial);
				}
			}
			if(activeLayer == LAYER_SYMBOL) {
				if(activeSymbol == SYMBOL_EMPTY || activeMaterial == MATERIAL_EMPTY) {
					hexarray[col][row].clearSymbols();
				}
				else {
					hexarray[col][row].addSymbol(activeSymbol, activeMaterial);
				}
			}
			if(activeLayer == LAYER_ANNOTATE) {
				if(activeAnnotation == ANNOTATION_BASE || activeMaterial == MATERIAL_EMPTY) {
					hexarray[col][row].clearAnnotations();
				}
				else {
					hexarray[col][row].addAnnotation(activeAnnotation, activeMaterial);
				}
			}
		}
		plotHex(buf, col, row, false);
	}
}

function eraseHex(col, row, isBulkErase) {
	if(col >= 0 && col < hexarray.length && row >= 0  && row < hexarray[0].length) {
		if(isBulkErase) {
			hexarray[col][row].clearAll();
		}
		else {
			if(activeLayer == LAYER_MEGAHEX) { return; }
			if(activeLayer == LAYER_BASE) {
				hexarray[col][row].clearBases();
			}
			if(activeLayer == LAYER_TERRAIN) {
				hexarray[col][row].clearTerrains();
			}
			if(activeLayer == LAYER_OVERLAY) {
				hexarray[col][row].clearOverlays();
			}
			if(activeLayer == LAYER_FEATURE) {
				hexarray[col][row].clearFeatures();
			}
			if(activeLayer == LAYER_DECOR) {
				hexarray[col][row].clearDecors();
			}
			if(activeLayer == LAYER_SYMBOL) {
				hexarray[col][row].clearSymbols();
			}
			if(activeLayer == LAYER_ANNOTATE) {
				hexarray[col][row].clearAnnotations();
			}
		}
		plotHex(buf, col, row, false);
	}
}

function assignHexComponent(col, row, component, rotation) {
	var currLayer = activeLayer;
	if(activeLayer == LAYER_BASE) {
		var currBase = activeBase;
		if(component > HEX_WHOLE) {
			activeBase = component + rotation;
		}
		else {
			activeBase = component;
		}
		setHex(col, row);
		activeBase = currBase;
	}
	else {
		activeLayer = LAYER_OVERLAY;
		var currOverlay = activeOverlay;
		if(component > HEX_WHOLE) {
			activeOverlay = component + rotation;
		}
		else {
			activeOverlay = component;
		}
		setHex(col, row);
		activeOverlay = currOverlay;
	}
	activeLayer = currLayer;
}

function processMouseMove(evt) {
	if(activeLayer == LAYER_MEGAHEX) { return; }
	var mh = getMouseHex(evt.clientX, evt.clientY);
	var col = mh.x;
	var row = mh.y;
	if(col >= 0 && col < hexarray.length && row >= 0  && row < hexarray[0].length) {
		if(col != lastHexCol || row != lastHexRow) {
			lastHexCol = col;
			lastHexRow = row;
			document.getElementById('mouse-tracker').innerHTML = '<span>' + col + ' / ' + row + '</span>';
			if(paintOn) {
				if(activeMode == MODE_ERASE || activeMode == MODE_BULKE) {
					eraseHex(col, row, (activeMode == MODE_BULKE));
				}
				else {
					setHex(col, row);
				}
			}
			doubleBuffer();
			plotTool(cnv, col, row);
		}
	}
}

function togglePaintMode(evt, toggleval) {
	paintOn = toggleval;
	if(paintOn) {
		processMousePaint(evt);
	}
}

function processMousePaint(evt) {
	if(activeLayer == LAYER_MEGAHEX) { return; }
	var mh = getMouseHex(evt.clientX, evt.clientY);
	var col = mh.x;
	var row = mh.y;
	if(col >= 0 && col < hexarray.length && row >= 0  && row < hexarray[0].length) {
		storeUndo();
		if(activeMode != MODE_PAINT) {
			if(activeMode == MODE_ERASE || activeMode == MODE_BULKE) {
				eraseHex(col, row, (activeMode == MODE_BULKE));
			}
			else if(activeMode == MODE_CLONE) {
				if(dupeHex != null) {
					setHex(col, row);
					doubleBuffer();
				}
				else {
					dupeHex = hexarray[col][row].clone();
				}
			}
			else {
				if(modeStartX < 0 || modeStartY < 0) {
					paintOn = false;
					modeStartX = col;
					modeStartY = row;
				}
				else {
					paintOn = true;
					if(activeMode == MODE_LINE) {
						drawLine(modeStartX, modeStartY, col, row);
					}
					else if(activeMode == MODE_RECT) {
						drawRect(modeStartX, modeStartY, col, row);
					}
					else if(activeMode == MODE_AREA) {
						fillRect(modeStartX, modeStartY, col, row);
					}
					else if(activeMode == MODE_SKIP) {
						if(mapGrain == GRAIN_VERTICAL) {
							drawSkipLine(modeStartX, modeStartY, col, modeStartY);
						}
						else {
							drawSkipLine(modeStartX, modeStartY, modeStartX, row);
						}
					}
					else if(activeMode == MODE_LROOM || activeMode == MODE_SROOM) {
						if(Math.abs(modeStartX - col) % 2 > 0) {
							if(col < modeStartX) {
								if(col > 0) {
									col = col - 1;
								}
								else {
									col = col + 1;
								}
							}
							else {
								if(col < (gridx - 1)) {
									col = col + 1;
								}
								else {
									col = col - 1;
								}
							}
						}
						if((Math.abs(modeStartX - col) > 0) && (Math.abs(modeStartY - row) > 0)) {
							drawRoom(modeStartX, modeStartY, col, row, (activeMode == MODE_SROOM));
						}
					}
					modeStartX = -1;
					modeStartY = -1;
					doubleBuffer();
				}
			}
		}
		else {
			setHex(col, row);
			doubleBuffer();
		}
	}
}

function toggleLayer(lyr) {
	layerVisibility[lyr] = !layerVisibility[lyr];
	document.getElementById(LAYER_TOGGLE_PREFIX + lyr).className = CSS_LYR_BUTTON + (layerVisibility[lyr] ? ' ' + CSS_LYR_BUTTON_ACTIVE : '');
	drawMap();
}

function storeUndo() {
	hexundo.length = 0;
	for(var x = 0; x < hexarray.length; x++) {
		hexundo[x] = [];
		for(var y = 0; y < hexarray[x].length; y++) {
			hexundo[x].push(hexarray[x][y].clone());
		}
	}
	haveUndo = true;
}

function recallUndo() {
	hexarray.length = 0;
	for(var x = 0; x < hexundo.length; x++) {
		hexarray[x] = [];
		for(var y = 0; y < hexundo[x].length; y++) {
			hexarray[x].push(hexundo[x][y].clone());
		}
	}
	haveUndo = false;
}

function undo() {
	if(haveUndo) {
		recallUndo();
		drawMap();
	}
}

function recalculate() {
	drawscale = drawsize;
	halfscale = drawscale / 2;
	xoffset = drawscale;
	yoffset = (drawscale * CALCTHIN);
	outlineThickness = getOutlineThickness();
}

function resizeCanvas() {
	var newx = (mapGrain == GRAIN_VERTICAL
					? Math.ceil(((drawsize * 1.5) * gridx) + (leftmargin * 2) + (drawsize / 2))
					: Math.ceil(((drawsize * CALCSQRT) * gridx) + (leftmargin * 2) + (drawsize / 2) + (drawsize / 4))
				);
	var newy = (mapGrain == GRAIN_VERTICAL
					? Math.ceil(((drawsize * CALCSQRT) * gridy) + (topmargin * 2) + (drawsize / 2) + (drawsize / 4))
					: Math.ceil(((drawsize * 1.5) * gridy) + (topmargin * 2) + (drawsize / 2))
				);
	canvas.width = newx;
	canvas.height = newy;
	cnv = canvas.getContext('2d');
	buffer.width = newx;
	buffer.height = newy;
	buf = buffer.getContext('2d');
	buf.clearRect(0, 0, buffer.width, buffer.height);
	mhexes.width = newx;
	mhexes.height = newy;
	mhx = mhexes.getContext('2d');
	mhx.clearRect(0, 0, mhexes.width, mhexes.height);
	if(megahexesOn) { imposeMegahexes(); }
	doubleBuffer();
}

function zoomslide(src) {
	drawsize = parseInt(src.value);
	if(src.id == 'zoomlevel') {
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		if(document.getElementById('prefZoomlevel') != null && document.getElementById('prefZoomlevel') != 'undefined') {
			document.getElementById('prefZoomlevel').value = drawsize;
			document.getElementById('showZoomlevelPref').innerHTML = drawsize + 'px';
		}
	}
	else if(src.id == 'prefZoomlevel') {
		document.getElementById('zoomlevel').value = drawsize;
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		document.getElementById('showZoomlevelPref').innerHTML = drawsize + 'px';
	}
	recalculate();
	resizeCanvas();
	drawMap();
}

function zoomecho(src) {
	drawsize = parseInt(document.getElementById('zoomlevel').value);
	document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
}

function translide() {
	TRANSLUCE = parseFloat(document.getElementById('translevel').value);
	document.getElementById('showOpacity').innerHTML = TRANSLUCE;
	doubleBuffer();
}

function transecho() {
	TRANSLUCE = parseFloat(document.getElementById('translevel').value);
	document.getElementById('showOpacity').innerHTML = TRANSLUCE;
}

function setTranslucence(transval) {
	if(transval != null && transval >= 0.0 && transval <= 1.0) {
		TRANSLUCE = transval;
		document.getElementById('translevel').value = TRANSLUCE;
		document.getElementById('showOpacity').innerHTML = TRANSLUCE;
	}
}

function zoom(increase) {
	var changed = false;
	if(increase && drawsize < ZOOM_MAX) {
		drawsize += ZOOM_STEP;
		changed = true;
	}
	if(!increase && drawsize > ZOOM_MIN) {
		drawsize -= ZOOM_STEP;
		changed = true;
	}
	if(changed) {
		document.getElementById('zoomlevel').value = drawsize;
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		recalculate();
		resizeCanvas();
		drawMap();
	}
}

function updateOutlineColor(src) {
	if(src != null && src.value != "undefined") {
		var greyscale = parseInt(src.value);
		document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
		defaultGridGrey = greyscale;
		defaultGridColor = "rgb(" + defaultGridGrey + "," + defaultGridGrey + "," + defaultGridGrey + ")";
		hexOutline = defaultGridColor;
		updatePreview();
		drawMap();
	}
}

function outlineColorEcho() {
	var greyscale = parseInt(document.getElementById('hexshade').value);
	document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
}

function mhexWeightEcho() {
	document.getElementById('showMhexThick').innerHTML = megahexWeight + 'px';
}

function mhexslide(src) {
	megahexWeight = parseInt(src.value);
	mhexWeightEcho();
	imposeMegahexes();
}

function mhexWeightUpdate(refresh) {
	var mhsliderelem = document.getElementById("mhexThick");
	mhsliderelem.value = megahexWeight;
	mhexWeightEcho();
	if(refresh) { imposeMegahexes(); }
}

function mdotWeightEcho() {
	document.getElementById('showDotWeight').innerHTML = megahexDotWeight + 'px';
}

function mdotslide(src) {
	megahexDotWeight = parseInt(src.value);
	mdotWeightEcho();
	imposeMegahexes();
}

function mdotWeightUpdate(refresh) {
	var mdsliderelem = document.getElementById("mhexDotWt");
	mdsliderelem.value = megahexDotWeight;
	mdotWeightEcho();
	if(refresh) { imposeMegahexes(); }
}

function editMaterial(materialId) {
	if(materialId != null && materialId >= CUSTOM_PALETTE_OFFSET) {
		materialToEdit = parseInt(materialId);
		document.getElementById('material-editor-picker').value = materialDef[materialId];
		document.getElementById('material-editor-name').value = materialNames[materialId];
		document.getElementById('material-editor-label').innerHTML = document.getElementById('material-editor-picker').value;
		showMaterialEditor();
	}
}

function updateMaterial(cpicker) {
	document.getElementById('material-editor-label').innerHTML = document.getElementById('material-editor-picker').value;
}

function loadMaterial(materialIndex, materialName, materialColor) {
	if(materialIndex != null && materialIndex >= CUSTOM_PALETTE_OFFSET) {
		materialNames[materialIndex] = materialName;
		materialDef[materialIndex] = materialColor;
		var materialButtonSvg = document.getElementById('svgmaterial' + materialIndex);
		materialButtonSvg.style.fill = materialColor;
		var materialButtonLabel = document.getElementById('lblmaterial' + materialIndex);
		materialButtonLabel.innerHTML = materialName;
	}
}

function postMaterial() {
	var materialIndex = materialToEdit;
	var materialName = document.getElementById('material-editor-name').value;
	var materialColor = document.getElementById('material-editor-picker').value;
	document.getElementById('customColorMessage').innerHTML = '';
	if(checkSimpleString(materialName) && materialColor != 'undefined' && materialColor != null) {
		materialNames[materialIndex] = materialName;
		materialDef[materialIndex] = materialColor;
		var materialButtonSvg = document.getElementById('svgmaterial' + materialIndex);
		materialButtonSvg.style.fill = materialColor;
		var materialButtonLabel = document.getElementById('lblmaterial' + materialIndex);
		materialButtonLabel.innerHTML = materialName;
		hideMaterialEditor();
		drawMap();
	}
	else {
		var errorMsg = '';
		if(materialColor == 'undefined' || materialColor == null) {
			errorMsg = errorMsg + 'Color is invalid.';
		}
		else if(materialName.length <= 0) {
			errorMsg = errorMsg + 'Name cannot be blank.';
		}
		else {
			errorMsg = errorMsg + 'Name cannot contain extended characters.';
		}
		if(errorMsg != '') {
			document.getElementById('customColorMessage').innerHTML = errorMsg;
		}
	}
}

function getSlope(startX, startY, endX, endY) {
	var slope = ((endY - startY) * 1.0) / ((endX - startX) * 1.0);
	if(slope < 0.0) { slope = slope * -1.0; }
	return slope;
}

function drawLine(startX, startY, endX, endY) {
	var slope = getSlope(startX, startY, endX, endY);
	if(mapGrain == GRAIN_VERTICAL) {
		if(slope <= 0.25) { endY = startY; }
		else if(slope >= 0.75) { endX = startX; }
	}
	else {
		if(slope <= 1.0) { endY = startY; }
		else if(slope >= 3.0) { endX = startX; }
	}
	var linelen = Math.max(Math.abs(startX - endX), Math.abs(startY - endY)) + 1;
	var linedir = 0;
	var lerror  = 0.0;
	if(linelen == 1) {
		setHex(startX, startY);
	}
	else if(startX == endX) {
		// vertical line
		linedir = (startY > endY ? -1 : 1);
		for(var i = startY; i != endY; i += linedir) {
			setHex(endX, i);
		}
		setHex(endX, endY);
	}
	else if(startY == endY) {
		// horizontal line
		linedir = (startX > endX ? -1 : 1);
		for(var i = startX; i != endX; i += linedir) {
			setHex(i, endY);
		}
		setHex(endX, endY);
	}
	else {
		// perfect diagonal
		if(mapGrain == GRAIN_VERTICAL) {
			endY = startY + (Math.round(Math.abs(startX - endX) * 0.5) * (endY < startY ? -1 : 1)) + (endY < startY ? ((startX % 2) * Math.abs((endX % 2) - 1)) : -(endX % 2) + Math.abs((startX % 2) * (endX % 2)));
		}
		else {
			endX = startX + (Math.round(Math.abs(startY - endY) * 0.5) * (endX < startX ? -1 : 1)) + (endX < startX ? ((startY % 2) * Math.abs((endY % 2) - 1)) : -(endY % 2) + Math.abs((startY % 2) * (endY % 2)));
		}
		if(startY > endY) {
			var tmpX = startX;
			var tmpY = startY;
			startX = endX;
			startY = endY;
			endX = tmpX;
			endY = tmpY;
		}
		linedir      = (startY > endY ? -1 : 1);
		var linedir2 = (startX > endX ? -1 : 1);
		var px       = startX;
		var py       = startY;
		var ullr     = (Math.sign(linedir) == Math.sign(linedir2) ? 0 : 1);
		for(var i = 0; i < linelen; i++) {
			setHex(px, py);
			if(mapGrain == GRAIN_VERTICAL) {
				px += linedir2;
				if(px % 2 == 0) {
					py += linedir;
				}
			}
			else {
				py += linedir;
				if(py % 2 == ullr) {
					px += linedir2;
				}
			}
		}
	}
}

function drawRect(upperX, upperY, lowerX, lowerY, val) {
	if(upperX > lowerX) {
		var tmp = upperX;
		upperX = lowerX;
		lowerX = tmp;
	}
	if(upperY > lowerY) {
		var tmp = upperY;
		upperY = lowerY;
		lowerY = tmp;
	}
	for(var x = upperX; x <= lowerX; x++) {
		setHex(x, upperY);
		setHex(x, lowerY);
	}
	for(var y = upperY; y <= lowerY; y++) {
		setHex(upperX, y);
		setHex(lowerX, y);
	}
}

function fillRect(upperX, upperY, lowerX, lowerY, val) {
	if(upperX > lowerX) {
		var tmp = upperX;
		upperX = lowerX;
		lowerX = tmp;
	}
	if(upperY > lowerY) {
		var tmp = upperY;
		upperY = lowerY;
		lowerY = tmp;
	}
	for(var x = upperX; x <= lowerX; x++) {
		for(var y = upperY; y <= lowerY; y++) {
			setHex(x, y);
		}
	}
}

function drawSkipLine(startX, startY, endX, endY) {
	var linelen = Math.max(Math.abs(startX - endX), Math.abs(startY - endY)) + 1;
	if(linelen == 1) {
		setHex(startX, startY);
	}
	else {
		if(mapGrain == GRAIN_VERTICAL) {
			if(startX > endX) {
				for(x = startX; x >= endX; x = x - 2) {
					setHex(x, startY);
				}
			}
			else {
				for(x = startX; x <= endX; x = x + 2) {
					setHex(x, startY);
				}
			}
		}
		else {
			if(startY > endY) {
				for(y = startY; y >= endY; y = y - 2) {
					setHex(startX, y);
				}
			}
			else {
				for(y = startY; y <= endY; y = y + 2) {
					setHex(startX, y);
				}
			}
		}
	}
}

function drawRoom(upperX, upperY, lowerX, lowerY, isSolid) {
	if(upperX > lowerX) {
		var tmp = upperX;
		upperX = lowerX;
		lowerX = tmp;
	}
	if(upperY > lowerY) {
		var tmp = upperY;
		upperY = lowerY;
		lowerY = tmp;
	}
	if(isSolid) {
		assignHexComponent(upperX, upperY, HEX_SRCNR, 2);
		assignHexComponent(lowerX, upperY, HEX_SRCNR, 3);
		assignHexComponent(upperX, lowerY, HEX_SRCNR, 0);
		assignHexComponent(lowerX, lowerY, HEX_SRCNR, 5);
	}
	else {
		assignHexComponent(upperX, upperY, HEX_LRCNR, 2);
		assignHexComponent(lowerX, upperY, HEX_LRCNR, 3);
		assignHexComponent(upperX, lowerY, HEX_LRCNR, 0);
		assignHexComponent(lowerX, lowerY, HEX_LRCNR, 5);
	}
	var borderYTop = upperY;
	var borderYBottom = lowerY - 1;
	if(upperX % 2 == 1) {
		borderYTop = upperY + 1;
		borderYBottom = lowerY;
	}
	var isBorderPhase = true;
	var rBorderComp = (isSolid ? HEX_WHOLE : HEX_BORDR);
	var rWallComp = (isSolid ? HEX_FHALF : HEX_LRWAL);
	var rSideComp = (isSolid ? HEX_SRSID : HEX_LRSID);
	for(var x = upperX + 1; x <= lowerX - 1; x++) {
		if(isBorderPhase) {
			assignHexComponent(x, borderYTop, rBorderComp, 0);
			assignHexComponent(x, borderYBottom, rBorderComp, 3);
		}
		else {
			assignHexComponent(x, upperY, rWallComp, 3);
			assignHexComponent(x, lowerY, rWallComp, 0);
		}
		isBorderPhase = !isBorderPhase;
	}
	for(var y = upperY + 1; y <= lowerY - 1; y++) {
		assignHexComponent(upperX, y, rSideComp, 0);
		assignHexComponent(lowerX, y, rSideComp, 3);
	}
	if(isSolid) {
		for(var y = upperY + 1; y <= lowerY - 1; y++) {
			for(var x = upperX + 1; x <= lowerX - 1; x++) {
				assignHexComponent(x, y, HEX_WHOLE, 0);
			}
		}
	}
}

function endModeDraw() {
	modeStartX = -1;
	modeStartY = -1;
}

function newMap() {
	if(confirm("Start new map? Any unsaved work will be lost.")) {
		initialise();
		resetBackdrop();
	}
}

function clearMap() {
	if(confirm("Clear the map?")) {
		for(x = 0; x < hexarray.length; x++) {
			for(y = 0; y < hexarray[x].length; y++) {
				hexarray[x][y].clearAll();
			}
		}
		drawMap();
	}
}

function reinitialiseGrid() {
	var hexprev = hexarray.slice();
	hexarray = [];
	for(x = 0; x < gridx; x++) {
		hexarray[x] = [];
		for(y = 0; y < gridy; y++) {
			if(x < hexprev.length && y < hexprev[0].length) {
				hexarray[x][y] = hexprev[x][y];
			}
			else {
				var hex = new Hex(x, y);
				hex.addBase(HEX_EMPTY, MATERIAL_EMPTY);
				hexarray[x][y] = hex;
			}
		}
	}
	hexundo = hexarray.slice();
}

function resizeGrid(rows, cols, isExisting) {
	gridx = cols;
	gridy = rows;
	if(isExisting) {
		reinitialiseGrid();
	}
	else {
		initialiseGrid();
	}
	resizeCanvas();
}

function toggleRecalc(enable) {
	document.getElementById("btnupdategrid").disabled = !enable;
}

function recalcMap() {
	var newrows = parseInt(document.getElementById("maprows").value);
	var newcols = parseInt(document.getElementById("mapcols").value);
	if(isNaN(newrows) || isNaN(newcols) || newrows <= 0 || newcols <= 0 || newrows > MAX_GRID || newcols > MAX_GRID) {
		document.getElementById("maprows").value = gridy;
		document.getElementById("mapcols").value = gridx;
		toggleRecalc(false);
		window.alert("Row and Column values must be between 1 and " + MAX_GRID);
		return;
	}
	if((newcols >= gridx && newrows >= gridy) || confirm("Reducing the grid size will trim the map to fit.\r\nDo you wish to proceed?")) {
		resizeGrid(newrows, newcols, true);
		drawMap();
		toggleRecalc(false);
	}
}

function shiftMapElement(ocol, orow, ncol, nrow) {
	hexarray[ncol][nrow] = hexarray[ocol][orow].clone();
	hexarray[ocol][orow].clearAll();
}

function shiftMap(shiftDir) {
	var xrange = hexarray.length;
	var yrange = hexarray[0].length;
	var shiftStep = 1;
	if(
		(mapGrain == GRAIN_VERTICAL && (shiftDir == SHIFTMAP_LEFT || shiftDir == SHIFTMAP_RIGHT || shiftDir == SHIFTMAP_RIGHTK))
		||
		(mapGrain == GRAIN_HORIZONTAL && (shiftDir == SHIFTMAP_UP || shiftDir == SHIFTMAP_DOWN || shiftDir == SHIFTMAP_DOWNK))
	) {
		shiftStep = 2;
	}
	if(shiftDir == SHIFTMAP_RIGHT) { xrange = xrange - shiftStep; }
	if(shiftDir == SHIFTMAP_DOWN) { yrange = yrange - shiftStep; }
	var addedHexes = false;
	if(shiftDir == SHIFTMAP_RIGHTK) {
		// add columns to the right
		gridx = hexarray.length + shiftStep;
		gridy = hexarray[0].length;
		addedHexes = true;
	}
	if(shiftDir == SHIFTMAP_DOWNK) {
		// add rows to the bottom
		gridx = hexarray.length;
		gridy = hexarray[0].length + shiftStep;
		addedHexes = true;
	}
	if(addedHexes) {
		resizeGrid(gridy, gridx, true);
		resizeCanvas();
		drawMap();
		document.getElementById("maprows").value = gridy;
		document.getElementById("mapcols").value = gridx;
	}
	if(shiftDir == SHIFTMAP_LEFT) {
		// shift map left
		for(x = shiftStep; x < xrange; x++) {
			for(y = 0; y < yrange; y++) {
				shiftMapElement(x, y, x - shiftStep, y);
			}
		}
	}
	else if(shiftDir == SHIFTMAP_RIGHT || shiftDir == SHIFTMAP_RIGHTK) {
		// shift map right
		for(x = xrange - 1; x >= 0; x--) {
			for(y = 0; y < yrange; y++) {
				shiftMapElement(x, y, x + shiftStep, y);
			}
		}
	}
	else if(shiftDir == SHIFTMAP_UP) {
		// shift map up
		for(x = 0; x < xrange; x++) {
			for(y = 1; y < yrange; y++) {
				shiftMapElement(x, y, x, y - shiftStep);
			}
		}
	}
	else if(shiftDir == SHIFTMAP_DOWN || shiftDir == SHIFTMAP_DOWNK) {
		// shift map down
		for(x = 0; x < xrange; x++) {
			for(y = yrange - 1; y >= 0; y--) {
				shiftMapElement(x, y, x, y + shiftStep);
			}
		}
	}
	drawMap();
}

function changeBackground(newBackground) {
	hexBackground = newBackground;
	materialDef[0] = hexBackground;
	document.getElementById("canvas").style.background = newBackground;
	document.getElementById("buffer").style.background = newBackground;
}

function changeTemplate(backgroundColor, outlineColor, rows, cols, scaling) {
	if(confirm('This will erase all unsaved map changes!')) {
		hexOutline = outlineColor;
		document.getElementById('hexshade').value = hexOutline;
		changeBackground(backgroundColor);
		resizeGrid(rows, cols, false);
		document.getElementById("maprows").value = rows;
		document.getElementById("mapcols").value = cols;
		drawsize = scaling;
		document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
		recalculate();
		resizeCanvas();
		drawMap();
	}
}

function getMapName() {
	var mapname = document.getElementById('mapname').value;
	if(mapname.length < 1) { mapname = "default"; }
	return mapname;
}

function getMapAuthor() {
	var mapauthor = document.getElementById('mapauthor').value;
	if(mapauthor.length < 1) { mapauthor = "default"; }
	return mapauthor;
}

function toggleRoomDrawMode(isForceDisable) {
	if(mapGrain == GRAIN_VERTICAL && !isForceDisable) {
		document.getElementById("mode" + MODE_LROOM).disabled = false;
		document.getElementById("mode" + MODE_LROOM).className = CSS_PAL_BUTTON;
		document.getElementById("mode" + MODE_SROOM).disabled = false;
		document.getElementById("mode" + MODE_SROOM).className = CSS_PAL_BUTTON;
	}
	else {
		resetMode(MODE_PAINT);
		document.getElementById("mode" + MODE_LROOM).disabled = true;
		document.getElementById("mode" + MODE_LROOM).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_DISABLED;
		document.getElementById("mode" + MODE_SROOM).disabled = true;
		document.getElementById("mode" + MODE_SROOM).className = CSS_PAL_BUTTON + ' ' + CSS_PAL_BUTTON_DISABLED;
	}
}

function toggleGrain(selector) {
	mapGrain = selector.options[selector.selectedIndex].value;
	toggleRoomDrawMode(false);
	correctMegahexOffsetsForGrain();
	resizeCanvas();
	drawMap();
	updatePreview();
}

function changeLineWeight(selector) {
	lineWeight = selector.options[selector.selectedIndex].value;
	recalculate();
	drawMap();
}

function isEmptyHex(hex) {
	var isEmpty = true;
	if(hex.overlays.length > 0) { isEmpty = false; }
	if(hex.features.length > 0) { isEmpty = false; }
	if(hex.decors.length > 0) { isEmpty = false; }
	if(hex.terrains.length > 0) { isEmpty = false; }
	if(hex.symbols.length > 0) { isEmpty = false; }
	if(hex.annotations.length > 0) { isEmpty = false; }
	if(hex.bases.length > 1) {
		isEmpty = false;
	}
	else if(hex.bases.length == 1) {
		if(
			hex.bases[0].ctype != HEX_EMPTY ||
			hex.bases[0].material != MATERIAL_EMPTY
		) {
			isEmpty = false;
		}
	}
	return isEmpty;
}

function encodeBackdrop(img) {
	if(img != null) {
		return img.src;
	}
	else {
		return '';
	}
}

function changeBackdropLockState(newstate) {
	backdropLocked = (newstate == true || newstate == "true");
	document.getElementById("backdroplock").className = CSS_PAL_BUTTON + (backdropLocked ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
}

function toggleBackdropLockState() {
	changeBackdropLockState(!backdropLocked);
}

function warnBackdropLocked() {
	window.alert("Cannot modify backdrop while locked.\nToggle the backdrop lock to edit.");
}

function encodeLayerComponents(lyr) {
	var comps = [];
	for(var c = 0; c < lyr.length; c++) {
		var dataPair = {};
		var hcomp = lyr[c];
		if(hcomp.ctype < 10) {
			// LEGACY - remove this IF and IF ELSE after early SII maps are migrated, keep only the ELSE portion
			if(hcomp.ctype == 0 && (hcomp.ctype % 10 == 1)) {
				dataPair[DATAPAIR_CTYPE] = 1;
				dataPair[DATAPAIR_ROTAT] = 0;
			}
			else if(hcomp.ctype == 1) {
				dataPair[DATAPAIR_CTYPE] = 1;
				dataPair[DATAPAIR_ROTAT] = 0;
			}
			else {
				dataPair[DATAPAIR_CTYPE] = hcomp.ctype;
				dataPair[DATAPAIR_ROTAT] = 0;
			}
		}
		else {
			dataPair[DATAPAIR_CTYPE] = hcomp.ctype - (hcomp.ctype % 10);
			dataPair[DATAPAIR_ROTAT] = hcomp.ctype % 10;
		}
		dataPair[DATAPAIR_MATER] = hcomp.material;
		comps.push(dataPair);
	}
	return comps;
}

function encodeAnnotations(lyr) {
	var annots = [];
	for(var c = 0; c < lyr.length; c++) {
		var dataPair = {};
		var hcomp = lyr[c];
		dataPair[DATAPAIR_CTYPE] = hcomp.ctype;
		dataPair[DATAPAIR_MATER] = hcomp.material;
		annots.push(dataPair);
	}
	return annots;
}

function saveMapData() {
	/* increase file iteration */
	fileIteration = fileIteration + 1;

	/* build savefile */
	var jsonOutput = {};
	var mapName = getMapName();
	// map info
	jsonOutput[FILEKEYVERS] = VERSION_NUMBER;
	jsonOutput[FILEKEYCOLS] = hexarray.length;
	jsonOutput[FILEKEYROWS] = hexarray[0].length;
	jsonOutput[FILEKEYNAME] = mapName;
	jsonOutput[FILEKEYAUTH] = getMapAuthor();
	jsonOutput[FILEKEYITER] = fileIteration;
	jsonOutput[FILEKEYGRAN] = mapGrain;
	jsonOutput[FILEKEYLWGT] = lineWeight;
	jsonOutput[FILEKEYSHAD] = parseInt(document.getElementById('hexshade').value);
	jsonOutput[FILEKEYZOOM] = parseInt(document.getElementById('zoomlevel').value);
	jsonOutput[FILEKEYTRNO] = translucencyEnabled;
	jsonOutput[FILEKEYTRNA] = TRANSLUCE;
	if(megahexesOn) {
		jsonOutput[FILEKEYMHEX] = megahexesOn + SEP_LIST + mhexImposeX + SEP_LIST + mhexImposeY + SEP_LIST + mhexImposeS + SEP_LIST + mhexType;
	}
	jsonOutput[FILEKEYMHWT] = megahexWeight;
	jsonOutput[FILEKEYMHDT] = megahexDotWeight;
	// map data
	var hexArray = [];
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			var hexData = {};
			hexData[HKEY_X] = x;
			hexData[HKEY_Y] = y;
			var hex = hexarray[x][y];
			if(isEmptyHex(hex)) {
				// don't save empty hexes
			}
			else {
				hexData[HKEY_BASE] = encodeLayerComponents(hex.bases);
				hexData[HKEY_TERRAIN] = encodeLayerComponents(hex.terrains);
				hexData[HKEY_OVERLAY] = encodeLayerComponents(hex.overlays);
				hexData[HKEY_FEATURE] = encodeLayerComponents(hex.features);
				hexData[HKEY_DECOR] = encodeLayerComponents(hex.decors);
				hexData[HKEY_SYMBOL] = encodeLayerComponents(hex.symbols);
				hexData[HKEY_ANNOT] = encodeAnnotations(hex.annotations);
				hexArray.push(hexData);
			}
		}
	}
	jsonOutput[DATAHEX] = hexArray;
	// custom colors
	var customColors = {};
	var customColor = {};
	var customName = '';
	var customValue = '';
	for(var c = CUSTOM_PALETTE_OFFSET; c < materialDef.length; c++) {
		customColor = {};
		customColor[materialNames[c]] = materialDef[c];
		customColors[c] = customColor;
	}
	jsonOutput[CUSTOMCOLORS] = customColors;
	// backdrop
	if(backdropImage != null) {
		var backdropObject = {};
		backdropObject[BDROPDIMNW] = backdropImage.naturalWidth;
		backdropObject[BDROPDIMNH] = backdropImage.naturalHeight;
		backdropObject[BDROPDIMRW] = backdropImage.width;
		backdropObject[BDROPDIMRH] = backdropImage.height;
		backdropObject[BDROPOFFX] = backdropOffset.x;
		backdropObject[BDROPOFFY] = backdropOffset.y;
		backdropObject[BDROPDATA] = encodeBackdrop(backdropImage);
		jsonOutput[BDROPOBJ] = backdropObject;
	}

	/* pseudolink for named download */
    var saveStr = JSON.stringify(jsonOutput);
    var pseudoanchor = document.createElement("a");
    pseudoanchor.setAttribute('href', FILEPREAMBLE + encodeURIComponent(saveStr));
    pseudoanchor.setAttribute('download', mapName + MAPFILE_EXTENSION);
    pseudoanchor.click();
    hideOverlay(); 
}

function saveCustomColorData() {
	/* build savefile */
	var jsonOutput = {};
	var mapName = getMapName();
	// color collection info
	jsonOutput[FILEKEYVERS] = VERSION_NUMBER;
	jsonOutput[FILEKEYAUTH]  = getMapAuthor();
	// custom colors
	var customColors = {};
	var customColor = {};
	var customName = '';
	var customValue = '';
	for(var c = CUSTOM_PALETTE_OFFSET; c < materialDef.length; c++) {
		customColor = {};
		customColor[materialNames[c]] = materialDef[c];
		customColors[c] = customColor;
	}
	jsonOutput[CUSTOMCOLORS] = customColors;

	/* pseudolink for named download */
    var saveStr = JSON.stringify(jsonOutput);
    var pseudoanchor = document.createElement("a");
    pseudoanchor.setAttribute('href', FILEPREAMBLE + encodeURIComponent(saveStr));
    pseudoanchor.setAttribute('download', mapName + COLORFILE_EXTENSION);
    pseudoanchor.click();
    hideOverlay(); 
}

function legacyKey(filekey) {
	return filekey.toUpperCase() + ':';
}

function legacySafeMaterial(mater) {
	/* turns custom colors into black, since the custom colors are out-of-range in original Shamat */
	if(mater >= CUSTOM_PALETTE_OFFSET) {
		return 1;
	}
	return mater;
}

function saveLegacyMapData() {
	/* increase file iteration */
	fileIteration = fileIteration + 1;

	/* build savefile */
	var csvfile = LEGACYFILEPREAMBLE + LEGACYFILEHEADER + NEWLINE;
	csvfile += LEGACYFILEKEYVERS + VERSION_NUMBER + NEWLINE;
	csvfile += LEGACYFILEKEYSIZE + hexarray.length + SEP_LIST + hexarray[0].length + NEWLINE;
	var mapname = getMapName();
	csvfile += legacyKey(FILEKEYNAME) + mapname + NEWLINE;
	var mapauthor = getMapAuthor();
	csvfile += legacyKey(FILEKEYAUTH) + mapauthor + NEWLINE;
	csvfile += legacyKey(FILEKEYITER) + fileIteration + NEWLINE;
	csvfile += legacyKey(FILEKEYGRAN) + mapGrain + NEWLINE;
	csvfile += legacyKey(FILEKEYLWGT) + lineWeight + NEWLINE;
	csvfile += legacyKey(FILEKEYSHAD) + document.getElementById('hexshade').value + NEWLINE;
	csvfile += legacyKey(FILEKEYZOOM) + document.getElementById('zoomlevel').value + NEWLINE;
	csvfile += legacyKey(FILEKEYTRNO) + translucencyEnabled + NEWLINE;
	csvfile += legacyKey(FILEKEYTRNA) + TRANSLUCE + NEWLINE;
	if(megahexesOn) {
		csvfile += legacyKey(FILEKEYMHEX) + mhexOverlayX + SEP_LIST + mhexOverlayY + SEP_LIST + mhexOverlayS + SEP_LIST + mhexType + SEP_LIST + NEWLINE;
	}
	csvfile += legacyKey(FILEKEYMHWT) + megahexWeight + NEWLINE;
	csvfile += legacyKey(FILEKEYMHDT) + megahexDotWeight + NEWLINE;
	if(backdropImage != null) {
		csvfile += LEGACYBDROPHDR + NEWLINE;
		csvfile += legacyKey(LEGACYBDROPDIMNW) + backdropImage.naturalWidth + NEWLINE;
		csvfile += legacyKey(LEGACYBDROPDIMNH) + backdropImage.naturalHeight + NEWLINE;
		csvfile += legacyKey(LEGACYBDROPDIMRW) + backdropImage.width + NEWLINE;
		csvfile += legacyKey(LEGACYBDROPDIMRH) + backdropImage.height + NEWLINE;
		csvfile += legacyKey(LEGACYBDROPOFFX) + backdropOffset.x + NEWLINE;
		csvfile += legacyKey(LEGACYBDROPOFFY) + backdropOffset.y + NEWLINE;
		csvfile += legacyKey(LEGACYBDROPDATA) + encodeBackdrop(backdropImage) + NEWLINE;
	}
	csvfile += LEGACYDATAHEADER + NEWLINE;
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			var row = LEGACYDATAHEX + x + SEP_DATA + y;
			var hex = hexarray[x][y];
			row += ",[";
			for(var b = 0; b < hex.bases.length; b++) {
				var bcomp = hex.bases[b];
				row += (b > 0 ? SEP_LIST : "") + bcomp.ctype + SEP_PAIR + legacySafeMaterial(bcomp.material);
			}
			row += "]";
			row += SEP_DATA;
			row += "{";
			for(var o = 0; o < hex.overlays.length; o++) {
				var ocomp = hex.overlays[o];
				row += (o > 0 ? SEP_LIST : "") + ocomp.ctype + SEP_PAIR + legacySafeMaterial(ocomp.material);
			}
			row += "}";
			row += SEP_DATA;
			row += "(";
			var elems = 0;
			for(var f = 0; f < hex.terrains.length; f++) {
				var terr = hex.terrains[f];
				row += (elems > 0 ? SEP_LIST : "") + terr.ctype + SEP_PAIR + legacySafeMaterial(terr.material);
				elems += 1;
			}
			for(var f = 0; f < hex.features.length; f++) {
				var feat = hex.features[f];
				row += (elems > 0 ? SEP_LIST : "") + feat.ctype + SEP_PAIR + legacySafeMaterial(feat.material);
				elems += 1;
			}
			for(var f = 0; f < hex.decors.length; f++) {
				var deco = hex.decors[f];
				row += (elems > 0 ? SEP_LIST : "") + deco.ctype + SEP_PAIR + legacySafeMaterial(deco.material);
				elems += 1;
			}
			for(var f = 0; f < hex.symbols.length; f++) {
				var symb = hex.symbols[f];
				if(symb.ctype == SYMBOL_MARKER) {
					row += (elems > 0 ? SEP_LIST : "") + OLD_SYMBOL_MARKER + SEP_PAIR + legacySafeMaterial(symb.material);
				}
				else if(symb.ctype == SYMBOL_EVENT) {
					row += (elems > 0 ? SEP_LIST : "") + OLD_SYMBOL_EVENT + SEP_PAIR + legacySafeMaterial(symb.material);
				}
				else {
					row += (elems > 0 ? SEP_LIST : "") + symb.ctype + SEP_PAIR + legacySafeMaterial(symb.material);
				}
				elems += 1;
			}
			for(var a = 0; a < hex.annotations.length; a++) {
				var anno = hex.annotations[a];
				row += (elems > 0 ? SEP_LIST : "") + anno.ctype + SEP_PAIR + legacySafeMaterial(anno.material);
				elems += 1;
			}
			row += ")";
			csvfile += row + NEWLINE;
		}
	}

	/* pseudolink for named download */
	var csvUrl = encodeURI(csvfile);
    var pseudoanchor = document.createElement("a");
    pseudoanchor.setAttribute('href', csvUrl);
    pseudoanchor.setAttribute('download', mapname + OLD_MAPFILE_EXTENSION);
    pseudoanchor.click();
    hideOverlay(); 
}

/*
const EXPORT_NAMES_HEXES = [];
EXPORT_NAMES_HEXES[HEX_EMPTY] = 'empty';
EXPORT_NAMES_HEXES[HEX_WHOLE] = 'whole';
EXPORT_NAMES_HEXES[HEX_FHALF] = 'flat_half';
EXPORT_NAMES_HEXES[HEX_PHALF] = 'point_half';
EXPORT_NAMES_HEXES[HEX_APEX] = 'apex';
EXPORT_NAMES_HEXES[HEX_BULK] = 'bulk';
EXPORT_NAMES_HEXES[HEX_CHANL] = 'channel';
EXPORT_NAMES_HEXES[HEX_JUNCT] = 'junction';
EXPORT_NAMES_HEXES[HEX_CORNR] = 'corner';
EXPORT_NAMES_HEXES[HEX_FINSH] = 'finish';
EXPORT_NAMES_HEXES[HEX_WIDE] = 'wide';
EXPORT_NAMES_HEXES[HEX_JOIN] = 'join';
EXPORT_NAMES_HEXES[HEX_SLICE] = 'slice';
EXPORT_NAMES_HEXES[HEX_TWALL] = 'thin_wall';
EXPORT_NAMES_HEXES[HEX_TEDGE] = 'thin_edge';
EXPORT_NAMES_HEXES[HEX_TCORN] = 'thin_corner';
EXPORT_NAMES_HEXES[HEX_BORDR] = 'hex_border';
EXPORT_NAMES_HEXES[HEX_LRSID] = 'room_side';
EXPORT_NAMES_HEXES[HEX_LRWAL] = 'room_wall';
EXPORT_NAMES_HEXES[HEX_LRCNR] = 'room_corner';
EXPORT_NAMES_HEXES[HEX_SRSID] = 'filled_side';
EXPORT_NAMES_HEXES[HEX_SRCNR] = 'filled_corner';
EXPORT_NAMES_HEXES[HEX_PSEUD] = 'mega_unit';

const EXPORT_NAMES_OBJECTS = [];
EXPORT_NAMES_OBJECTS[0] = 'empty';
EXPORT_NAMES_OBJECTS[FEATURE_DOOR] = 'door';
EXPORT_NAMES_OBJECTS[FEATURE_STAIRS] = 'stairs';
EXPORT_NAMES_OBJECTS[FEATURE_STAIR] = 'stair';
EXPORT_NAMES_OBJECTS[FEATURE_SHAFT] = 'shaft';
EXPORT_NAMES_OBJECTS[FEATURE_UNDERS] = 'unders';
EXPORT_NAMES_OBJECTS[FEATURE_UNDER] = 'under';
EXPORT_NAMES_OBJECTS[FEATURE_BANDS] = 'bands';
EXPORT_NAMES_OBJECTS[FEATURE_BAND] = 'band';
EXPORT_NAMES_OBJECTS[FEATURE_DNDOOR] = 'half_door';
EXPORT_NAMES_OBJECTS[FEATURE_SPIRAL] = 'spiral_stairs';
EXPORT_NAMES_OBJECTS[FEATURE_WINDOW] = 'window';
EXPORT_NAMES_OBJECTS[FEATURE_MIDOOR] = 'mid_door';
EXPORT_NAMES_OBJECTS[FEATURE_MIWIND] = 'mid_window';
EXPORT_NAMES_OBJECTS[FEATURE_AXONS] = 'axons';
EXPORT_NAMES_OBJECTS[FEATURE_AXON] = 'axon';
EXPORT_NAMES_OBJECTS[FEATURE_TDOOR] = 'thin_door';
EXPORT_NAMES_OBJECTS[FEATURE_TWIND] = 'thin_window';
EXPORT_NAMES_OBJECTS[FEATURE_EDOOR] = 'edge_door';
EXPORT_NAMES_OBJECTS[FEATURE_EWIND] = 'edge_window';
EXPORT_NAMES_OBJECTS[FEATURE_ADOOR] = 'angle_door';
EXPORT_NAMES_OBJECTS[FEATURE_AWIND] = 'angle_window';
EXPORT_NAMES_OBJECTS[FEATURE_PILLAR] = 'pillar';
EXPORT_NAMES_OBJECTS[TERRAIN_BOUNDER] = 'bounder';
EXPORT_NAMES_OBJECTS[TERRAIN_WATER] = 'water';
EXPORT_NAMES_OBJECTS[TERRAIN_MOUNTOUT] = 'mountain_outline';
EXPORT_NAMES_OBJECTS[TERRAIN_MOUNTFIL] = 'mountain_fill';
EXPORT_NAMES_OBJECTS[TERRAIN_TREES] = 'forest';
EXPORT_NAMES_OBJECTS[TERRAIN_SWAMP] = 'swamp';
EXPORT_NAMES_OBJECTS[TERRAIN_PLAINS] = 'plains';
EXPORT_NAMES_OBJECTS[TERRAIN_DESERT] = 'desert';
EXPORT_NAMES_OBJECTS[TERRAIN_CAPITOL] = 'capitol';
EXPORT_NAMES_OBJECTS[TERRAIN_CITY] = 'city';
EXPORT_NAMES_OBJECTS[TERRAIN_VILLAGE] = 'village';
EXPORT_NAMES_OBJECTS[TERRAIN_FORT] = 'fort';
EXPORT_NAMES_OBJECTS[TERRAIN_MINE] = 'mine';
EXPORT_NAMES_OBJECTS[TERRAIN_RUIN] = 'ruin';
EXPORT_NAMES_OBJECTS[TERRAIN_INTEREST] = 'interest';
EXPORT_NAMES_OBJECTS[DECOR_TRAP] = 'trap';
EXPORT_NAMES_OBJECTS[DECOR_TRAPDOOR] = 'trapdoor';
EXPORT_NAMES_OBJECTS[DECOR_FLOORPIT] = 'floor_pit';
EXPORT_NAMES_OBJECTS[DECOR_TRIPWIRE] = 'tripwire';
EXPORT_NAMES_OBJECTS[DECOR_CHESTL] = 'chest_large';
EXPORT_NAMES_OBJECTS[DECOR_CHESTS] = 'chest_small';
EXPORT_NAMES_OBJECTS[DECOR_KEY] = 'key';
EXPORT_NAMES_OBJECTS[DECOR_PENTAGRM] = 'pentagram';
EXPORT_NAMES_OBJECTS[DECOR_GATE] = 'gate';
EXPORT_NAMES_OBJECTS[DECOR_LIGHT] = 'light';
EXPORT_NAMES_OBJECTS[DECOR_SCONCE] = 'sconce';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWEL] = 'arrow_edge_large';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWES] = 'arrow_edge_small';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWVL] = 'arrow_vertex_large';
EXPORT_NAMES_OBJECTS[SYMBOL_ARROWVS] = 'arrow_vertex_small';
EXPORT_NAMES_OBJECTS[SYMBOL_PROHIBL] = 'prohibit_large';
EXPORT_NAMES_OBJECTS[SYMBOL_PROHIBS] = 'prohibit_small';
EXPORT_NAMES_OBJECTS[SYMBOL_MARKER] = 'marker';
EXPORT_NAMES_OBJECTS[SYMBOL_EVENT] = 'event';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTL] = 'accent_linear';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTC] = 'accent_circle';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTT] = 'accent_triangle';
EXPORT_NAMES_OBJECTS[SYMBOL_ACCENTD] = 'accent_diamond';
EXPORT_NAMES_OBJECTS[OLD_SYMBOL_MARKER] = 'marker';
EXPORT_NAMES_OBJECTS[OLD_SYMBOL_EVENT] = 'event';
*/

function encodeExportComponents(lyr, isobject) {
	var comps = [];
	for(var c = 0; c < lyr.length; c++) {
		var dataPair = {};
		var hcomp = lyr[c];
		var comptype = hcomp.ctype - (hcomp.ctype % 10);
		var rotation = hcomp.ctype % 10;
		if(hcomp.ctype < 10) {
			// LEGACY - remove this IF and IF ELSE after early SII maps are migrated, keep only the ELSE portion
			if(hcomp.ctype == 0 && (hcomp.ctype % 10 == 1)) {
				comptype = 1;
				rotation = 0;
			}
			else if(hcomp.ctype == 1) {
				comptype = 1;
				rotation = 0;
			}
			else {
				comptype = hcomp.ctype;
				rotation = 0;
			}
		}
		dataPair['component'] = (isobject ? EXPORT_NAMES_OBJECTS[comptype] : EXPORT_NAMES_HEXES[comptype]);
		dataPair['facing'] = rotation;
		dataPair['color'] = materialDef[hcomp.material];
		comps.push(dataPair);
	}
	return comps;
}

function encodeExportAnnotations(lyr) {
	var annots = [];
	for(var c = 0; c < lyr.length; c++) {
		var dataPair = {};
		var hcomp = lyr[c];
		dataPair['character'] = String.fromCharCode(hcomp.ctype - ANNOTATION_BASE);
		dataPair['color'] = materialDef[hcomp.material];
		annots.push(dataPair);
	}
	return annots;
}

function exportSharedMap() {
	/* increase file iteration */
	fileIteration = fileIteration + 1;

	/* build savefile */
	var jsonOutput = {};
	var mapName = getMapName();
	// map info
	jsonOutput[FILEKEYVERS] = VERSION_NUMBER;
	jsonOutput[FILEKEYCOLS] = hexarray.length;
	jsonOutput[FILEKEYROWS] = hexarray[0].length;
	jsonOutput[FILEKEYNAME] = mapName;
	jsonOutput[FILEKEYAUTH] = getMapAuthor();
	jsonOutput[FILEKEYITER] = fileIteration;
	jsonOutput[FILEKEYGRAN] = mapGrain;
	jsonOutput[FILEKEYLWGT] = lineWeight;
	jsonOutput[FILEKEYSHAD] = parseInt(document.getElementById('hexshade').value);
	jsonOutput[FILEKEYZOOM] = parseInt(document.getElementById('zoomlevel').value);
	jsonOutput[FILEKEYTRNO] = translucencyEnabled;
	jsonOutput[FILEKEYTRNA] = TRANSLUCE;
	if(megahexesOn) {
		jsonOutput[FILEKEYMHEX] = megahexesOn + SEP_LIST + mhexImposeX + SEP_LIST + mhexImposeY + SEP_LIST + mhexImposeS + SEP_LIST + mhexType;
	}
	jsonOutput[FILEKEYMHWT] = megahexWeight;
	jsonOutput[FILEKEYMHDT] = megahexDotWeight;
	// map data
	var hexArray = [];
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			var hexData = {};
			hexData[HKEY_X] = x;
			hexData[HKEY_Y] = y;
			var hex = hexarray[x][y];
			hexData['base'] = encodeExportComponents(hex.bases, false);
			hexData['terrain'] = encodeExportComponents(hex.terrains, true);
			hexData['overlay'] = encodeExportComponents(hex.overlays, false);
			hexData['feature'] = encodeExportComponents(hex.features, true);
			hexData['decor'] = encodeExportComponents(hex.decors, true);
			hexData['symbol'] = encodeExportComponents(hex.symbols, true);
			hexData['annotation'] = encodeExportAnnotations(hex.annotations);
			hexArray.push(hexData);
		}
	}
	jsonOutput['hexes'] = hexArray;
	// custom colors
	var customColors = {};
	var customColor = {};
	var customName = '';
	var customValue = '';
	for(var c = CUSTOM_PALETTE_OFFSET; c < materialDef.length; c++) {
		customColor = {};
		customColor[materialNames[c]] = String(materialDef[c]).substring(1, materialDef[c].length); // must remove the leading #, otherwise Javascript treats as a URL subanchor and doesn't write the value
		customColors[c] = customColor;
	}
	jsonOutput[CUSTOMCOLORS] = customColors;
	// backdrop
	if(backdropImage != null) {
		var backdropObject = {};
		backdropObject[BDROPDIMNW] = backdropImage.naturalWidth;
		backdropObject[BDROPDIMNH] = backdropImage.naturalHeight;
		backdropObject[BDROPDIMRW] = backdropImage.width;
		backdropObject[BDROPDIMRH] = backdropImage.height;
		backdropObject[BDROPOFFX] = backdropOffset.x;
		backdropObject[BDROPOFFY] = backdropOffset.y;
		backdropObject[BDROPDATA] = encodeBackdrop(backdropImage);
		jsonOutput[BDROPOBJ] = backdropObject;
	}

	/* pseudolink for named download */
    var saveStr = JSON.stringify(jsonOutput);
    var pseudoanchor = document.createElement("a");
    pseudoanchor.setAttribute('href', FILEPREAMBLE + encodeURIComponent(saveStr));
    pseudoanchor.setAttribute('download', mapName + EXPORTMAP_EXTENSION);
    pseudoanchor.click();
    hideOverlay(); 
}

function loadFileData(cmp) {
	var loadFile = cmp.files[0];
	var lastPeriod = loadFile.name.lastIndexOf(".");
	var extn = (lastPeriod < 0 ? "null" : loadFile.name.substr(lastPeriod).toLowerCase());
	if(MAPFILE_EXTENSION != extn && OLD_MAPFILE_EXTENSION != extn && COLORFILE_EXTENSION != extn) {
		document.getElementById("load-file-message").innerHTML = "<div class='warning'><b>" + loadFile.name + "</b> does not have a supported file extension</div>";
	}
	else {
		document.getElementById("load-file-message").innerHTML = "<div class='success'>Loading file <b>" + loadFile.name + "</b> ...</div>";
		backdropImage = null;
		var filereader = new FileReader();
		filereader.onload = function(lf) {
			parseFileData(lf.target.result.trim(), extn);
		}
		filereader.readAsText(loadFile);
		hideFileLoader();
		drawMap();
	}
}

function parseFileData(raw, extn) {
	if(extn == MAPFILE_EXTENSION) {
		parseShamatMap(raw);
	}
	else if(extn == COLORFILE_EXTENSION) {
		parseCustomColorData(raw);
	}
	else if(extn == OLD_MAPFILE_EXTENSION) {
		parseLegacyShamatMap(raw);
	}
}

function getJsonValue(json, key) {
	if(json[key] > '') {
		return json[key];
	}
	else if(json[key.toUpperCase()] > '') {
		return json[key.toUpperCase()];
	}
	else if(json[key.toLowerCase()] > '') {
		return json[key.toLowerCase()];
	}
	return null;
}

function parseShamatMap(raw) {
	try {
		var mapData = JSON.parse(raw);
		var loadedMapVersion = VERSION_NUMBER;
		loadedMapName = defaultMapName;
		loadedAuthorName = defaultAuthorName;
		var maprows = 0;
		var mapcols = 0;
		fileIteration = 1;
		mapGrain = GRAIN_VERTICAL;
		lineWeight = LINEWEIGHT_PROPORTIONAL;
		megahexesOn = false;
		megahexWeight = MEGAHEX_WEIGHT_DEFAULT;
		megahexDotWeight = MEGADOT_WEIGHT_DEFAULT;
		resetBackdrop();
		var hexArray = [];

		// Map Metadata
		if(getJsonValue(mapData, FILEKEYVERS) > '') {
			loadedMapVersion = getJsonValue(mapData, FILEKEYVERS);
		}
		if(getJsonValue(mapData, FILEKEYNAME) > '') {
			loadedMapName = getJsonValue(mapData, FILEKEYNAME);
		}
		if(getJsonValue(mapData, FILEKEYAUTH) > '') {
			loadedAuthorName = getJsonValue(mapData, FILEKEYAUTH);
		}
		if(getJsonValue(mapData, FILEKEYROWS) > '') {
			maprows = parseInt(getJsonValue(mapData, FILEKEYROWS));
			gridy = maprows;
			document.getElementById('maprows').value = maprows;
			if(maprows > 0 && mapcols > 0) { initialiseGrid(); }
		}
		if(getJsonValue(mapData, FILEKEYCOLS) > '') {
			mapcols = parseInt(getJsonValue(mapData, FILEKEYCOLS));
			gridx = mapcols;
			document.getElementById('mapcols').value = mapcols;
			if(maprows > 0 && mapcols > 0) { initialiseGrid(); }
		}
		// LEGACY - supports early SII keypair, replaced by rows/cols fields now
		if(getJsonValue(mapData, FILEKEYSIZE) > '') {
			var sizePair = getJsonValue(mapData, FILEKEYSIZE);
			mapcols = parseInt(sizePair.substr(0, sizePair.indexOf(SEP_PAIR)));
			maprows = parseInt(sizePair.substr(sizePair.indexOf(SEP_PAIR) + 1));
			gridx = mapcols;
			gridy = maprows;
			initialiseGrid();
		}
		if(getJsonValue(mapData, FILEKEYITER) > '') {
			try {
				fileIteration = parseInt(getJsonValue(mapData, FILEKEYITER));
			}
			catch {
				fileIteration = 1;
			}
		}
		if(getJsonValue(mapData, FILEKEYGRAN) > '') {
			mapGrain = parseInt(getJsonValue(mapData, FILEKEYGRAN));
			document.getElementById("mapgrain").selectedIndex = mapGrain;
		}
		if(getJsonValue(mapData, FILEKEYLWGT) > '') {
			lineWeight = parseInt(getJsonValue(mapData, FILEKEYLWGT));
		}
		if(getJsonValue(mapData, FILEKEYSHAD) > '') {
			var greyscale = parseInt(getJsonValue(mapData, FILEKEYSHAD));
			if(isNaN(greyscale)) {}
			else {
				document.getElementById('hexshade').value = greyscale;
				document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
				hexOutline = "rgb(" + greyscale + "," + greyscale + "," + greyscale + ")";
			}
		}
		if(getJsonValue(mapData, FILEKEYZOOM) > '') {
			var zoomload = parseInt(getJsonValue(mapData, FILEKEYZOOM));
			if(isNaN(zoomload)) {}
			else {
				drawsize = zoomload;
				document.getElementById('zoomlevel').value = drawsize;
				document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
				recalculate();
				resizeCanvas();
			}
		}
		if(getJsonValue(mapData, FILEKEYTRNO) > '') {
			var transload = getJsonValue(mapData, FILEKEYTRNO);
			setTranslucency((transload == true || transload == 'true'));
		}
		if(getJsonValue(mapData, FILEKEYTRNA) > '') {
			var transamt = parseFloat(getJsonValue(mapData, FILEKEYTRNA));
			if(isNaN(transamt)) {} else { setTranslucence(transamt); }
		}
		if(getJsonValue(mapData, FILEKEYMHEX) > '') {
			var mhargs = getJsonValue(mapData, FILEKEYMHEX).split(SEP_LIST);
			megahexesOn = (mhargs[0] == true || mhargs[0] == 'true');
			mhexImposeX = parseInt(mhargs[1]);
			mhexImposeY = parseInt(mhargs[2]);
			mhexImposeS = parseInt(mhargs[3]);
			mhexType = (typeof mhargs[4] === 'undefined' ? MEGAHEX_TYPE_STANDARD : parseInt(mhargs[4]));
			if(isNaN(mhexType)) { mhexType = MEGAHEX_TYPE_STANDARD; }
		}
		if(getJsonValue(mapData, FILEKEYMHWT) > '') {
			var mhexwght = parseInt(getJsonValue(mapData, FILEKEYMHWT));
			if(isNaN(mhexwght)) {} else {
				megahexWeight = mhexwght;
				mhexWeightUpdate(false);
			}
		}
		if(getJsonValue(mapData, FILEKEYMHDT) > '') {
			var mdotwght = parseInt(getJsonValue(mapData, FILEKEYMHDT));
			if(isNaN(mdotwght)) {} else {
				megahexDotWeight = mdotwght;
				mdotWeightUpdate(false);
			}
		}
		// Map Hexes
		if(getJsonValue(mapData, DATAHEX) > '') {
			var dataHexes = getJsonValue(mapData, DATAHEX);
			var loadhex;
			var hexpair;
			var hexX = 0;
			var hexY = 0;
			var hexBases = [];
			var hexTerrains = [];
			var hexOverlays = [];
			var hexFeatures = [];
			var hexDecors = [];
			var hexSymbols = [];
			var hexAnnotations = [];
			for(var h = 0; h < dataHexes.length; h++) {
				loadhex = dataHexes[h];
				hexX = parseInt(loadhex[HKEY_X]);
				hexY = parseInt(loadhex[HKEY_Y]);
				var hex = new Hex(hexX, hexY);
				hexBases = loadhex[HKEY_BASE];
				if(hexBases != undefined && hexBases != null && hexBases.length > 0) {
					for(var i = 0; i < hexBases.length; i++) {
						hexpair = hexBases[i];
						var rot = 0; if(hexpair[DATAPAIR_ROTAT] > '') { rot = parseInt(hexpair[DATAPAIR_ROTAT]); }
						hex.addBase(hexpair[DATAPAIR_CTYPE] + rot, hexpair[DATAPAIR_MATER]);
					}
				}
				hexTerrains = loadhex[HKEY_TERRAIN];
				if(hexTerrains != undefined && hexTerrains != null && hexTerrains.length > 0) {
					for(var i = 0; i < hexTerrains.length; i++) {
						hexpair = hexTerrains[i];
						var rot = 0; if(hexpair[DATAPAIR_ROTAT] > '') { rot = parseInt(hexpair[DATAPAIR_ROTAT]); }
						hex.addTerrain(hexpair[DATAPAIR_CTYPE] + rot, hexpair[DATAPAIR_MATER]);
					}
				}
				hexOverlays = loadhex[HKEY_OVERLAY];
				if(hexOverlays != undefined && hexOverlays != null && hexOverlays.length > 0) {
					for(var i = 0; i < hexOverlays.length; i++) {
						hexpair = hexOverlays[i];
						var rot = 0; if(hexpair[DATAPAIR_ROTAT] > '') { rot = parseInt(hexpair[DATAPAIR_ROTAT]); }
						hex.addOverlay(hexpair[DATAPAIR_CTYPE] + rot, hexpair[DATAPAIR_MATER]);
					}
				}
				hexFeatures = loadhex[HKEY_FEATURE];
				if(hexFeatures != undefined && hexFeatures != null && hexFeatures.length > 0) {
					for(var i = 0; i < hexFeatures.length; i++) {
						hexpair = hexFeatures[i];
						var rot = 0; if(hexpair[DATAPAIR_ROTAT] > '') { rot = parseInt(hexpair[DATAPAIR_ROTAT]); }
						hex.addFeature(hexpair[DATAPAIR_CTYPE] + rot, hexpair[DATAPAIR_MATER]);
					}
				}
				hexDecors = loadhex[HKEY_DECOR];
				if(hexDecors != undefined && hexDecors != null && hexDecors.length > 0) {
					for(var i = 0; i < hexDecors.length; i++) {
						hexpair = hexDecors[i];
						var rot = 0; if(hexpair[DATAPAIR_ROTAT] > '') { rot = parseInt(hexpair[DATAPAIR_ROTAT]); }
						hex.addDecor(hexpair[DATAPAIR_CTYPE] + rot, hexpair[DATAPAIR_MATER]);
					}
				}
				hexSymbols = loadhex[HKEY_SYMBOL];
				if(hexSymbols != undefined && hexSymbols != null && hexSymbols.length > 0) {
					for(var i = 0; i < hexSymbols.length; i++) {
						hexpair = hexSymbols[i];
						var rot = 0; if(hexpair[DATAPAIR_ROTAT] > '') { rot = parseInt(hexpair[DATAPAIR_ROTAT]); }
						hex.addSymbol(hexpair[DATAPAIR_CTYPE] + rot, hexpair[DATAPAIR_MATER]);
					}
				}
				hexAnnotations = loadhex[HKEY_ANNOT];
				if(hexAnnotations != undefined && hexAnnotations != null && hexAnnotations.length > 0) {
					for(var i = 0; i < hexAnnotations.length; i++) {
						hexpair = hexAnnotations[i];
						hex.addAnnotation(hexpair[DATAPAIR_CTYPE], hexpair[DATAPAIR_MATER]);
					}
				}
				hexarray[hexX][hexY] = hex.clone();
			}
		}
		// Custom Colors
		if(getJsonValue(mapData, CUSTOMCOLORS) > '') {
			try {
				for(var [ckey, cdef] of Object.entries(getJsonValue(mapData, CUSTOMCOLORS))) {
					var customColorIndex = parseInt(ckey);
					var customColorName = Object.keys(cdef)[0];
					var customColorValue = Object.values(cdef)[0];
					// LEGACY - support for early SII maps which didn't include the #
					if(!customColorValue.startsWith('#')) { customColorValue = '#' + customColorValue; }
					loadMaterial(customColorIndex, customColorName, customColorValue);
				}
			}
			catch(excustclr) {
				console.log('Exception occurred loading custom colors.');
				console.log(excustclr);
			}
		}
		// Map Backdrop
		if(getJsonValue(mapData, BDROPOBJ) > '') {
			var backdropObject = getJsonValue(mapData, BDROPOBJ);
			backdropImage = new Image();
			var valload;
			if(getJsonValue(backdropObject, BDROPDIMNW) > '') {
				valload = parseInt(getJsonValue(backdropObject, BDROPDIMNW));
				if(isNaN(valload)) {} else { backdropImage.naturalWidth = valload; }
			}
			if(getJsonValue(backdropObject, BDROPDIMNH) > '') {
				valload = parseInt(getJsonValue(backdropObject, BDROPDIMNH));
				if(isNaN(valload)) {} else { backdropImage.naturalHeight = valload; }
			}
			if(getJsonValue(backdropObject, BDROPDIMRW) > '') {
				valload = parseInt(getJsonValue(backdropObject, BDROPDIMRW));
				if(isNaN(valload)) {} else { backdropImage.width = valload; }
			}
			if(getJsonValue(backdropObject, BDROPDIMRH) > '') {
				valload = parseInt(getJsonValue(backdropObject, BDROPDIMRH));
				if(isNaN(valload)) {} else { backdropImage.height = valload; }
			}
			if(getJsonValue(backdropObject, BDROPOFFX) > '') {
				valload = parseInt(getJsonValue(backdropObject, BDROPOFFX));
				if(isNaN(valload)) {} else { backdropOffset.x = valload; }
			}
			if(getJsonValue(backdropObject, BDROPOFFY) > '') {
				valload = parseInt(getJsonValue(backdropObject, BDROPOFFY));
				if(isNaN(valload)) {} else { backdropOffset.y = valload; }
			}
			if(getJsonValue(backdropObject, BDROPDATA) > '') {
				backdropImage.src = getJsonValue(backdropObject, BDROPDATA);
				backdropImage.naturalWidth = backdropImage.naturalWidth;
				backdropImage.naturalHeight = backdropImage.naturalHeight;
				backdropImage.width = backdropImage.width;
				backdropImage.height = backdropImage.height;
				updateBackdropWidgets();
			}
			if(backdropImage != null) {
				backdropImage.addEventListener('load', function() {
					backdropShowing = true;
					document.getElementById("backdroptoggle").className = CSS_PAL_BUTTON + (backdropShowing ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
				}, false);
			}
		}
		else if(getJsonValue(mapData, LEGACYBDROPHDR) > '') {
			backdropImage = new Image();
			var valload;
			if(getJsonValue(mapData, LEGACYBDROPDIMNW) > '') {
				valload = parseInt(getJsonValue(mapData, LEGACYBDROPDIMNW));
				if(isNaN(valload)) {} else { backdropImage.naturalWidth = valload; }
			}
			if(getJsonValue(mapData, LEGACYBDROPDIMNH) > '') {
				valload = parseInt(getJsonValue(mapData, LEGACYBDROPDIMNH));
				if(isNaN(valload)) {} else { backdropImage.naturalHeight = valload; }
			}
			if(getJsonValue(mapData, LEGACYBDROPDIMRW) > '') {
				valload = parseInt(getJsonValue(mapData, LEGACYBDROPDIMRW));
				if(isNaN(valload)) {} else { backdropImage.width = valload; }
			}
			if(getJsonValue(mapData, LEGACYBDROPDIMRH) > '') {
				valload = parseInt(getJsonValue(mapData, LEGACYBDROPDIMRH));
				if(isNaN(valload)) {} else { backdropImage.height = valload; }
			}
			if(getJsonValue(mapData, LEGACYBDROPOFFX) > '') {
				valload = parseInt(getJsonValue(mapData, LEGACYBDROPOFFX));
				if(isNaN(valload)) {} else { backdropOffset.x = valload; }
			}
			if(getJsonValue(mapData, LEGACYBDROPOFFY) > '') {
				valload = parseInt(getJsonValue(mapData, LEGACYBDROPOFFY));
				if(isNaN(valload)) {} else { backdropOffset.y = valload; }
			}
			if(getJsonValue(mapData, LEGACYBDROPDATA) > '') {
				backdropImage.src = getJsonValue(mapData, LEGACYBDROPDATA);
				backdropImage.naturalWidth = backdropImage.naturalWidth;
				backdropImage.naturalHeight = backdropImage.naturalHeight;
				backdropImage.width = backdropImage.width;
				backdropImage.height = backdropImage.height;
				updateBackdropWidgets();
			}
			if(backdropImage != null) {
				backdropImage.addEventListener('load', function() {
					backdropShowing = true;
					document.getElementById("backdroptoggle").className = CSS_PAL_BUTTON + (backdropShowing ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
				}, false);
			}
		}
		// Interface Updates
		drawMap();
		document.getElementById('mapname').value = loadedMapName;
		document.getElementById('mapauthor').value = loadedAuthorName;
	}
	catch(excpt) {
		console.log(excpt);
	}
}

function parseCustomColorData(raw) {
	try {
		var fileData = JSON.parse(raw);
		var loadedCustomColorsVersion = VERSION_NUMBER;
		var loadedCustomColorsAuthorName = defaultAuthorName;

		// Color Collection Metadata
		if(getJsonValue(fileData, FILEKEYVERS) > '') {
			loadedCustomColorsVersion = getJsonValue(fileData, FILEKEYVERS);
		}
		if(getJsonValue(fileData, FILEKEYAUTH) > '') {
			loadedCustomColorsAuthorName = getJsonValue(fileData, FILEKEYAUTH);
		}
		// Custom Colors
		if(getJsonValue(fileData, CUSTOMCOLORS) > '') {
			try {
				for(var [ckey, cdef] of Object.entries(getJsonValue(fileData, CUSTOMCOLORS))) {
					var customColorIndex = parseInt(ckey);
					var customColorName = Object.keys(cdef)[0];
					var customColorValue = Object.values(cdef)[0];
					// LEGACY - support for early SII maps which didn't include the #
					if(!customColorValue.startsWith('#')) { customColorValue = '#' + customColorValue; }
					loadMaterial(customColorIndex, customColorName, customColorValue);
				}
			}
			catch(excustclrload) {
				console.log('Exception occurred loading custom colors.');
				console.log(excustclr);
			}
		}
		// Interface Updates
		deactivatePalette(TOOL_BUTTON_MATSTND);
		activatePalette(TOOL_BUTTON_MATCUST);
		drawMap();
	}
	catch(excpt) {
		console.log(excpt);
	}
}

function parseLegacyShamatMap(raw) {
	var lines = raw.split(NEWLINE);
	if(lines.length < 4) {
		return;
	}
	var maprows = 0;
	var mapcols = 0;
	loadedMapName = defaultMapName;
	loadedAuthorName = defaultAuthorName;
	fileIteration = 1;
	mapGrain = GRAIN_VERTICAL;
	lineWeight = LINEWEIGHT_PROPORTIONAL;
	megahexWeight = MEGAHEX_WEIGHT_DEFAULT;
	megahexDotWeight = MEGADOT_WEIGHT_DEFAULT;
	megahexesOn = false;
	resetBackdrop();
	// process the map header
	var linenum = 0;
	var procline = '';
	var fileVersionNumber = 0;
	while(lines[linenum] != LEGACYDATAHEADER) {
		procline = lines[linenum].trim();
		if(procline == LEGACYFILEHEADER) {
		}
		else if(procline.startsWith(LEGACYFILEKEYVERS)) {
			fileVersionNumber = procline.substr(LEGACYFILEKEYVERS.length);
		}
		else if(procline.startsWith(LEGACYFILEKEYSIZE)) {
			var sizePair = procline.substr(LEGACYFILEKEYSIZE.length);
			mapcols = parseInt(sizePair.substr(0, sizePair.indexOf(SEP_LIST)));
			maprows = parseInt(sizePair.substr(sizePair.indexOf(SEP_LIST) + 1));
			var gridcount = maprows * mapcols;
		}
		else if(procline.startsWith(legacyKey(FILEKEYNAME))) {
			loadedMapName = procline.substr(legacyKey(FILEKEYNAME).length);
		}
		else if(procline.startsWith(legacyKey(FILEKEYAUTH))) {
			loadedAuthorName = procline.substr(legacyKey(FILEKEYAUTH).length);
		}
		else if(procline.startsWith(legacyKey(FILEKEYITER))) {
			try {
				fileIteration = parseInt(procline.substr(legacyKey(FILEKEYITER).length));
			}
			catch(err) {
				fileIteration = 1;
			}
		}
		else if(procline.startsWith(legacyKey(FILEKEYGRAN))) {
			var mapGrainVal = procline.substr(legacyKey(FILEKEYGRAN).length);
			mapGrain = parseInt(mapGrainVal);
			document.getElementById("mapgrain").selectedIndex = mapGrain;
		}
		else if(procline.startsWith(legacyKey(FILEKEYLWGT))) {
			var lineWeightVal = procline.substr(legacyKey(FILEKEYLWGT).length);
			lineWeight = parseInt(lineWeightVal);
			document.getElementById("lineweight").selectedIndex = lineWeight;
		}
		else if(procline.startsWith(legacyKey(FILEKEYSHAD))) {
			var greyscale = parseInt(procline.substr(legacyKey(FILEKEYSHAD).length));
			if(isNaN(greyscale)) {}
			else {
				document.getElementById('hexshade').value = greyscale;
				document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((greyscale / 255) * 100)) + '%';
				hexOutline = "rgb(" + greyscale + "," + greyscale + "," + greyscale + ")";
			}
		}
		else if(procline.startsWith(legacyKey(FILEKEYZOOM))) {
			var zoomload = parseInt(procline.substr(legacyKey(FILEKEYZOOM).length));
			if(isNaN(zoomload)) {}
			else {
				drawsize = zoomload;
				document.getElementById('zoomlevel').value = drawsize;
				document.getElementById('showZoomLevel').innerHTML = drawsize + 'px';
				recalculate();
				resizeCanvas();
			}
		}
		else if(procline.startsWith(legacyKey(FILEKEYTRNO))) {
			var transload = procline.substr(legacyKey(FILEKEYTRNO).length);
			setTranslucency((transload == true || transload == 'true'));
		}
		else if(procline.startsWith(legacyKey(FILEKEYTRNA))) {
			var transamt = parseFloat(procline.substr(legacyKey(FILEKEYTRNA).length));
			if(isNaN(transamt)) {} else { setTranslucence(transamt); }
		}
		else if(procline.startsWith(legacyKey(FILEKEYMHEX))) {
			var mhargs = procline.substr(legacyKey(FILEKEYMHEX).length).split(SEP_LIST);
			mhexImposeX = parseInt(mhargs[0]);
			mhexImposeY = parseInt(mhargs[1]);
			mhexImposeS = parseInt(mhargs[2]);
			mhexType = (typeof mhargs[3] === 'undefined' ? MEGAHEX_TYPE_STANDARD : parseInt(mhargs[3]));
			if(isNaN(mhexType)) { mhexType = MEGAHEX_TYPE_STANDARD; }
			megahexesOn = true;
		}
		else if(procline.startsWith(legacyKey(FILEKEYMHWT))) {
			var mhexwght = parseInt(procline.substr(legacyKey(FILEKEYMHWT).length));
			if(isNaN(mhexwght)) {} else {
				megahexWeight = mhexwght;
				mhexWeightUpdate(false);
			}
		}
		else if(procline.startsWith(legacyKey(FILEKEYMHDT))) {
			var mdotwght = parseInt(procline.substr(legacyKey(FILEKEYMHDT).length));
			if(isNaN(mdotwght)) {} else {
				megahexDotWeight = mdotwght;
				mdotWeightUpdate(false);
			}
		}
		else if(procline.startsWith(LEGACYBDROPHDR)) {
			backdropImage = new Image();
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPDIMNW))) {
			var valload = parseInt(procline.substr(legacyKey(LEGACYBDROPDIMNW).length));
			if(isNaN(valload)) {} else { backdropImage.naturalWidth = valload; }
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPDIMNH))) {
			var valload = parseInt(procline.substr(legacyKey(LEGACYBDROPDIMNH).length));
			if(isNaN(valload)) {} else { backdropImage.naturalHeight = valload; }
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPDIMRW))) {
			var valload = parseInt(procline.substr(legacyKey(LEGACYBDROPDIMRW).length));
			if(isNaN(valload)) {} else { backdropImage.width = valload; }
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPDIMRH))) {
			var valload = parseInt(procline.substr(legacyKey(LEGACYBDROPDIMRH).length));
			if(isNaN(valload)) {} else { backdropImage.height = valload; }
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPDIMRH))) {
			var valload = parseInt(procline.substr(legacyKey(LEGACYBDROPDIMRH).length));
			if(isNaN(valload)) {} else { backdropImage.height = valload; }
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPOFFX))) {
			var valload = parseInt(procline.substr(acyKey(LEGACYBDROPOFFX).length));
			if(isNaN(valload)) {} else { backdropOffset.x = valload; }
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPOFFY))) {
			var valload = parseInt(procline.substr(legacyKey(LEGACYBDROPOFFY).length));
			if(isNaN(valload)) {} else { backdropOffset.y = valload; }
		}
		else if(procline.startsWith(legacyKey(LEGACYBDROPDATA))) {
			var imgrawdata = procline.substr(legacyKey(LEGACYBDROPDATA).length);
			if(imgrawdata != null && imgrawdata != '') {
				backdropImage.src = imgrawdata;
				backdropImage.naturalWidth = backdropImage.naturalWidth;
				backdropImage.naturalHeight = backdropImage.naturalHeight;
				backdropImage.width = backdropImage.width;
				backdropImage.height = backdropImage.height;
				updateBackdropWidgets();
			}
		}
		else {
			// unsupported header line
			debugMsg("UNKNOWN HEADER: " + procline);
		}
		linenum++;
	}
	// header looks good, reinitialise grid for incoming map data
	if(maprows > 0 && mapcols > 0) {
		resizeGrid(maprows, mapcols, false);
		document.getElementById("maprows").value = maprows;
		document.getElementById("mapcols").value = mapcols;
	}
	else {
		document.getElementById("load-file-message").innerHTML = "<span class='warning'>Map has zero length size setting</span>";
		return;
	}
	linenum++;
	// process the map data
	var dvals = [];
	var hexX = 0;
	var hexY = 0;
	var gstr = '';
	var grns = [];
	var wstr = '';
	var wals = [];
	var fstr = '';
	var ftrs = [];
	var comp = 0;
	var matr = 0;
	for(var l = linenum; l < lines.length; l++) {
		if(lines[l].substr(0, LEGACYDATAHEX.length) == LEGACYDATAHEX) {
			// hex data
			dvals = lines[l].substr(LEGACYDATAHEX.length).split(SEP_DATA);
			hexX = parseInt(dvals[0]);
			hexY = parseInt(dvals[1]);
			var hex = new Hex(hexX, hexY);
			gstr = dvals[2];
			wstr = dvals[3];
			fstr = (dvals.length < 5 ? '' : dvals[4]);
			grns = [];
			wals = [];
			ftrs = [];
			if(gstr.length > 2) {
				gstr = gstr.substr(1, gstr.length - 2);
				grns = gstr.split(SEP_LIST);
				for(var g = 0; g < grns.length; g++) {
					comp = parseInt(grns[g].substr(0, grns[g].indexOf(SEP_PAIR)));
					matr = parseInt(grns[g].substr(grns[g].indexOf(SEP_PAIR) + 1));
					hex.addBase(comp, matr);
				}
			}
			if(wstr.length > 2) {
				wals = wstr.substr(1, wstr.length - 2).split(SEP_LIST);
				for(var w = 0; w < wals.length; w++) {
					comp = parseInt(wals[w].substr(0, wals[w].indexOf(SEP_PAIR)));
					matr = parseInt(wals[w].substr(wals[w].indexOf(SEP_PAIR) + 1));
					hex.addOverlay(comp, matr);
				}
			}
			if(fstr.length > 2) {
				ftrs = fstr.substr(1, fstr.length - 2).split(SEP_LIST);
				for(var f = 0; f < ftrs.length; f++) {
					feat = parseInt(ftrs[f].substr(0, ftrs[f].indexOf(SEP_PAIR)));
					matr = parseInt(ftrs[f].substr(ftrs[f].indexOf(SEP_PAIR) + 1));
					if(feat >= ANNOTATION_BASE) {
						hex.addAnnotation(feat, matr);
					}
					else if(feat == OLD_SYMBOL_MARKER) {
						// migrate legacy symbol to new scheme
						hex.addSymbol(SYMBOL_MARKER, matr);
					}
					else if(feat == OLD_SYMBOL_EVENT) {
						// migrate legacy symbol to new scheme
						hex.addSymbol(SYMBOL_EVENT, matr);
					}
					else if(feat >= TERRAIN_BOUND_B && feat <= TERRAIN_BOUND_T) {
						hex.addTerrain(feat, matr);
					}
					else if(feat >= DECOR_BOUND_B && feat <= DECOR_BOUND_T) {
						hex.addDecor(feat, matr);
					}
					else if(feat >= SYMBOL_BOUND_B && feat <= SYMBOL_BOUND_T) {
						hex.addSymbol(feat, matr);
					}
					else {
						hex.addFeature(feat, matr);
					}
				}
			}
			hexarray[hexX][hexY] = hex;
		}
	}
	hideFileLoader();
	toggleMegahexes(megahexesOn);
	if(backdropImage != null) {
		backdropImage.addEventListener('load', function() {
			backdropShowing = true;
			document.getElementById("backdroptoggle").className = CSS_PAL_BUTTON + (backdropShowing ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
			drawMap();
		}, false);
	}
	drawMap();
	document.getElementById('mapname').value = loadedMapName;
	document.getElementById('mapauthor').value = loadedAuthorName;
}

function hideOverlay() {
	document.getElementById("load-file-selector").style.display = CSS_BLOCK_HIDE;
	document.getElementById("save-file-selector").style.display = CSS_BLOCK_HIDE;
	document.getElementById("export-file-selector").style.display = CSS_BLOCK_HIDE;
	document.getElementById("image-save-gadget").style.display = CSS_BLOCK_HIDE;
	document.getElementById("load-backdrop-selector").style.display = CSS_BLOCK_HIDE;
	document.getElementById('material-editor').style.display = CSS_BLOCK_HIDE;
	document.getElementById('about-modal').style.display = CSS_BLOCK_HIDE;
	overlayDiv.style.display = CSS_BLOCK_HIDE;
	shaderDiv.style.display = CSS_BLOCK_HIDE;
	materialToEdit = null;
}

function showMaterialEditor() {
	document.getElementById('customColorMessage').innerHTML = '';
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById('material-editor').style.display = CSS_BLOCK_SHOW;
}

function hideMaterialEditor() {
	hideOverlay();
}

function showAboutModal() {
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById('about-modal').style.display = CSS_BLOCK_SHOW;
}

function hideAboutModal() {
	hideOverlay();
}

function displayFileLoader() {
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("load-file-selector").style.display = CSS_BLOCK_SHOW;
}

function hideFileLoader() {
	document.getElementById("load-file-form").reset();
	document.getElementById("load-file-message").innerHTML = '';
	hideOverlay();
}

function displayFileSaver() {
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("save-file-selector").style.display = CSS_BLOCK_SHOW;
}

function hideFileSaver() {
	document.getElementById("save-file-form").reset();
	hideOverlay();
}

function displayFileExporter() {
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("export-file-selector").style.display = CSS_BLOCK_SHOW;
}

function hideFileExporter() {
	document.getElementById("export-file-form").reset();
	hideOverlay();
}

function displayBackdropLoader() {
	if(backdropLocked) {
		warnBackdropLocked();
	}
	else {
		shaderDiv.style.display = CSS_BLOCK_SHOW;
		overlayDiv.style.display = CSS_BLOCK_SHOW;
		document.getElementById("load-backdrop-selector").style.display = CSS_BLOCK_SHOW;
	}
}

function hideBackdropLoader() {
	document.getElementById("load-backdrop-form").reset();
	document.getElementById("load-backdrop-message").innerHTML = '';
	hideOverlay();
}

function saveImage() {
	var currMapSize = drawsize;
	if(drawsize != parseInt(document.getElementById("imagesavesize").value)) {
		drawsize = parseInt(document.getElementById("imagesavesize").value);
		recalculate();
		resizeCanvas();
		drawMap();
	}
	var mapname = getMapName();
	var png = canvas.toDataURL("image/png").replace("image/png", "image/octet-stream");
	var pseudoanchor = document.createElement("a");
	pseudoanchor.href = png;
	pseudoanchor.download = mapname + ".png";
	pseudoanchor.style.display = CSS_BLOCK_HIDE;
	document.body.appendChild(pseudoanchor);
	pseudoanchor.click();
	document.body.removeChild(pseudoanchor);
	if(drawsize != currMapSize) {
		drawsize = currMapSize;
		recalculate();
		resizeCanvas();
		drawMap();
	}
	hideImageSaveSelector();
}

function imageSaveSizeSet() {
	var newsavesize = document.getElementById("imagesavesize").value;
	var savesizex = Math.ceil(((newsavesize * 1.5) * gridx) + (leftmargin * 2) + (newsavesize / 2));
	var savesizey = Math.ceil(((newsavesize * CALCSQRT) * gridy) + (topmargin * 2) + (newsavesize / 2) + (newsavesize / 4));
	document.getElementById("imgsavemsg").innerHTML = "<span class='sizemsg'>" + savesizex + " x " + savesizey + " pixels</span>";
}

function displayImageSaveSelector() {
	hideFileSaver();
	document.getElementById("imagesavesize").value = drawsize;
	imageSaveSizeSet();
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("image-save-gadget").style.display = CSS_BLOCK_SHOW;
}

function hideImageSaveSelector() {
	document.getElementById("imgsavemsg").innerHTML = "&nbsp;";
	document.getElementById("image-save-gadget").style.display = CSS_BLOCK_HIDE;
	overlayDiv.style.display = CSS_BLOCK_HIDE;
	shaderDiv.style.display = CSS_BLOCK_HIDE;
}

function updateBackdropWidgets() {
	if(backdropImage != null) {
		document.getElementById('showBackdropWidth').value = backdropImage.width;
		document.getElementById('showBackdropHeight').value = backdropImage.height;
		document.getElementById('showBackdropOffsetX').value = backdropOffset.x;
		document.getElementById('showBackdropOffsetY').value = backdropOffset.y;
	}
	else {
		document.getElementById('showBackdropWidth').value = 0;
		document.getElementById('showBackdropHeight').value = 0;
		document.getElementById('showBackdropOffsetX').value = 0;
		document.getElementById('showBackdropOffsetY').value = 0;
	}
}

function loadBackdropData(cmp) {
	var loadFile = cmp.files[0];
	var lastPeriod = loadFile.name.lastIndexOf(".");
	var extn = (lastPeriod < 0 ? "null" : loadFile.name.substr(lastPeriod).toLowerCase());
	if(".png" != extn && ".jpg" != extn && ".jpeg" != extn) {
		document.getElementById("load-backdrop-message").innerHTML = "<span class='warning'><b>" + loadFile.name + "</b> does not have supported extension.<br/>Image must be a PNG or JPEG.</span>";
	}
	else {
		backdropImage = new Image();
		document.getElementById("load-backdrop-message").innerHTML = "<span class='success'>Loading image <b>" + loadFile.name + "</b> ...</span>";
		var filereader = new FileReader();
		filereader.onload = function() {
			backdropImage.src = filereader.result;
		}
		filereader.readAsDataURL(loadFile);
		backdropImage.addEventListener('load', function() {
			hideOverlay();
			backdropShowing = true;
			document.getElementById("backdroptoggle").className = CSS_PAL_BUTTON + (backdropShowing ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
			setTranslucency(true);
			backdropToCorner();
			updateBackdropWidgets();
			drawMap();
		}, false);
		hideBackdropLoader();
	}
}

function toggleBackdrop() {
	backdropShowing = !backdropShowing;
	document.getElementById("backdroptoggle").className = CSS_PAL_BUTTON + (backdropShowing ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
	if(backdropShowing != translucencyEnabled) {
		toggleTranslucency();
	}
	drawMap();
}

function resetBackdrop() {
	backdropImage = null;
	changeBackdropLockState(false);
	updateBackdropWidgets();
	setTranslucency(false);
}

function removeBackdrop() {
	if(backdropLocked) {
		warnBackdropLocked();
	}
	else {
		resetBackdrop();
		drawMap();
	}
}

function backdropFitWidth() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = canvas.width;
			backdropImage.height = backdropImage.naturalHeight * (canvas.width / backdropImage.naturalWidth);
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropFitHeight() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = backdropImage.naturalWidth * (canvas.height / backdropImage.naturalHeight);
			backdropImage.height = canvas.height;
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropFitBoth() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = canvas.width;
			backdropImage.height = canvas.height;
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropFitActual() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropImage.width = backdropImage.naturalWidth;
			backdropImage.height = backdropImage.naturalHeight;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function resizeBackdropWidth(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newWidth = backdropImage.width + amt;
			if(newWidth > 0 && newWidth < 10000) {
				backdropImage.width = newWidth;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function resizeBackdropHeight(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newHeight = backdropImage.height + amt;
			if(newHeight > 0 && newHeight < 10000) {
				backdropImage.height = newHeight;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function shiftBackdropHorizontal(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newOffsetX = backdropOffset.x + amt;
			if(newOffsetX > -(backdropImage.width + 1) && newOffsetX < canvas.width) {
				backdropOffset.x = newOffsetX;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function shiftBackdropVertical(amt) {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			var newOffsetY = backdropOffset.y + amt;
			if(newOffsetY > -(backdropImage.height + 1) && newOffsetY < canvas.height) {
				backdropOffset.y = newOffsetY;
				updateBackdropWidgets();
				drawMap();
			}
		}
	}
}

function backdropToCorner() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropOffset.x = 0;
			backdropOffset.y = 0;
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function backdropToCenter() {
	if(backdropImage != null) {
		if(backdropLocked) {
			warnBackdropLocked();
		}
		else {
			backdropOffset.x = Math.round((canvas.width - backdropImage.width) / 2);
			backdropOffset.y = Math.round((canvas.height - backdropImage.height) / 2);
			updateBackdropWidgets();
			drawMap();
		}
	}
}

function setTranslucency(enableTrans) {
	translucencyEnabled = enableTrans;
	document.getElementById("translucenttoggle").className = CSS_PAL_BUTTON + (translucencyEnabled ? ' ' + CSS_PAL_BUTTON_ACTIVE : "");
}

function toggleTranslucency() {
	setTranslucency(!translucencyEnabled);
	drawMap();
}

function writeVersionNumber() {
	document.write('v' + VERSION_NUMBER);
}

function setAnnotation(annot) {
	if(activeAnnotation != ANNOTATION_BASE) { document.getElementById('annotate_' + activeAnnotation).className = 'annotatetool'; }
	activeAnnotation = annot;
	if(activeAnnotation != ANNOTATION_BASE) { document.getElementById('annotate_' + activeAnnotation).className = 'annotatetool annotatetool-active'; resetMode(MODE_PAINT); }
}

function goToHelp(sect) {
	var helpdiv = document.getElementById('paletteHelp');
	var helpsect = document.getElementById(sect);
	helpdiv.scrollTop = helpsect.offsetTop - helpdiv.offsetTop;
	return false;
}

function displayPrefSetter() {
	prefRestoreCreator = defaultAuthorName;
	prefRestoreGridshade = defaultGridGrey;
	overlayDiv.className = "setprefdiv";
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("prefSetter").style.display = CSS_BLOCK_SHOW;
	document.getElementById('prefInputCreator').value = getMapAuthor();
	document.getElementById('prefInputRows').value = gridy;
	document.getElementById('prefInputCols').value = gridx;
	document.getElementById('prefMapgrain').value = mapGrain;
	document.getElementById('prefLineweight').value = lineWeight;
	document.getElementById('prefHexshade').value = defaultGridGrey;
	document.getElementById('showShadeLevelPref').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
}

function cancelPrefSetter() {
	defaultAuthorName = prefRestoreCreator;
	document.getElementById('prefInputCreator').value = defaultAuthorName;
	defaultGridGrey = prefRestoreGridshade;
	document.getElementById('hexshade').value = defaultGridGrey;
	document.getElementById('showShadeLevel').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
	document.getElementById('prefHexshade').value = defaultGridGrey;
	document.getElementById('showShadeLevelPref').innerHTML = (100 - Math.round((defaultGridGrey / 255) * 100)) + '%';
	defaultGridColor = "rgb(" + defaultGridGrey + "," + defaultGridGrey + "," + defaultGridGrey + ")";
	hexOutline = defaultGridColor;
	drawMap(true);
}

function checkVersion() {
	var req = new XMLHttpRequest();
//	req.open('GET', 'blob:file%3A///D%3A/files/USB/shadekeep/shadekeep/shamat/v.html', true);
	req.onload = function() {
		debugMsg('Fetching...');
		debugMsg(req);
		debugMsg(req.response);
	}
	req.open('GET', 'http://shadekeep.com/index.html', true);
	req.send();
	window.alert("Version: " + VERSION_NUMBER);
}

function getLineIntersection(line1Start, line1End, line2Start, line2End) {
	var line1YDiff = line1End.y - line1Start.y;
	var line1XDiff = line1Start.x - line1End.x;
	var line1Slope = (line1YDiff * line1Start.x) + (line1XDiff * line1Start.y);
	var line2YDiff = line2End.y - line2Start.y;
	var line2XDiff = line2Start.x - line2End.x;
	var line2Slope = (line2YDiff * line2Start.x) + (line2XDiff * line2Start.y);
	var determin   = (line1YDiff * line2XDiff) - (line1XDiff * line2YDiff);
	if(determin == 0) {
		// parallel lines, no intersection
		return line1Start;
	}
	else {
		var interX = ((line2XDiff * line1Slope) - (line1XDiff * line2Slope)) / determin;
		var interY = ((line1YDiff * line2Slope) - (line2YDiff * line1Slope)) / determin;
		return new Point(Math.round(interX), Math.round(interY));
	}
}

function clearMapGrid() {
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			hexarray[x][y].clearAll();
		}
	}
}

function generateRandomMap() {
	hideRandomMapOverlay();
	var maptype = "color";
	var maptypes = document.getElementsByName('rndmaptype');
	for(m = 0; m < maptypes.length; m++) {
		if(maptypes[m].checked) {
			maptype = maptypes[m].value;
		}
	}
	var colorMap = (maptype == "color");
	clearMapGrid();
	randomMapSeed(colorMap);
	while(randomMapHasBlanks()) {
		randomMapIterate();
	}
	if(document.getElementById('rndriver').checked) {
		randomMapRivers(colorMap);
	}
	if(document.getElementById('rndpath').checked) {
		randomMapPaths(colorMap);
	}
	randomMapTerrains(colorMap);
	drawMap();
}

function getRandom(upperval) {
	return Math.floor(Math.random() * upperval);
}

function randomMapSeed(colorMap) {
	var seedCount = Math.ceil((hexarray.length * hexarray[0].length) / 50);
	var seedPlant = seedCount;
	var seedValue = TERRAIN_EMPTY;
	var seedMater = 1;
	var baseValue = HEX_EMPTY;
	var baseMater = MATERIAL_EMPTY;
	var seedPool = [];
	if(document.getElementById('rndblank').checked) {
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
		seedPool.push('EMPTY');
	}
	if(document.getElementById('rndwater').checked) {
		seedPool.push('WATER');
	}
	if(document.getElementById('rndmountains').checked) {
		seedPool.push('MOUNTAIN');
	}
	if(document.getElementById('rndforest').checked) {
		seedPool.push('FOREST');
		seedPool.push('FOREST');
		seedPool.push('FOREST');
	}
	if(document.getElementById('rndswamp').checked) {
		seedPool.push('SWAMP');
	}
	if(document.getElementById('rndplains').checked) {
		seedPool.push('PLAINS');
		seedPool.push('PLAINS');
	}
	if(document.getElementById('rnddesert').checked) {
		seedPool.push('DESERT');
	}
	if(seedPool.length < 1) {
		seedPool.push('EMPTY');
	}
	while(seedPlant > 0) {
		var findHex = true;
		switch(seedPool[getRandom(seedPool.length)]) {
			case 'EMPTY':
				baseValue = HEX_EMPTY;
				baseMater = MATERIAL_EMPTY;
				seedValue = TERRAIN_EMPTY;
				seedMater = 0;
				break;
			case 'PLAINS':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 5 : MATERIAL_EMPTY);
				seedValue = TERRAIN_PLAINS;
				seedMater = (colorMap ? 15 : 9);
				break;
			case 'FOREST':
				baseValue = HEX_EMPTY;
				baseMater = (colorMap ? MATERIAL_EMPTY : MATERIAL_EMPTY);
				seedValue = TERRAIN_TREES;
				seedMater = (colorMap ? 7 : 3);
				break;
			case 'SWAMP':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 14 : MATERIAL_EMPTY);
				seedValue = TERRAIN_SWAMP;
				seedMater = (colorMap ? 1 : 1);
				break;
			case 'DESERT':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 10 : MATERIAL_EMPTY);
				seedValue = TERRAIN_DESERT;
				seedMater = (colorMap ? 4 : 9);
				break;
			case 'MOUNTAIN':
				baseValue = HEX_WHOLE;
				baseMater = (colorMap ? 3 : MATERIAL_EMPTY);
				seedValue = TERRAIN_MOUNTFIL;
				seedMater = (colorMap ? 6 : 3);
				break;
			case 'WATER':
				baseValue = HEX_EMPTY;
				baseMater = (colorMap ? MATERIAL_EMPTY : MATERIAL_EMPTY);
				seedValue = TERRAIN_WATER;
				seedMater = (colorMap ? 13 : 1);
				break;
		}
		while(findHex) {
			var randX = getRandom(hexarray.length);
			var randY = getRandom(hexarray[0].length);
			if(hexarray[randX][randY].terrains.length == 0) {
				hexarray[randX][randY].addBase(baseValue, baseMater);
				hexarray[randX][randY].addTerrain(seedValue, seedMater);
				findHex = false;
			}
		}
		seedPlant--;
	}
	drawMap();
}

function randomMapHasBlanks() {
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			if(hexarray[x][y].terrains.length == 0) {
				return true;
			}
		}
	}
	return false;
}

function randomMapIterate() {
	var nextPt;
	for(x = 0; x < hexarray.length; x++) {
		for(y = 0; y < hexarray[x].length; y++) {
			nextPt = getHexAt(x, y, getRandom(6));
			if(nextPt.x != OFFGRID && nextPt.y != OFFGRID) {
				if(hexarray[x][y].terrains.length > 0) {
					if(hexarray[nextPt.x][nextPt.y].terrains.length == 0) {
						hexarray[nextPt.x][nextPt.y].addBase(hexarray[x][y].bases[0].ctype, hexarray[x][y].bases[0].material);
						hexarray[nextPt.x][nextPt.y].addTerrain(hexarray[x][y].terrains[0].ctype, hexarray[x][y].terrains[0].material);
					}
				}
				else {
					if(hexarray[nextPt.x][nextPt.y].terrains.length > 0) {
						hexarray[x][y].addBase(hexarray[nextPt.x][nextPt.y].bases[0].ctype, hexarray[nextPt.x][nextPt.y].bases[0].material);
						hexarray[x][y].addTerrain(hexarray[nextPt.x][nextPt.y].terrains[0].ctype, hexarray[nextPt.x][nextPt.y].terrains[0].material);
					}
				}
			}
		}
	}
}

function randomMapPaths(colorMap) {
	var pathCount = Math.max((Math.ceil((hexarray.length * hexarray[0].length) / 500) + (getRandom(3) - 2)), 1);
	var pathLoop = 0;
	var pathMater = (colorMap ? 15 : 9);
	var addPath = true;
	while(pathLoop < pathCount) {
		addPath = true;
		var randX = getRandom(hexarray.length);
		var randY = getRandom(hexarray[0].length);
		var rota = getRandom(6);
		if(hexarray[randX][randY].terrains.length > 0) {
			if(hexarray[randX][randY].terrains[0].ctype == TERRAIN_WATER || hexarray[randX][randY].terrains[0].ctype == TERRAIN_MOUNTFIL) {
				addPath = false;
			}
		}
		if(addPath) {
			hexarray[randX][randY].addBase(HEX_CHANL + rota, pathMater);
			randomMapPathPlot(randX, randY, rota, pathMater);
			rota = rota - 3;
			if(rota < 0) { rota = rota + 6; }
			randomMapPathPlot(randX, randY, rota, pathMater);
			pathLoop++;
		}
	}
}

function randomMapPathPlot(x, y, direction, pathMater) {
	var hexPt = getHexAt(x, y, direction);
	if(hexPt.x == OFFGRID || hexPt.y == OFFGRID) {
		return;
	}
	if(hexarray[hexPt.x][hexPt.y].terrains.length > 0) {
		if(hexarray[hexPt.x][hexPt.y].terrains[0].ctype == TERRAIN_WATER || hexarray[hexPt.x][hexPt.y].terrains[0].ctype == TERRAIN_MOUNTFIL) {
			return;
		}
	}
	var pathAct = getRandom(8);
	switch(pathAct) {
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
			// Continue path
			hexarray[hexPt.x][hexPt.y].addBase(HEX_CHANL + direction, pathMater);
			break;
		case 6:
			// Turn counterclockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			direction = direction - 2;
			if(direction < 0) { direction += 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			break;
		case 7:
			// Turn clockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			direction = direction + 2;
			if(direction > 5) { direction -= 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JUNCT + direction, pathMater);
			break;
	}
	randomMapPathPlot(hexPt.x, hexPt.y, direction, pathMater);
}

function randomMapRivers(colorMap) {
	var pathCount = Math.max((Math.ceil((hexarray.length * hexarray[0].length) / 750) + (getRandom(3) - 2)), 1);
	var pathLoop = 0;
	var pathMater = (colorMap ? 13 : 3);
	var addPath = true;
	while(pathLoop < pathCount) {
		addPath = true;
		var randX = getRandom(hexarray.length);
		var randY = getRandom(hexarray[0].length);
		var rota = getRandom(6);
		if(hexarray[randX][randY].terrains.length > 0) {
			if(hexarray[randX][randY].terrains[0].ctype == TERRAIN_WATER || hexarray[randX][randY].terrains[0].ctype == TERRAIN_MOUNTFIL) {
				addPath = false;
			}
		}
		if(addPath) {
			hexarray[randX][randY].addBase(HEX_WIDE + rota, pathMater);
			randomMapRiverPlot(randX, randY, rota, pathMater);
			rota = rota - 3;
			if(rota < 0) { rota = rota + 6; }
			randomMapRiverPlot(randX, randY, rota, pathMater);
			pathLoop++;
		}
	}
}

function randomMapRiverPlot(x, y, direction, pathMater) {
	var hexPt = getHexAt(x, y, direction);
	if(hexPt.x == OFFGRID || hexPt.y == OFFGRID) {
		return;
	}
	if(hexarray[hexPt.x][hexPt.y].terrains.length > 0) {
		if(hexarray[hexPt.x][hexPt.y].terrains[0].ctype == TERRAIN_WATER || hexarray[hexPt.x][hexPt.y].terrains[0].ctype == TERRAIN_MOUNTFIL) {
			return;
		}
	}
	var pathAct = getRandom(8);
	switch(pathAct) {
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
			// Continue path
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WIDE + direction, pathMater);
			break;
		case 6:
			// Turn counterclockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			direction = direction - 2;
			if(direction < 0) { direction += 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			break;
		case 7:
			// Turn clockwise
			direction = direction - 3;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			direction = direction + 2;
			if(direction > 5) { direction -= 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_JOIN + direction, pathMater);
			break;
	}
	randomMapRiverPlot(hexPt.x, hexPt.y, direction, pathMater);
}

function randomMapTerrains(colorMap) {
	var randX = 0;
	var randY = 0;
	var isOkay = false;

	if(document.getElementById('rndcapitol').checked) {
		var searching = true;
		var attempts = 0;
		while(searching && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].terrains.length; i++) {
				if(
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MOUNTFIL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addTerrain(TERRAIN_CAPITOL, (colorMap ? 1 : 1)); // was 16
				searching = false;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndcity').checked) {
		var cityCount = Math.ceil((hexarray.length * hexarray[0].length) / 500);
		var ccounter = 0;
		var attempts = 0;
		while(ccounter < cityCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].terrains.length; i++) {
				if(
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MOUNTFIL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addTerrain(TERRAIN_CITY, (colorMap ? 1 : 1)); // was 13
				ccounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndvillage').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 250);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].terrains.length; i++) {
				if(
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addTerrain(TERRAIN_VILLAGE, (colorMap ? 1 : 1)); // was 11
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndfort').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 500);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].terrains.length; i++) {
				if(
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addTerrain(TERRAIN_FORT, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndmine').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 750);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].terrains.length; i++) {
				if(
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_WATER ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_SWAMP ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_TREES ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_DESERT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addTerrain(TERRAIN_MINE, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndruin').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 500);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].terrains.length; i++) {
				if(
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addTerrain(TERRAIN_RUIN, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}

	if(document.getElementById('rndinterest').checked) {
		var townCount = Math.ceil((hexarray.length * hexarray[0].length) / 600);
		var tcounter = 0;
		var attempts = 0;
		while(tcounter < townCount && attempts < 100) {
			randX = getRandom(hexarray.length);
			randY = getRandom(hexarray[0].length);
			isOkay = true;
			for(var i = 0; i < hexarray[randX][randY].terrains.length; i++) {
				if(
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CAPITOL ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_CITY ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_VILLAGE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_FORT ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_MINE ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_RUIN ||
					hexarray[randX][randY].terrains[i].ctype == TERRAIN_INTEREST
				) {
					isOkay = false;
				}
			}
			if(isOkay) {
				hexarray[randX][randY].addTerrain(TERRAIN_INTEREST, (colorMap ? 1 : 1));
				tcounter++;
				attempts = 0;
			}
			else {
				attempts++;
			}
		}
	}
}

function displayRandomMapSelector() {
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("random-map-terrain").style.display = CSS_BLOCK_SHOW;
}

function hideRandomMapOverlay() {
	document.getElementById("random-map-terrain").style.display = CSS_BLOCK_HIDE;
	overlayDiv.style.display = CSS_BLOCK_HIDE;
	shaderDiv.style.display = CSS_BLOCK_HIDE;
}

function generateRandomRoomDungeon() {
	hideRandomRoomsSettings();
	var minRooms = 5;
	var maxRooms = 12;
	var minRoomWidth = 4;
	var minRoomHeight = 4;
	var maxRoomWidth = 12;
	var maxRoomHeight = 12;
	var roomSpacing = 1;
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomCountMin.value)))  { minRooms = parseInt(document.randomroomelems.inpRoomCountMin.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomCountMax.value)))  { maxRooms = parseInt(document.randomroomelems.inpRoomCountMax.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomWidthMin.value)))  { minRoomWidth = parseInt(document.randomroomelems.inpRoomWidthMin.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomWidthMax.value)))  { maxRoomWidth = parseInt(document.randomroomelems.inpRoomWidthMax.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomHeightMin.value))) { minRoomHeight = parseInt(document.randomroomelems.inpRoomHeightMin.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomHeightMax.value))) { maxRoomHeight = parseInt(document.randomroomelems.inpRoomHeightMax.value); }
	if(Number.isInteger(parseInt(document.randomroomelems.inpRoomSpacing.value)))   { roomSpacing = parseInt(document.randomroomelems.inpRoomSpacing.value); }
	if(maxRooms < minRooms) { var tmp = maxRooms; maxRooms = minRooms; minRooms = tmp; }
	if(maxRoomWidth < minRoomWidth) { var tmp = maxRoomWidth; maxRoomWidth = minRoomWidth; minRoomWidth = tmp; }
	if(maxRoomHeight < minRoomHeight) { var tmp = maxRoomHeight; maxRoomHeight = minRoomHeight; minRoomHeight = tmp; }
	// generate random room set
	var rooms = [];
	rooms[0] = randomRoom(minRoomWidth, minRoomHeight, maxRoomWidth, maxRoomHeight);
	var numRooms = minRooms + getRandom(maxRooms - minRooms);
	var r = 1;
	var breakout = 0;
	while(r < numRooms && breakout < 500) {
		var newRoom = randomRoom(minRoomWidth, minRoomHeight, maxRoomWidth, maxRoomHeight);
		var isDistinct = true;
		for(var tr = 0; tr < rooms.length; tr++) {
			if(newRoom.overlaps(rooms[tr], roomSpacing)) {
				isDistinct = false;
			}
		}
		if(isDistinct) {
			rooms[r] = newRoom;
			r++;
			breakout = 0;
		}
		else {
			breakout++;
		}
	}
	// fill rooms on map
	clearMapGrid();
	var roomFill = activeMaterial;
	for(var rd = 0; rd < rooms.length; rd++) {
		for(var x = rooms[rd].startX; x <= rooms[rd].endX; x++) {
			for(var y = rooms[rd].startY; y <= rooms[rd].endY; y++) {
				hexarray[x][y].addBase(HEX_WHOLE, roomFill);
			}
		}
	}
	joinRooms(rooms, roomFill);
	drawMap();
}

function randomRoom(minW, minH, maxW, maxH) {
	var rWidth  = minW + getRandom(maxW - minW);
	var rHeight = minH + getRandom(maxH - minH);
	var rX = getRandom(hexarray.length - rWidth);
	var rY = getRandom(hexarray[0].length - rHeight);
	return new RoomDef(rX, rY, rWidth, rHeight);
}

function joinRooms(roomArray, mater) {
	var connections = [];
	var currentRoom = null;
	var compareRoom = null;
	var closestRoom = -1;
	var dist = Math.max(hexarray.length, hexarray[0].length);
	var thisdist = Math.max(hexarray.length, hexarray[0].length);
	var canAdd = true;
	for(var r = 0; r < roomArray.length; r++) {
		currentRoom = roomArray[r];
		closestRoom = -1;
		dist = Math.max(hexarray.length, hexarray[0].length);
		for(var r2 = 0; r2 < roomArray.length; r2++) {
			compareRoom = roomArray[r2];
			thisdist = Math.max(hexarray.length, hexarray[0].length);
			if(r == r2) {
				// same room, skip
			}
			else if((compareRoom.startX > currentRoom.endX || compareRoom.endX < currentRoom.startX) && (compareRoom.startY > currentRoom.endY || compareRoom.endY < currentRoom.startY)) {
				// not aligned
			}
			else {
				if(compareRoom.startX > currentRoom.endX) {
					thisdist = compareRoom.startX - currentRoom.endX;
				}
				else if(compareRoom.endX < currentRoom.startX) {
					thisdist = currentRoom.startX - compareRoom.endX;
				}
				else if(compareRoom.startY > currentRoom.endY) {
					thisdist = compareRoom.startY - currentRoom.endY;
				}
				else if(compareRoom.endY < currentRoom.startY) {
					thisdist = currentRoom.startY - compareRoom.endY;
				}
				if(thisdist < dist) {
					canAdd = true;
					if(typeof connections[r2] === 'undefined') {
						// not set yet, is okay
					}
					else {
						if(connections[r2] == r) {
							// already connected
							canAdd = false;
						}
					}
					if(canAdd) {
						dist = thisdist;
						closestRoom = r2;
					}
				}
			}
		}
		if(closestRoom > -1) {
		}
		connections[r] = closestRoom;
	}
	connectRooms(roomArray, connections, mater);
}

function connectRooms(roomArray, connections, mater) {
	var currRoom = null;
	var joinRoom = null;
	for(var r = 0; r < roomArray.length; r++) {
		if(connections[r] > -1) {
			currRoom = roomArray[r];
			joinRoom = roomArray[connections[r]];
			if(joinRoom.startX > currRoom.endX || joinRoom.endX < currRoom.startX) {
				// horizontal join
				var placeFrom = Math.max(currRoom.startY, joinRoom.startY);
				var placeTo = Math.min(currRoom.endY, joinRoom.endY);
				var place = getRandom(placeTo - placeFrom) + placeFrom;
				if(joinRoom.startX > currRoom.endX) {
					for(var x = currRoom.endX + 1; x <= joinRoom.startX - 1; x++) {
						hexarray[x][place].addBase(HEX_WHOLE, mater);
					}
				}
				else {
					for(var x = joinRoom.endX + 1; x <= currRoom.startX - 1; x++) {
						hexarray[x][place].addBase(HEX_WHOLE, mater);
					}
				}
			}
			else {
				// vertical join
				var placeFrom = Math.max(currRoom.startX, joinRoom.startX);
				var placeTo = Math.min(currRoom.endX, joinRoom.endX);
				var place = getRandom(placeTo - placeFrom) + placeFrom;
				if(joinRoom.startY > currRoom.endY) {
					for(var y = currRoom.endY + 1; y <= joinRoom.startY - 1; y++) {
						hexarray[place][y].addBase(HEX_WHOLE, mater);
					}
				}
				else {
					for(var y = joinRoom.endY + 1; y <= currRoom.startY - 1; y++) {
						hexarray[place][y].addBase(HEX_WHOLE, mater);
					}
				}
			}
		}
	}
}

function displayRandomRoomsSettings() {
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("random-map-rooms").style.display = CSS_BLOCK_SHOW;
}

function hideRandomRoomsSettings() {
	document.getElementById("random-map-rooms").style.display = CSS_BLOCK_HIDE;
	overlayDiv.style.display = CSS_BLOCK_HIDE;
	shaderDiv.style.display = CSS_BLOCK_HIDE;
}

function generateRandomTunnelDungeon() {
	hideRandomTunnelsSettings();
	var saturationPercent = 0.25;
	var densitySelector = document.randomtunnelelems.tunneldensity;
	for(var d = 0; d < densitySelector.length; d++) {
		if(densitySelector[d].checked) {
			saturationPercent = densitySelector[d].value;
		}
	}
	var totalHexes = hexarray.length * hexarray[0].length;
	var toFillHexes = Math.round(totalHexes * saturationPercent);
	clearMapGrid();
	var pathMater = activeMaterial;
	while(countMaterialFills(pathMater) < toFillHexes) {
		randomTunnel(pathMater);
	}
	var fillMater = (pathMater == 16 ? 12 : 16);
	joinTunnels(pathMater, fillMater);
	var fillSeed = fillFinder(fillMater);
	floodFill(hexarray, fillSeed.x, fillSeed.y, fillMater, pathMater);
	drawMap();
}

function randomTunnel(pathMater) {
	var addPath = false;
	while(!addPath) {
		addPath = true;
		var randX = getRandom(hexarray.length);
		var randY = getRandom(hexarray[0].length);
		var rota = getRandom(6);
		if(hexarray[randX][randY].bases[0].material != MATERIAL_EMPTY) {
			addPath = false;
		}
		if(addPath) {
			randomTunnelPlot(randX, randY, rota, pathMater);
		}
	}
}

function randomTunnelPlot(x, y, direction, pathMater) {
	var hexPt = getHexAt(x, y, direction);
	if(hexPt.x == OFFGRID || hexPt.y == OFFGRID) {
		return;
	}
	if(hexarray[hexPt.x][hexPt.y].bases[0].material != MATERIAL_EMPTY) {
		return;
	}
	var pathAct = getRandom(8);
	switch(pathAct) {
		case 0:
		case 1:
		case 2:
		case 3:
		case 4:
		case 5:
			// Continue path
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WHOLE, pathMater);
			break;
		case 6:
			// Turn counterclockwise
			direction = direction - 5;
			if(direction < 0) { direction += 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WHOLE, pathMater);
			break;
		case 7:
			// Turn clockwise
			direction = direction - 1;
			if(direction < 0) { direction = direction + 6; }
			hexarray[hexPt.x][hexPt.y].addBase(HEX_WHOLE, pathMater);
			break;
	}
	randomTunnelPlot(hexPt.x, hexPt.y, direction, pathMater);
}

function joinTunnels(pathMater, fillMater) {
	// fill from initial seed point
	var seeking = true;
	var seekCol = 0;
	var seekRow = 0;
	while(seeking) {
		if(hexarray[seekCol][seekRow].bases[0].material == pathMater) {
			floodFill(hexarray, seekCol, seekRow, pathMater, fillMater);
			seeking = false;
		}
		if(seeking) {
			seekRow++;
			if(seekRow >= hexarray[seekCol].length) {
				seekRow = 0;
				seekCol++;
				if(seekCol >= hexarray.length) {
					seeking = false;
				}
			}
		}
	}
	var tunnelCount = countMaterialFills(pathMater);
	while(tunnelCount > 0) {
		var closestFillHex = new Point(0, 0);
		var closestTunnelHex = new Point(0, 0);
		var shortestDistance = hexarray.length * hexarray[0].length;
		var testDistance = shortestDistance;
		for(var x = 0; x < hexarray.length; x++) {
			for(var y = 0; y < hexarray[x].length; y++) {
				if(hexarray[x][y].bases[0].material == fillMater) {
					for(var x2 = 0; x2 < hexarray.length; x2++) {
						for(var y2 = 0; y2 < hexarray[x2].length; y2++) {
							if(hexarray[x2][y2].bases[0].material == pathMater) {
								testDistance = Math.abs(x - x2) + Math.abs(y - y2);
								if(testDistance < shortestDistance) {
									shortestDistance = testDistance;
									closestFillHex.x = x;
									closestFillHex.y = y;
									closestTunnelHex.x = x2;
									closestTunnelHex.y = y2;
								}
							}
						}
					}
				}
			}
		}
		connectTunnels(hexarray, closestFillHex.x, closestFillHex.y, closestTunnelHex.x, closestTunnelHex.y, pathMater, fillMater);
		tunnelCount = countMaterialFills(pathMater);
	}
}

function connectTunnels(dataarray, startX, startY, endX, endY, pathMater, fillMater) {
	var posX = startX;
	var posY = startY;
	var nextX = 0;
	var nextY = 0;
	var fillX = -1;
	var fillY = -1;
	while(!(posX == endX && posY == endY)) {
		if(startX == endX) {
			nextX = 0;
		}
		else {
			if(posX < endX) {
				nextX = 1;
			}
			else if(posX > endX) {
				nextX = -1;
			}
			else {
				nextX = 0;
			}
		}
		if(startY == endY) {
			nextY = 0;
		}
		else {
			if(posY < endY) {
				nextY = 1;
			}
			else if(posY > endY) {
				nextY = -1;
			}
			else {
				nextY = 0;
			}
		}
		if(nextX == 0 && nextY == 1) {
			posY++;
		}
		else if(nextX == 0 && nextY == -1) {
			posY--;
		}
		else {
			// different horizontal rules for odd and even hexes
			if(posX % 2 == 1) {
				if(nextX == -1 && nextY == -1) {
					posX--;
				}
				else if(nextX == -1 && nextY == 0) {
					posX--;
				}
				else if(nextX == -1 && nextY == 1) {
					posX--;
					posY++;
				}
				else if(nextX == 1 && nextY == -1) {
					posX++;
				}
				else if(nextX == 1 && nextY == 0) {
					posX++;
				}
				else if(nextX == 1 && nextY == 1) {
					posX++;
					posY++;
				}
			}
			else {
				if(nextX == -1 && nextY == -1) {
					posX--;
					posY--;
				}
				else if(nextX == -1 && nextY == 0) {
					posX--;
				}
				else if(nextX == -1 && nextY == 1) {
					posX--;
				}
				else if(nextX == 1 && nextY == -1) {
					posX++;
					posY--;
				}
				else if(nextX == 1 && nextY == 0) {
					posX++;
				}
				else if(nextX == 1 && nextY == 1) {
					posX++;
				}
			}
		}
		dataarray[posX][posY].addBase(HEX_WHOLE, pathMater);
		if(fillX == -1) { fillX = posX; }
		if(fillY == -1) { fillY = posY; }
	}
	floodFill(dataarray, fillX, fillY, pathMater, fillMater);
}

function displayRandomTunnelsSettings() {
	shaderDiv.style.display = CSS_BLOCK_SHOW;
	overlayDiv.style.display = CSS_BLOCK_SHOW;
	document.getElementById("random-map-tunnels").style.display = CSS_BLOCK_SHOW;
}

function hideRandomTunnelsSettings() {
	document.getElementById("random-map-tunnels").style.display = CSS_BLOCK_HIDE;
	overlayDiv.style.display = CSS_BLOCK_HIDE;
	shaderDiv.style.display = CSS_BLOCK_HIDE;
}

function floodFill(dataarray, posX, posY, matchMater, fillMater) {
	if(dataarray[posX][posY].bases[0].material == matchMater) {
		dataarray[posX][posY].addBase(HEX_WHOLE, fillMater);
	}
	else {
		return;
	}
	if(posX >= 1) {
		floodFill(dataarray, posX - 1, posY, matchMater, fillMater);
	}
	if(posY >= 1) {
		floodFill(dataarray, posX, posY - 1, matchMater, fillMater);
	}
	if(posX < dataarray.length - 1) {
		floodFill(dataarray, posX + 1, posY, matchMater, fillMater);
	}
	if(posY < dataarray[posX].length - 1) {
		floodFill(dataarray, posX, posY + 1, matchMater, fillMater);
	}
	if(posX % 2 == 0) {
		if(posX >= 1 && posY >= 1) {
			floodFill(dataarray, posX - 1, posY - 1, matchMater, fillMater);
		}
		if(posX < dataarray.length - 1 && posY >= 1) {
			floodFill(dataarray, posX + 1, posY - 1, matchMater, fillMater);
		}
	}
	else {
		if(posX >= 1 && posY < dataarray[posX].length - 1) {
			floodFill(dataarray, posX - 1, posY + 1, matchMater, fillMater);
		}
		if(posX < dataarray.length - 1 && posY < dataarray[posX].length - 1) {
			floodFill(dataarray, posX + 1, posY + 1, matchMater, fillMater);
		}
	}
}

function countMaterialFills(mater) {
	var mcount = 0;
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			if(hexarray[x][y].bases[0].material == mater) {
				mcount++;
			}
		}
	}
	return mcount;
}

function fillFinder(mater) {
	for(var x = 0; x < hexarray.length; x++) {
		for(var y = 0; y < hexarray[x].length; y++) {
			if(hexarray[x][y].bases[0].material == mater) {
				return new Point(x, y);
			}
		}
	}
	return new Point(-1, -1);
}